import datetime
import html
import os
import webbrowser

import addonHandler
import api
import config
import globalPluginHandler
import globalVars
import gui
import inputCore
import ui
import vision
import wx
from locationHelper import RectLTRB
from logHandler import log

try:
    import vision
except ImportError:
    raise RuntimeError("No vision framework support")
from . import isUSARunning, libUSA


def skipEventAndCall(handler):
    def wrapWithEventSkip(event):
        if event:
            event.Skip()
        return handler()

    return wrapWithEventSkip


# We need to initialize translation and localization support:
addonHandler.initTranslation()


class GlobalPlugin(globalPluginHandler.GlobalPlugin):
    """
    USA globalPlugin class of which an instance is created, i.e. upon loading of NVDA.
    This is used to initialize USA Remote if applicable. Also it sets the help file direction in the settings.
    For the Object Reader we also need the mouseMove event.
    """

    def isUSAEnabled(self):
        veps = vision.handler.getActiveProviderInstances()
        for vep in veps:
            try:
                if vep.getSettings().getId().lower().strip() == "usa":
                    return True
            except:
                pass
        return False

    def enableUSA(self):
        if self.isUSAEnabled():
            return
        vepsInfo = vision.handler.getProviderList()
        usaProviderInfo = None
        for vep in vepsInfo:
            if vep.providerId == "usa":
                usaProviderInfo = vep

        if usaProviderInfo:
            vision.handler.initializeProvider(usaProviderInfo)
        else:
            log.error("USA provider not found! Could not enable provider")

    def disableUSA(self):
        if not self.isUSAEnabled():
            return
        vepsInfo = vision.handler.getProviderList()
        usaProviderInfo = None
        for vep in vepsInfo:
            if vep.providerId == "usa":
                usaProviderInfo = vep

        if usaProviderInfo:
            vision.handler.terminateProvider(usaProviderInfo)
        else:
            log.error("USA provider not found! could not disable provider")

    def getLicenseDialog(self):
        from .licenseDialog import licenseDialog

        return licenseDialog.USALicenseDialog
        # get the license dialog so that the vision enhancement provider can use it.

    def licenseUSA(self):
        gui.mainFrame.prePopup()
        licenseDialog = self.getLicenseDialog()
        with licenseDialog(gui.mainFrame) as dlg:
            dlg.ShowModal()
        gui.mainFrame.postPopup()

    def USASimpleSettings(self, startup=False):
        from . import simpleSettings

        gui.mainFrame.prePopup()
        with simpleSettings.USASimpleSettings(gui.mainFrame, startup) as dlg:
            dlg.ShowModal()
        gui.mainFrame.postPopup()

    def USAAdvancedSettings(self):
        gui.mainFrame.prePopup()

        def onVisionCommand():
            gui.mainFrame.popupSettingsDialog(
                gui.NVDASettingsDialog, gui.settingsDialogs.VisionSettingsPanel
            )

        wx.CallAfter(onVisionCommand)
        gui.mainFrame.postPopup()

    def updateMenu(self, evt):
        self.quickUSA_menuitem.Enable(self.isUSAEnabled())
        self.advanceUSA_menuitem.Enable(self.isUSAEnabled())

    def create_menu(self):
        self.menu = wx.Menu()

        self.quickUSA_menuitem = self.menu.Append(wx.ID_ANY, _("Quick settings"))

        gui.mainFrame.sysTrayIcon.Bind(
            wx.EVT_MENU,
            skipEventAndCall(self.USASimpleSettings),
            self.quickUSA_menuitem,
        )

        self.licenseUSA_menuItem = self.menu.Append(wx.ID_ANY, _("License"))

        gui.mainFrame.sysTrayIcon.Bind(
            wx.EVT_MENU, skipEventAndCall(self.licenseUSA), self.licenseUSA_menuItem
        )
        self.help_item = self.menu.Append(
            wx.ID_ANY,
            _("Help"),
            _("Manual"),
        )

        gui.mainFrame.sysTrayIcon.Bind(
            wx.EVT_MENU,
            lambda e: os.startfile(addonHandler.getCodeAddon().getDocFilePath()),
            self.help_item,
        )

        def gesture_to_html(g):
            g = g.split(":", 1)[1].title()
            keys = g.split("+")

            # Custom sorting explained: https://stackoverflow.com/a/64593268
            custom_order = {
                c: i for i, c in enumerate(["Nvda", "Control", "Alt", "Shift"])
            }

            def customSort(l):
                l.sort(key=lambda c: custom_order.get(c, len(custom_order)))

            customSort(keys)
            g = " + ".join(keys)
            return g

        def category_has_script_gestures(category_scripts):
            for gestures_script_info in category_scripts.values():
                if len(gestures_script_info.gestures) > 0:
                    return True
            return False

        def create_gesture_help(e):
            str_title = _("USA Shortcut List")
            str_description = _("Description")
            str_shortcuts = _("Shortcuts")
            str_category = _("Common Shortcuts")
            html_content = f"""
			<!DOCTYPE html>
			<html lang="en">
			<head>
				<meta charset="UTF-8">
				<meta name="viewport" content="width=device-width, initial-scale=1.0">
				<title>{str_title}</title>
				<style>
					body {{
						font-family: Arial, sans-serif;
						margin: 20px;
					}}
					table {{
						width: 100%;
						border-collapse: collapse;
					}}
					th, td {{
						border: 1px solid #ddd;
						padding: 8px;
						text-align: left;
					}}
					th {{
						background-color: #f2f2f2;
					}}
				</style>
			</head>
			<body>
				<h1>{str_title}</h1>
                <h2>{str_category}</h2>
				<table>
					<thead>
					<tr>
						<th>{str_description}</th>
						<th colspan="5">{str_shortcuts}</th>
					</tr>
					</thead>
					<tbody>"""

            def add_from_gesture_mapping(html_content, gesture_mapping):
                for description, gestures_script_info in gesture_mapping:
                    gestures: list[str] = gestures_script_info.gestures
                    # Skip scripts without assigned gestures
                    if len(gestures) == 0:
                        continue
                    html_content += f"""
						<tr>
							<td>{description}</td>"""

                    # Truncate gestures if there are more than 3 for the script
                    if len(gestures) > 3:
                        gestures = gestures[:3]

                    for g in gestures:
                        if not g.startswith("kb"):
                            log.warning(
                                f"Gesture '{g}' for '{description}' not starting with 'kb'"
                            )
                            continue
                        try:
                            g = gesture_to_html(g)

                        except IndexError:
                            log.error(f"Invalid gesture without colon: {g}")

                        html_content += f"""
							<td>{g}</td>"""
                    html_content += "</tr>"
                return html_content

            try:
                scripts = inputCore.manager.getAllGestureMappings()[
                    "Ultimate Screen Access"
                ]

                # Add most common shortcut gestures at the top
                common_script_names = [
                    "increaseMagnification",
                    "decreaseMagnification",
                    "toggleColorScheme",
                    "toggleMagnifier",
                    "switchMagnifierMode",
                ]

                # Sort scripts to show common ones first
                def filter_scripts(item):
                    _, gestures_script_info = item
                    script_name = gestures_script_info.scriptName
                    if script_name in common_script_names:
                        return True
                    else:
                        return False

                common_scripts = sorted(
                    filter(filter_scripts, scripts.items()),
                    key=lambda x: common_script_names.index(x[1].scriptName),
                )

                html_content = add_from_gesture_mapping(html_content, common_scripts)

                # Manually add mouse scroll zoom gesture
                str_zoom = _("Zoom")
                str_mouse_wheel = _("Mouse Wheel")
                html_content += f"""
						<tr>
							<td>{str_zoom}</td>
							<td>Nvda + {str_mouse_wheel}</td>
						</tr>"""

                str_category = _("All USA Shortcuts")

                html_content += f"""
						</tbody>
					</table>
                    <h2>{str_category}</h2>
                    <table>
                        <thead>
                        <tr>
                            <th>{str_description}</th>
                            <th colspan="5">{str_shortcuts}</th>
                        </tr>
                        </thead>
                    <tbody>"""

                html_content = add_from_gesture_mapping(html_content, scripts.items())

                html_content += """
						</tbody>
					</table>"""

                # Add all other NVDA shortcuts to own table

                str_nvda_title = _("NVDA Shortcut List")
                html_content += f"""
                <h1>{str_nvda_title}</h1>"""

                for (
                    category,
                    scripts,
                ) in inputCore.manager.getAllGestureMappings().items():
                    if category == "Ultimate Screen Access":
                        continue

                    if not category_has_script_gestures(scripts):
                        continue

                    html_content += f"""
                    <h2>{category}</h2>
                    <table>
                        <thead>
                        <tr>
                            <th>{str_description}</th>
                            <th colspan="5">{str_shortcuts}</th>
                        </tr>
                        </thead>
                    <tbody>"""

                    for description, gestures_script_info in scripts.items():
                        gestures: list[str] = gestures_script_info.gestures
                        # Skip scripts without assigned gestures
                        if len(gestures) == 0:
                            continue
                        html_content += f"""
                            <tr>
                                <td>{description}</td>"""

                        # Truncate gestures if there are more than 3 for the script
                        if len(gestures) > 3:
                            gestures = gestures[:3]

                        for g in gestures:
                            try:
                                # Ignore touch screen gestures unless we are
                                # in the touchscreen category
                                if g.startswith("ts") and category != "Touchscreen":
                                    continue

                                g = gesture_to_html(g)

                            except IndexError:
                                log.error(f"Invalid gesture without colon: {g}")

                            html_content += f"""
                                <td>{g}</td>"""
                        html_content += "</tr>"

                    html_content += """
                            </tbody>
                        </table>"""

                html_content += """
				</body>
				</html>
                """

                # Write the HTML content to a file
                html_path = f"{globalVars.appArgs.configPath}\\usa-keybinds.html"
                with open(html_path, "w", encoding="utf-8") as file:
                    file.write(html_content)

                webbrowser.open(f"file://{html_path}")
            except Exception as e:
                log.error(
                    f"No valid gesture mappings for 'Ultimate Screen Access' found: {e}"
                )

        self.shortcut_list_item = self.menu.Append(
            wx.ID_ANY,
            _("Shortcut list"),
            _("Shortcut list"),
        )

        gui.mainFrame.sysTrayIcon.Bind(
            wx.EVT_MENU, create_gesture_help, self.shortcut_list_item
        )

        self.advanceUSA_menuitem = self.menu.Append(wx.ID_ANY, _("Advanced settings"))

        gui.mainFrame.sysTrayIcon.Bind(
            wx.EVT_MENU,
            skipEventAndCall(self.USAAdvancedSettings),
            self.advanceUSA_menuitem,
        )

        self.submenu_item = gui.mainFrame.sysTrayIcon.menu.Insert(
            2, wx.ID_ANY, _("&Ultimate Screen Access"), self.menu
        )
        gui.mainFrame.sysTrayIcon.Bind(wx.EVT_MENU_HIGHLIGHT_ALL, self.updateMenu)

    def checkIsRemote(self) -> bool:
        allProviders = vision.handler.getProviderList()
        for provider in allProviders:
            if provider.providerId == "usa":
                log.debug(
                    f"USA isRemote is {provider.providerClass.getSettings().isRemote }"
                )
                return provider.providerClass.getSettings().isRemote
        return False

    def checkDemoAndRunning(self):
        if not self.checkIsRemote():
            other_popups_open = self.checkForOtherPopups()
            licenseDialog = self.getLicenseDialog()
            with licenseDialog(gui.mainFrame) as dlg:
                if dlg.shouldShowAtStart:
                    # Only check for other popups if the license dialog should be shown
                    if other_popups_open:
                        wx.CallLater(1000, self.checkDemoAndRunning)
                        return
                    # If the license dialog would be shown, do not start USA in a secure desktop env
                    if globalVars.appArgs.secure:
                        return
                    gui.mainFrame.prePopup()
                    dlg.ShowModal()
                    gui.mainFrame.postPopup()
                else:
                    log.debugWarning("not showing license dialog on start")

        wx.CallLater(1000, self.checkRunning)

    def checkRunning(self):
        if self.checkIsRemote():

            allProviders = vision.handler._allProviders
            for provider in allProviders:
                if provider.providerId == "usa" and (
                    not provider in vision.handler.getActiveProviderInfos()
                ):
                    log.info("starting USA since it runs in Remote mode")
                    vision.handler.initializeProvider(provider)
            return

        if isUSARunning.shouldAskForStart():
            if self.checkForOtherPopups():
                wx.CallLater(1000, self.checkRunning)
                return
            with isUSARunning.isUSARunning(gui.mainFrame) as dlg:
                gui.mainFrame.prePopup()
                dlg.ShowModal()
                gui.mainFrame.postPopup()

        usaInst = None
        veps = vision.handler.getActiveProviderInstances()
        for vep in veps:
            try:
                if vep.getSettings().getId().lower().strip() == "usa":
                    usaInst = vep

            except:
                pass
        if not usaInst:
            return

        if usaInst.getSettings().get_quick_settings_on_startup():
            self.USASimpleSettings(True)

    def checkForOtherPopups(self):
        active_popups = []

        if len(gui.mainFrame.Children) > 0:
            for c in gui.mainFrame.Children:
                try:
                    # Check if window is valid, not being deleted, and actually visible
                    if (
                        c
                        and hasattr(c, "IsShown")
                        and c.IsShown()
                        and not c.IsBeingDeleted()
                        and hasattr(c, "IsModal")
                        and c.IsModal()
                    ):

                        active_popups.append(c)
                        log.info(
                            f"Active modal dialog: {getattr(c, 'Title', 'Unknown')}"
                        )
                except (AttributeError, RuntimeError) as e:
                    # Window might have been destroyed between checks
                    log.debug(f"Window no longer valid: {e}")
                    continue

        if active_popups:
            log.info(f"Found {len(active_popups)} active popup(s)")
            return True
        else:
            return False

    def enableLibUSA(self):
        if self.checkIsRemote():
            log.info("not initializing libUSA since it runs in Remote mode")
            self.canStartUSA_VEP = True
            return
        try:
            libUSA.handler.Load()
            log.info(f"load libUSA")
        except Exception as e:
            log.error(f"could not load libUSA: {e}")
            return
        self.canStartUSA_VEP = True
        wx.CallAfter(libUSA.handler.IsLicensed)
        wx.CallLater(1000, self.checkDemoAndRunning)

    def __init__(self):
        self.canStartUSA_VEP = False
        super().__init__()

        self.create_menu()
        self.remoteHandler = None
        wx.CallAfter(self.initializeRemote)
        wx.CallLater(int(config.conf.get("USA_timeout", 100)), self.enableLibUSA)
        log.info("USA: plugin initialized")

    def terminate(self):
        if self.remoteHandler:
            self.remoteHandler.terminate()
            self.remoteHandler = None
        if self.submenu_item:
            gui.mainFrame.sysTrayIcon.menu.Remove(self.submenu_item)
            self.submenu_item.Destroy()
            self.submenu_item = None
        super().terminate()

    def initializeRemote(self):
        try:
            self.remoteHandler = RemoteHandler()
        except (NotImplementedError, RuntimeError):
            log.info("USA: USA remote capabilities not available")
            self.remoteHandler = None
        else:
            log.info("USA: USA remote capabilities available")


class RemoteHandler(object):

    def __init__(self):
        self.connectionType = None
        self.transport = None
        # check if unicorn is running
        try:
            self.remoteClient = next(
                p
                for p in globalPluginHandler.runningPlugins
                if p.__module__ == "globalPlugins.unicorn"
            )
        except StopIteration:
            raise RuntimeError
        # register callbacks
        self.remoteClient.callback_manager.register_callback(
            "transport_connected", self.onTransportConnected
        )
        self.remoteClient.callback_manager.register_callback(
            "transport_disconnected", self.onTransportDisconnected
        )

        if self.remoteClient.slave_transport:
            self.onTransportConnected(
                connection_type="slave", transport=self.remoteClient.slave_transport
            )
        elif self.remoteClient.master_transport:
            self.onTransportConnected(
                connection_type="master", transport=self.remoteClient.master_transport
            )

    def sendVisionCommandOverUnicorn(self, command, **kwargs):
        if self.transport:
            self.transport.send(type="vision", command=command, **kwargs)
        else:
            log.warning(
                "Unicorn transport for USA not initialized, nvda isn't connected as client and server on both sides."
            )

    def terminate(self):
        self.unregisterRemoteTransport()
        self.remoteClient.callback_manager.unregister_callback(
            "transport_connected", self.onTransportConnected
        )
        self.remoteClient.callback_manager.unregister_callback(
            "transport_disconnected", self.onTransportDisconnected
        )
        self.remoteClient = None

    def unregisterRemoteTransport(self):
        if self.transport:
            try:
                self.transport.callback_manager.unregister_callback(
                    "msg_vision", self.handleVisionPacket
                )
            except ValueError:
                pass

    def onTransportConnected(self, connection_type, transport, *args, **kwargs):
        log.debugWarning("unicorn/USA: onTransportConnected")
        transport.callback_manager.register_callback(
            "msg_vision", self.handleVisionPacket
        )
        self.connectionType = connection_type
        self.transport = transport

    def onTransportDisconnected(self, *args, **kwargs):
        # We do not have to unregister callbacks here
        log.debugWarning("unicorn/USA: onTransportDisconnected")
        if self.transport:
            self.transport.callback_manager.unregister_callback(
                "msg_vision", self.handleVisionPacket
            )
        self.connectionType = None
        self.transport = None

    def handleVisionPacket(self, command, *args, **kwargs):
        if not vision.handler:
            return
        try:
            usa = vision.handler.getProviderInstance(
                vision.handler.getProviderInfo("usa")
            )
        except LookupError:
            usa = None
        if not usa:
            return
        commandMapping = usa.getUnicornCommandMapping()
        kwargs = usa.deserializeUnicornArguments(command, **kwargs)
        log.info(f"{command} {kwargs}")
        try:
            if kwargs:
                commandMapping[command](**kwargs)
            else:
                commandMapping[command]()
        except Exception as e:
            log.error(
                f"An error occured when handling a vision package for USA: \n {e}"
            )
