import os
from ctypes import *
from ctypes.wintypes import *

import addonHandler
import config
import gui
import ui
import vision
import watchdog
import wx
from gui import nvdaControls
from logHandler import log

from . import libUSA
from .dialogs import dialogs

addonHandler.initTranslation()


class USASimpleSettings(wx.Dialog):

    def __init__(self, parent, shownOnStartup=False):
        usaInst = None
        veps = vision.handler.getActiveProviderInstances()
        for vep in veps:
            try:
                if vep.getSettings().getId().lower().strip() == "usa":
                    usaInst = vep
                    break
            except:
                pass
        if not usaInst:
            return

        self.settings = usaInst.getSettings()
        self.dialogs = [
            dialogs.USACustomColorSchemeDialog,
            dialogs.USAFGColorSchemeDialog,
        ]
        self.EnableLayoutAdaptation(wx.DIALOG_ADAPTATION_MODE_ENABLED)

        super().__init__(parent, id=wx.ID_ANY, title=_("Quick settings"))

        font = wx.Font(18, wx.DEFAULT, wx.NORMAL, wx.BOLD)

        main_sizer = wx.BoxSizer(wx.VERTICAL)

        main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)

        self.shouldShowOnStartupCheckbox = None
        ### general USA settings ###
        if shownOnStartup:
            startupGroupText = _("Show on Startup")
            startupGroupSizer = wx.StaticBoxSizer(
                wx.VERTICAL, self, label=startupGroupText
            )
            startupBox = startupGroupSizer.GetStaticBox()
            startupGroup = gui.guiHelper.BoxSizerHelper(self, sizer=startupGroupSizer)
            main_sizer_helper.addItem(startupGroup)

            self.shouldShowOnStartupCheckbox = startupGroup.addItem(
                wx.CheckBox(startupBox, label=_("Show this dialog on startup."))
            )
            self.shouldShowOnStartupCheckbox.SetValue(
                self.settings.get_quick_settings_on_startup()
            )
        ### general USA settings ###
        GeneralGroupText = _("General")
        GeneralGroupSizer = wx.StaticBoxSizer(wx.VERTICAL, self, label=GeneralGroupText)
        GeneralBox = GeneralGroupSizer.GetStaticBox()
        GeneralGroup = gui.guiHelper.BoxSizerHelper(self, sizer=GeneralGroupSizer)
        main_sizer_helper.addItem(GeneralGroup)
        GeneralGroup.addItem(wx.StaticText(GeneralBox, label=_("General settings")))

        try:
            self.bitmapSmoothing = GeneralGroup.addItem(
                wx.CheckBox(
                    GeneralBox,
                    label=_(
                        "Enable bitmap &smoothing to increase magnification quality"
                    ),
                )
            )
            self.bitmapSmoothing.SetValue(self.settings.get_bitmap_smoothing())
            self.bitmapSmoothing.Bind(wx.EVT_CHECKBOX, self.on_bitmap_smoothing_changed)
        except Exception as e:
            log.error(f"failed bitmap smoothing checkbox creation with err {e}")

        try:
            self.setMagnificationByLevel = GeneralGroup.addItem(
                wx.CheckBox(
                    GeneralBox,
                    label=_("Let USA set the magnification level by fixed levels"),
                )
            )
            self.setMagnificationByLevel.SetValue(
                self.settings.get_magnification_fixed_levels()
            )
            self.setMagnificationByLevel.Bind(
                wx.EVT_CHECKBOX, self.on_magnification_fixed_levels_changed
            )
        except Exception as e:
            log.error(f"failed magnification by level checkbox creation with err {e}")

        try:
            stepLevelComboboxText = _("Step Level")
            self.stepLevelStrings = [
                ii.displayName
                for ii in list(self.settings.availableSteplevels.values())
            ]
            self.stepLevelCombobox = GeneralGroup.addLabeledControl(
                stepLevelComboboxText, wx.Choice, choices=self.stepLevelStrings
            )
            self.stepLevelCombobox.SetSelection(
                self.stepLevelStrings.index(
                    str(self.settings.get_magnification_step_level())
                )
            )
            self.stepLevelCombobox.Bind(wx.EVT_CHOICE, self.onMagStepLevelChanged)
        except Exception as e:
            log.error(f"failed step level combobox creation with err {e}")

        try:
            mag_level_combo_box_text = _("Magnification &level")
            self.mag_level_strings = [
                ii.displayName
                for ii in list(
                    self.settings._get_availableMagnificationlevels().values()
                )
            ]
            self.mag_level_combo_box = GeneralGroup.addLabeledControl(
                mag_level_combo_box_text, wx.Choice, choices=self.mag_level_strings
            )
            self.mag_level_combo_box.SetSelection(
                self.mag_level_strings.index(
                    str(self.settings.get_magnification_level_by_window(0))
                )
            )
            self.mag_level_combo_box.Bind(wx.EVT_CHOICE, self.on_mag_level_changed)
        except Exception as e:
            log.error(f"failed mag level with err {e}")

        try:
            boundary_combobox_text = _("Mouse boun&dary")
            self.boundary_level_choices = [
                ii.displayName
                for ii in list(self.settings._availableBoundaries.values())
            ]
            self.boundary_level_values = [
                ii.id for ii in list(self.settings._availableBoundaries.values())
            ]
            self.boundary_combobox = GeneralGroup.addLabeledControl(
                boundary_combobox_text, wx.Choice, choices=self.boundary_level_choices
            )
            boundary_value = int(self.settings.get_boundary() * 100)
            if boundary_value == -1:
                boundary_index = self.boundary_level_values.index([-1, -1, -1, -1])
            else:
                boundary_index = self.boundary_level_values.index(
                    [
                        boundary_value,
                        boundary_value,
                        100 - boundary_value,
                        100 - boundary_value,
                    ]
                )
            self.boundary_combobox.SetSelection(boundary_index)
            self.boundary_combobox.Bind(wx.EVT_CHOICE, self.on_boundary_changed)
        except Exception as e:
            log.error(f"failed boundary combobox creation with err {e}")

        ### settings for Object Reader ###
        objectReaderGroupText = _("Object Reader")
        objectReaderGroupSizer = wx.StaticBoxSizer(
            wx.VERTICAL, self, label=objectReaderGroupText
        )
        objectReaderBox = objectReaderGroupSizer.GetStaticBox()
        objectReaderGroup = gui.guiHelper.BoxSizerHelper(
            self, sizer=objectReaderGroupSizer
        )
        main_sizer_helper.addItem(objectReaderGroup)
        objectReaderGroup.addItem(
            wx.StaticText(
                objectReaderBox,
                label=_(
                    "Enable users to apply continuous reading for objects below the mouse."
                ),
            )
        )
        self.objectReaderMode = objectReaderGroup.addItem(
            wx.CheckBox(objectReaderBox, label=_("Activate Object Reader mode"))
        )
        self.objectReaderMode.SetValue(self.settings.get_object_reader())
        self.objectReaderMode.Bind(wx.EVT_CHECKBOX, self.on_object_reader_changed)
        self.followWord = objectReaderGroup.addItem(
            wx.CheckBox(
                objectReaderBox,
                label=_("Enable following of words in continuous reading mode"),
            )
        )
        self.followWord.SetValue(self.settings.get_word_follower())
        self.followWord.Bind(wx.EVT_CHECKBOX, self.on_word_follower_changed)
        self.followWordRect = objectReaderGroup.addItem(
            wx.CheckBox(
                objectReaderBox,
                label=_("Enable highlighting of words in continuous reading mode"),
            )
        )
        self.followWordRect.SetValue(self.settings.get_word_follower_show_rect())
        self.followWordRect.Bind(
            wx.EVT_CHECKBOX, self.on_word_follower_show_rect_changed
        )
        self.followWordJumpCaret = objectReaderGroup.addItem(
            wx.CheckBox(
                objectReaderBox,
                label=_(
                    "Let the caret jump with the read words in continuous reading mode"
                ),
            )
        )
        self.followWordJumpCaret.SetValue(self.settings.get_word_follower_caret_jump())
        self.followWordJumpCaret.Bind(
            wx.EVT_CHECKBOX, self.on_word_follower_caret_jump_changed
        )

        aboutObjectReader = objectReaderGroup.addItem(
            wx.Button(objectReaderBox, label=_("About"))
        )
        aboutObjectReader.Bind(wx.EVT_BUTTON, self.aboutObjectReader)

        ### settings for color schemes capabilities ###
        colorGroupText = _("Color schemes")
        colorGroupSizer = wx.StaticBoxSizer(wx.VERTICAL, self, label=colorGroupText)
        colorBox = colorGroupSizer.GetStaticBox()
        colorGroup = gui.guiHelper.BoxSizerHelper(self, sizer=colorGroupSizer)
        main_sizer_helper.addItem(colorGroup)
        colorGroup.addItem(
            wx.StaticText(colorBox, label=_("Settings for color schemes"))
        )

        try:
            color_scheme_combo_box_text = _("Color sc&heme")
            self.color_scheme_strings = [
                ii.displayName
                for ii in list(
                    self.settings._get_available_color_schemes_by_window(0).values()
                )
            ]
            self.color_scheme_key_strings = [
                ii
                for ii in list(
                    self.settings._get_available_color_schemes_by_window(0).keys()
                )
            ]
            self.color_scheme_combo_box = colorGroup.addLabeledControl(
                color_scheme_combo_box_text,
                wx.Choice,
                choices=self.color_scheme_strings,
            )
            self.color_scheme_combo_box.SetSelection(
                self.color_scheme_key_strings.index(
                    str(self.settings.get_color_scheme_by_window(0))
                )
            )
            self.color_scheme_combo_box.Bind(
                wx.EVT_CHOICE, self.on_selected_color_scheme_change
            )
        except Exception as e:
            log.error(f"failed color scheme with err {e}")

        modifierBoxLabel = _("&Select color schemes")
        colorSchemeDict = self.settings._get_all_color_schemes_by_window(0)

        self.modifierChoices = [
            colorSchemeDict[key].displayName for key in colorSchemeDict
        ]
        self.modifierList = colorGroup.addLabeledControl(
            modifierBoxLabel,
            nvdaControls.CustomCheckListBox,
            choices=self.modifierChoices,
        )
        checked_items = []
        try:
            get_scs = self.settings.get_selected_color_scheme_by_window_key
            if get_scs(0, "default") == "True":
                checked_items.append(
                    self.modifierChoices.index(colorSchemeDict["default"].displayName)
                )
            if get_scs(0, "inverted") == "True":
                checked_items.append(
                    self.modifierChoices.index(colorSchemeDict["inverted"].displayName)
                )
            if get_scs(0, "gray") == "True":
                checked_items.append(
                    self.modifierChoices.index(colorSchemeDict["gray"].displayName)
                )
            if get_scs(0, "high-contrast") == "True":
                checked_items.append(
                    self.modifierChoices.index(
                        colorSchemeDict["high-contrast"].displayName
                    )
                )
            if get_scs(0, "inverted-high-contrast") == "True":
                checked_items.append(
                    self.modifierChoices.index(
                        colorSchemeDict["inverted-high-contrast"].displayName
                    )
                )
            if get_scs(0, "blue-light") == "True":
                checked_items.append(
                    self.modifierChoices.index(
                        colorSchemeDict["blue-light"].displayName
                    )
                )
            if get_scs(0, "custom") == "True":
                checked_items.append(
                    self.modifierChoices.index(colorSchemeDict["custom"].displayName)
                )
        except Exception as e:
            log.error(f"creating failed with err {e}")

        self.modifierList.CheckedItems = checked_items
        self.modifierList.Select(0)
        self.modifierList.Bind(wx.EVT_CHECKLISTBOX, self.on_checkbox_changed)

        main_sizer_helper.addDialogDismissButtons(
            self.CreateButtonSizer(wx.OK | wx.CANCEL)
        )
        self.Bind(wx.EVT_BUTTON, self.on_ok, id=wx.ID_OK)
        self.Bind(wx.EVT_BUTTON, self.on_discard, id=wx.ID_CANCEL)
        main_sizer.Add(
            main_sizer_helper.sizer,
            border=gui.guiHelper.BORDER_FOR_DIALOGS,
            flag=wx.ALL,
        )
        main_sizer.Fit(self)
        self.SetSizer(main_sizer)
        self.Center(wx.BOTH)
        self.Show()

    def build_selected_color_scheme_settings(self):
        try:
            color_scheme_dict = self.settings._get_all_color_schemes_by_window(0)
            set_scsk = self.settings.set_selected_color_scheme_by_window_key

            for i in range(0, self.settings.get_total_amount_windows()):
                set_scsk(
                    i,
                    "default",
                    str(
                        self.modifierList.IsChecked(
                            self.modifierChoices.index(
                                color_scheme_dict["default"].displayName
                            )
                        )
                    ),
                )
                set_scsk(
                    i,
                    "inverted",
                    str(
                        self.modifierList.IsChecked(
                            self.modifierChoices.index(
                                color_scheme_dict["inverted"].displayName
                            )
                        )
                    ),
                )
                set_scsk(
                    i,
                    "gray",
                    str(
                        self.modifierList.IsChecked(
                            self.modifierChoices.index(
                                color_scheme_dict["gray"].displayName
                            )
                        )
                    ),
                )
                set_scsk(
                    i,
                    "high-contrast",
                    str(
                        self.modifierList.IsChecked(
                            self.modifierChoices.index(
                                color_scheme_dict["high-contrast"].displayName
                            )
                        )
                    ),
                )
                set_scsk(
                    i,
                    "inverted-high-contrast",
                    str(
                        self.modifierList.IsChecked(
                            self.modifierChoices.index(
                                color_scheme_dict["inverted-high-contrast"].displayName
                            )
                        )
                    ),
                )
                set_scsk(
                    i,
                    "custom",
                    str(
                        self.modifierList.IsChecked(
                            self.modifierChoices.index(
                                color_scheme_dict["custom"].displayName
                            )
                        )
                    ),
                )
                set_scsk(
                    i,
                    "blue-light",
                    str(
                        self.modifierList.IsChecked(
                            self.modifierChoices.index(
                                color_scheme_dict["blue-light"].displayName
                            )
                        )
                    ),
                )

            # Update available color schemes for selected color scheme
            available_color_scheme = (
                self.settings._get_available_color_schemes_by_window(0)
            )
            self.color_scheme_strings = [
                ii.displayName for ii in list(available_color_scheme.values())
            ]
            self.color_scheme_key_strings = [
                ii for ii in list(available_color_scheme.keys())
            ]
            value = self.settings.get_color_scheme_by_window(0)
            self.color_scheme_combo_box.Clear()
            self.color_scheme_combo_box.AppendItems(self.color_scheme_strings)
            self.color_scheme_combo_box.SetSelection(
                self.color_scheme_key_strings.index(str(value))
            )

            log.debug(f"new combobox choices {self.color_scheme_combo_box}")
        except Exception as e:
            log.error(f"error on color_scheme_combo_box {e}")

    def on_checkbox_changed(self, evt):
        self.build_selected_color_scheme_settings()

    def on_bitmap_smoothing_changed(self, evt):
        self.settings.backup_json()
        bitmap_smoothing_value = bool(evt.GetInt())
        try:
            self.settings.set_bitmap_smoothing(bitmap_smoothing_value)
            log.debug(f" set_bitmap_smoothing {bitmap_smoothing_value} succeeded")
        except Exception as e:
            log.error(
                f" set_bitmap_smoothing {bitmap_smoothing_value} failed with err {e}"
            )
        self.settings.UpdateConfig()

    def on_magnification_fixed_levels_changed(self, evt):
        self.settings.backup_json()
        mag_fixed_levels_value = bool(evt.GetInt())
        try:
            self.settings.set_magnification_fixed_levels(mag_fixed_levels_value)
            log.debug(
                f" set_magnification_fixed_levels {mag_fixed_levels_value} succeeded"
            )
        except Exception as e:
            log.error(
                f" set_magnification_fixed_levels {mag_fixed_levels_value} failed with err {e}"
            )
        self.settings.UpdateConfig()

    def onMagStepLevelChanged(self, evt):
        self.settings.backup_json()
        stepLevelIndex = self.stepLevelCombobox.CurrentSelection
        stepLevelValue = float(self.stepLevelStrings[stepLevelIndex])
        try:
            self.settings.set_magnification_step_level(stepLevelValue)
            log.debug(f" set_magnification_step_level {stepLevelValue} succeeded")
        except Exception as e:
            log.error(
                f" set_magnification_step_level {stepLevelValue} failed with err {e}"
            )
        self.settings.UpdateConfig()

    def on_mag_level_changed(self, evt):
        self.settings.backup_json()
        magLevelIndex = self.mag_level_combo_box.CurrentSelection
        magLevelValue = float(self.mag_level_strings[magLevelIndex])
        try:
            for i in range(0, self.settings.get_total_amount_windows()):
                self.settings.set_magnification_level_by_window(i, magLevelValue)
            log.debug(f" set_magnification_level_by_window {magLevelValue} suceeded")
        except Exception as e:
            log.error(
                f" set_magnification_level_by_window {magLevelValue} failed with err {e}"
            )
        self.settings.UpdateConfig()

    def on_boundary_changed(self, evt):
        self.settings.backup_json()
        boundaryIndex = self.boundary_combobox.CurrentSelection
        boundaryValue = self.boundary_level_values[boundaryIndex][0]
        boundaryValueFloat = float(boundaryValue) / 100
        try:
            self.settings.set_boundary(boundaryValueFloat)
            log.debug(f" set_boundary {boundaryValueFloat} succeeded")
        except Exception as e:
            log.error(f" set_boundary {boundaryValueFloat} failed with err {e}")
        self.settings.UpdateConfig()

    def on_object_reader_changed(self, evt):
        self.settings.backup_json()
        object_reader_value = bool(evt.GetInt())
        try:
            self.settings.set_object_reader(object_reader_value)
            log.debug(f" set_object_reader {object_reader_value} succeeded")
        except Exception as e:
            log.error(f" set_object_reader {object_reader_value} failed with err {e}")
        self.settings.UpdateConfig()

    def on_word_follower_changed(self, evt):
        self.settings.backup_json()
        word_follower_value = bool(evt.GetInt())
        try:
            self.settings.set_word_follower(word_follower_value)
            log.debug(f" set_word_follower {word_follower_value} succeeded")
        except Exception as e:
            log.error(f" set_word_follower {word_follower_value} failed with err {e}")
        self.settings.UpdateConfig()

    def on_word_follower_show_rect_changed(self, evt):
        self.settings.backup_json()
        word_follower_show_rect_value = bool(evt.GetInt())
        try:
            self.settings.set_word_follower_show_rect(word_follower_show_rect_value)
            log.debug(
                f" set_word_follower_show_rect {word_follower_show_rect_value} succeeded"
            )
        except Exception as e:
            log.error(
                f" set_word_follower_show_rect {word_follower_show_rect_value} failed with err {e}"
            )
        self.settings.UpdateConfig()

    def on_word_follower_caret_jump_changed(self, evt):
        self.settings.backup_json()
        word_follower_caret_jump_value = bool(evt.GetInt())
        try:
            self.settings.set_word_follower_caret_jump(word_follower_caret_jump_value)
            log.debug(
                f" set_word_follower_caret_jump {word_follower_caret_jump_value} succeeded"
            )
        except Exception as e:
            log.error(
                f" set_word_follower_caret_jump {word_follower_caret_jump_value} failed with err {e}"
            )
        self.settings.UpdateConfig()

    def on_selected_color_scheme_change(self, evt):
        self.settings.backup_json()
        try:
            colorSchemeIndex = self.color_scheme_combo_box.CurrentSelection
            colorSchemeValue = str(self.color_scheme_key_strings[colorSchemeIndex])

            for i in range(0, self.settings.get_total_amount_windows()):
                self.settings.set_color_scheme_by_window(i, colorSchemeValue)
            self.settings.UpdateConfig()
        except Exception as e:
            log.error(f"on_selected_color_scheme_change err: {e}")
        evt.Skip()

    def aboutObjectReader(self, evt):
        message = _(
            "Object Reader allows the user to start reading the text that is below the cursor. "
            "With NVDA+mouseclick the user can activate the continuous reading mode of NVDA starting with the text below the cursor. "
            "Enabling this function will do the following: \n"
            "-The speech in NVDA is automatically disabled (but can be enabled with NVDA+S)\n"
            "-When this is used in browsers, USA will automatically switch between focus and browsemode when NVDA+alt+R is pressed\n"
            "-An advanced setting for forcing UIA elements in Word and Outlook is enabled in NVDA\n"
            "Note: this functionality does not work well with PDFs!"
        )
        gui.messageBox(message, _("About Object Reader"))

    def on_ok(self, evt):
        if self.settings.get_object_reader():
            config.conf["UIA"]["allowInMSWord"] = 3  # always
        else:
            config.conf["UIA"]["allowInMSWord"] = 0  # default

        stepLevelIndex = self.stepLevelCombobox.CurrentSelection
        stepLevelValue = float(self.stepLevelStrings[stepLevelIndex])
        self.settings.stepLevel = stepLevelValue

        if self.shouldShowOnStartupCheckbox:
            log.info(
                f"set_quick_settings_on_startup({bool(self.shouldShowOnStartupCheckbox.GetValue())})"
            )
            self.settings.set_quick_settings_on_startup(
                bool(self.shouldShowOnStartupCheckbox.GetValue())
            )

        colorSchemeDict = self.settings._get_all_color_schemes_by_window(0)
        set_scsk = self.settings.set_selected_color_scheme_by_window_key

        for i in range(0, self.settings.get_total_amount_windows()):
            set_scsk(
                i,
                "default",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["default"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                i,
                "inverted",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["inverted"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                i,
                "gray",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(colorSchemeDict["gray"].displayName)
                    )
                ),
            )
            set_scsk(
                i,
                "high-contrast",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["high-contrast"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                i,
                "inverted-high-contrast",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["inverted-high-contrast"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                i,
                "custom",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["custom"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                i,
                "blue-light",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["blue-light"].displayName
                        )
                    )
                ),
            )

        self.settings.delete_backup()
        self.settings.UpdateConfig()

        evt.Skip()
        # Not calling close here because it triggers a wx.ID_CANCEL event which calls on_discard
        # See __init__ : self.Bind(wx.EVT_BUTTON, self.on_discard, id=wx.ID_CANCEL)
        # https://docs.wxpython.org/window_deletion_overview.html#default-window-close-behaviour
        self.DestroyLater()

    def on_discard(self, evt):
        self.settings.restore_from_backup()
        self.settings.UpdateConfig()

        evt.Skip()
        self.DestroyLater()
