# -*- coding: utf-8 -*-
"""
Babbage B.V., 2019-2025

visionEnhancmentProviders/_base/dialogs.py
Contains classes with dialogs and settings. Also the auto-filled vision-settings panel in the general NVDA settings is
somewhat managed here.

NOTE: If you add checkboxes, do NOT use .addLabeledControl, use .addItem or the _create_checkbox_wrapper / _create_checkbox helper functions
      When using .addLabeledControl the checkbox will NOT BE ACCESSIBLE (the label will not be announced), tested with NVDA 2024.4.1
"""


import copy

import addonHandler
import core
import globalPluginHandler
import gui
import queueHandler
import ui
import visionEnhancementProviders.usa as USA
import wx  # type: ignore
from gui import guiHelper, nvdaControls
from logHandler import log

from .. import MatrixOperations as mo
from .. import libUSA

addonHandler.initTranslation()
import vision


def defaultOnEnableFailure(self):
    """Initialization of highlighterPlus failed. Reset settings / GUI"""
    settingsStorage = self._getSettingsStorage()
    settingsStorage.highlightBrowseMode = False
    settingsStorage.highlightFocus = False
    settingsStorage.highlightNavigator = False
    self.updateDriverSettings()
    self._updateEnabledState()
    wx.CallAfter(
        gui.messageBox(
            parent=gui.mainFrame,
            caption=_("Cannot Enable"),
            # Translators: Message shown when cannot connect to the remote computer.
            message=_(
                "Unable to enable NVDA Highlighter. \nPlease make sure that your Highlighter Plus is disabled."
            ),
            style=wx.OK | wx.ICON_WARNING,
        )
    )


def defaultEnsureEnableState(self, shouldBeEnabled: bool) -> bool:
    try:
        settings = USA.findUSAInstance().getSettings()
        plusHighlighterEnabled = settings.get_highlighter_plus()

        if settings and settings.conf["enabled"] and plusHighlighterEnabled:
            return False
    except Exception:
        pass

    currentlyEnabled = bool(self._providerControl.getProviderInstance())
    if shouldBeEnabled and not currentlyEnabled:
        return self._providerControl.startProvider()
    elif not shouldBeEnabled and currentlyEnabled:
        return self._providerControl.terminateProvider()
    return True


class USASettingsPanel(gui.settingsDialogs.VisionProviderSubPanel_Wrapper):
    def _create_checkbox(self, gui_box, text: str, var_name: str, type, getter, setter):
        try:
            # Create on change callback
            def on_change(evt):
                settings = self._providerSettings.getSettings()
                settings.backup_json()
                value = type(evt.GetInt())
                setter_name: str = "set_" + var_name
                try:
                    setter(value)
                    log.debug(f" {setter_name} succeeded")
                except Exception as e:
                    log.error(f" {setter_name} failed with err {e}")
                if "cam_viewer" in var_name:
                    settings.write_cam_viewer_conf()
                else:
                    settings.UpdateConfig()

            checkbox_name: str = f"{var_name}_checkbox"
            callback_name: str = "on_" + checkbox_name + "_change"
            setattr(self, callback_name, on_change)

            # Create checkbox
            parent = gui_box._parentRef()
            if isinstance(gui_box.sizer, wx.StaticBoxSizer):
                parent = gui_box.sizer.GetStaticBox()
            setattr(
                self,
                checkbox_name,
                gui_box.addItem(wx.CheckBox(parent, label=text)),
            )
            checkbox = getattr(self, checkbox_name)
            checkbox.SetValue(getter())
            checkbox.Bind(wx.EVT_CHECKBOX, getattr(self, callback_name))
            return checkbox
        except Exception as e:
            log.error(f"{var_name} checkbox creation failed with err {e}")

    def _create_checkbox_wrapper(self, gui_box, text: str, var_name: str, type):
        try:
            settings = self._providerSettings.getSettings()
            return self._create_checkbox(
                gui_box,
                text,
                var_name,
                type,
                getattr(settings, "get_" + var_name),
                getattr(settings, "set_" + var_name),
            )
        except Exception as e:
            log.error(f"{var_name} _create_checkbox_wrapper failed with {e}")

    def _create_combobox(
        self, gui_box, text: str, var_name: str, type, choices, values, getter, setter
    ):
        combobox_name: str = f"{var_name}_combobox"
        choices_name: str = f"{var_name}_choices"
        values_name: str = f"{var_name}_values"
        callback_name: str = "on_" + combobox_name + "_change"

        # Create on change callback
        def on_change(evt):
            settings = self._providerSettings.getSettings()
            settings.backup_json()
            index = getattr(self, combobox_name).CurrentSelection
            value = type(getattr(self, values_name)[index])
            setter_name: str = "set_" + var_name
            try:
                setter(value)
                log.debug(f" {setter_name} {value} succeeded")
            except Exception as e:
                log.error(f" {setter_name} failed with err {e}")
            if "cam_viewer" in var_name:
                settings.write_cam_viewer_conf()
            else:
                settings.UpdateConfig()

        setattr(self, callback_name, on_change)

        # Create combobox
        try:
            setattr(self, choices_name, choices)
            setattr(self, values_name, values)
            setattr(
                self,
                combobox_name,
                gui_box.addLabeledControl(
                    text, wx.Choice, choices=getattr(self, choices_name)
                ),
            )
            combobox = getattr(self, combobox_name)
            combobox.SetSelection(getattr(self, values_name).index(type(getter())))
            combobox.Bind(wx.EVT_CHOICE, getattr(self, callback_name))
        except Exception as e:
            log.error(f"{var_name} combobox creation failed with err {e}")

    def _create_combobox_wrapper(
        self, gui_box, text: str, var_name: str, type, choices, values
    ):
        try:
            settings = self._providerSettings.getSettings()
            return self._create_combobox(
                gui_box,
                text,
                var_name,
                type,
                choices,
                values,
                getattr(settings, "get_" + var_name),
                getattr(settings, "set_" + var_name),
            )
        except Exception as e:
            log.error(f"{var_name} _create_combobox_wrapper failed with err {e}")

    def on_highlighter_plus_change(self, evt):
        usa_inst = USA.findUSAInstance()
        settings = self._providerSettings.getSettings()
        is_enabled = settings.get_highlighter_plus()
        try:
            import vision

            can_start = True
            providers = vision.handler.getActiveProviderInfos()
            for provider in providers:
                if provider.providerId == "NVDAHighlighter":
                    can_start = False
            if not is_enabled and can_start:
                settings.set_highlighter_plus(True)
                usa_inst.highlighterPlus.start()
                self.hp_focus_checkbox.SetValue(
                    settings.get_highlighter_plus_enabled_by_mode("focus")
                )
                self.hp_focus_checkbox.Enable(True)
                self.hp_navigator_checkbox.SetValue(
                    settings.get_highlighter_plus_enabled_by_mode("navigator")
                )
                self.hp_navigator_checkbox.Enable(True)
                self.hp_browse_checkbox.SetValue(
                    settings.get_highlighter_plus_enabled_by_mode("browseMode")
                )
                self.hp_browse_checkbox.Enable(True)
                # If all sub-settings are disabled, enable them
                if (
                    not settings.get_highlighter_plus_enabled_by_mode("focus")
                    and not settings.get_highlighter_plus_enabled_by_mode("navigator")
                    and not settings.get_highlighter_plus_enabled_by_mode("browseMode")
                ):
                    self.hp_focus_checkbox.SetValue(True)
                    self.hp_navigator_checkbox.SetValue(True)
                    self.hp_browse_checkbox.SetValue(True)
                    settings.set_highlighter_plus_enabled_by_mode("focus", True)
                    settings.set_highlighter_plus_enabled_by_mode("navigator", True)
                    settings.set_highlighter_plus_enabled_by_mode("browseMode", True)

            elif is_enabled:
                settings.set_highlighter_plus(False)
                usa_inst.highlighterPlus.terminate()
                self.highlighter_plus_checkbox.SetValue(False)
                self.hp_focus_checkbox.Enable(False)
                self.hp_navigator_checkbox.Enable(False)
                self.hp_browse_checkbox.Enable(False)
            else:
                settings.set_highlighter_plus(False)
                self.highlighter_plus_checkbox.SetValue(False)
                wx.CallAfter(
                    gui.messageBox(
                        parent=gui.mainFrame,
                        caption=_("Cannot Enable"),
                        # Translators: Message shown when cannot connect to the remote computer.
                        message=_(
                            "Unable to enable Highlighter Plus. \nPlease make sure that your NVDA Visual Highlighter is disabled."
                        ),
                        style=wx.OK | wx.ICON_WARNING,
                    )
                )
        except Exception as e:
            log.error(f"on_highlighter_plus_change failed with err {e}")
        settings.UpdateConfig()

    # Wrap window_id int into class so that is gets copied be reference instead of value
    # then call the wrapper with the actual current window id
    # Window id wrapper has to be same instance everywhere otherwise we copy on construction..
    # so we would somehow need to replace window.CurrentSelection with a IntWrapper instance
    # class IntWrapper:
    # 	value: int

    def _getter_by_window(self, getter):
        def wrapper():
            return getter(self.editable_window_combo_box.CurrentSelection)

        return wrapper

    def _setter_by_window(self, setter):
        def wrapper(val):
            setter(self.editable_window_combo_box.CurrentSelection, val)

        return wrapper

    def _getter_by_highlighter_mode(self, getter, mode):
        def wrapper():
            return getter(mode)

        return wrapper

    def _setter_by_highlighter_mode(self, setter, mode):
        def wrapper(val):
            setter(mode, val)

        return wrapper

    def _create_tracking_boundary_combobox(
        self,
        parent_group,
        label_text,
        context,
        getter,
        setter,
        allow_same_as_focus=False,
    ):
        """Create a tracking boundary combobox for a specific context."""
        try:
            combobox_name = f"tracking_boundary_{context}_combobox"
            callback_name = f"on_tracking_boundary_{context}_changed"

            choices = self.boundary_level_choices.copy()
            values = self.boundary_level_values.copy()
            # Add "Same as focus" option for non-focus contexts
            if allow_same_as_focus and context != "focus":
                choices.insert(0, _("Same as focus"))
                values.insert(0, [-2, -2, -2, -2])

            combobox = parent_group.addLabeledControl(
                label_text,
                wx.Choice,
                choices=choices,
            )

            current_value = getter()

            if current_value < 0:
                boundary_value = int(current_value)
            else:
                boundary_value = int(current_value * 100)
            if boundary_value == -1:
                boundary_index = values.index([-1, -1, -1, -1])
            elif boundary_value == -2:
                boundary_index = values.index([-2, -2, -2, -2])
            else:
                boundary_index = values.index(
                    [
                        boundary_value,
                        boundary_value,
                        100 - boundary_value,
                        100 - boundary_value,
                    ]
                )
            combobox.SetSelection(boundary_index)

            def on_change(evt):
                settings = self._providerSettings.getSettings()
                settings.backup_json()
                boundaryIndex = combobox.CurrentSelection

                if allow_same_as_focus and context != "focus" and boundaryIndex == 0:
                    # "Same as focus" option selected
                    boundaryValueFloat = -2
                else:
                    # Regular boundary value selected
                    value_index = boundaryIndex
                    if allow_same_as_focus and context != "focus":
                        value_index -= 1  # Adjust for the added "Same as focus" option

                    boundary_values = self.boundary_level_values
                    boundaryValue = boundary_values[value_index][0]
                    # Special case for relative (-1) and same as focus (-2) values
                    if boundaryValue < 0:
                        boundaryValueFloat = float(boundaryValue)
                    else:
                        boundaryValueFloat = float(boundaryValue) / 100

                try:
                    setter(boundaryValueFloat)
                    log.debug(
                        f" set_tracking_boundary_{context} {boundaryValueFloat} succeeded"
                    )
                except Exception as e:
                    log.error(
                        f" set_tracking_boundary_{context} {boundaryValueFloat} failed with err {e}"
                    )
                settings.UpdateConfig()

            combobox.Bind(wx.EVT_CHOICE, on_change)
            setattr(self, combobox_name, combobox)
            setattr(self, callback_name, on_change)

            return combobox
        except Exception as e:
            log.error(
                f"failed tracking boundary {context} combobox creation with err {e}"
            )
            return None

    def _createProviderSettings(self):
        super()._createProviderSettings()
        settings = self._providerSettings.getSettings()

        sHelper = guiHelper.BoxSizerHelper(self, sizer=self._providerSettingsSizer)

        self.isRemote = sHelper.addItem(
            wx.CheckBox(self, label=_("Run USA in Remote mode (with Unicorn)"))
        )
        self.isRemote.SetValue(settings.isRemote)

        licenseButton = sHelper.addItem(
            wx.Button(self, label=_("Manage USA l&icense..."))
        )
        licenseButton.Bind(wx.EVT_BUTTON, self.onLicense)

        self._create_combobox_wrapper(
            sHelper,
            _("Select which magnification engine to use"),
            "engine",
            str,
            [_("Multi-Window"), _("Fullscreen")],
            ["multiwindow", "fullscreen"],
        )

        self._create_checkbox_wrapper(
            sHelper,
            _("Show quick settings on startup"),
            "quick_settings_on_startup",
            bool,
        )

        self._create_checkbox_wrapper(
            sHelper,
            _("Auto-start USA on NVDA start"),
            "autostart",
            bool,
        )

        self._create_checkbox_wrapper(
            sHelper,
            _("Always update all screens, decreases performance"),
            "quality_mode",
            bool,
        )

        self._create_combobox_wrapper(
            sHelper,
            _("View settings capturer"),
            "mouse_focus_mode",
            bool,
            [_("Mouse position"), _("NVDA-Focus position")],
            [True, False],
        )

        crosshair_box_sizer_helper = wx.StaticBoxSizer(
            wx.VERTICAL, self, label=_("Crosshair settings")
        )
        crosshair_box = crosshair_box_sizer_helper.GetStaticBox()
        self.crosshair_group = guiHelper.BoxSizerHelper(
            self, sizer=crosshair_box_sizer_helper
        )
        sHelper.addItem(self.crosshair_group)

        self._create_checkbox_wrapper(
            self.crosshair_group, _("Enable crosshair"), "crosshair_enable", bool
        )

        try:
            crosshair_color_text = _("Color:")
            self.crosshair_color_label = self.crosshair_group.addItem(
                wx.StaticText(crosshair_box, label=crosshair_color_text)
            )
            color: int = settings.get_crosshair_color()
            a = (color >> 24) & 0xFF
            r = (color >> 16) & 0xFF
            g = (color >> 8) & 0xFF
            b = color & 0xFF
            self.crosshair_color = self.crosshair_group.addItem(
                wx.ColourPickerCtrl(crosshair_box, name="crosshair_color_picker")
            )
            self.crosshair_color.SetColour(wx.Colour(r, g, b, a))
            self.crosshair_color.Bind(
                wx.EVT_COLOURPICKER_CHANGED, self.on_crosshair_color_change
            )
        except Exception as e:
            log.error(f"failed crosshair color control creation with err {e}")

        ## Highlighter plus ##
        try:
            try:
                providers = vision.handler.getProviderList()
                for provider in providers:
                    if provider.providerId == "NVDAHighlighter":
                        provider.providerClass.getSettingsPanelClass()._onEnableFailure = (
                            defaultOnEnableFailure
                        )
                        provider.providerClass.getSettingsPanelClass()._ensureEnableState = (
                            defaultEnsureEnableState
                        )
            except Exception as e:
                log.warning(
                    f"Init: check if default highlighter is up failed with err: {e}"
                )

            highlighter_sizer = sHelper.addItem(
                wx.StaticBoxSizer(
                    wx.VERTICAL, self, label=_("Highlighter Plus settings")
                ),
                flag=wx.EXPAND,
            )

            highlighter_sizer.AddSpacer(guiHelper.SPACE_BETWEEN_VERTICAL_DIALOG_ITEMS)
            highlighter_group = guiHelper.BoxSizerHelper(self, sizer=highlighter_sizer)

            self.highlighter_plus_checkbox = self._create_checkbox_wrapper(
                highlighter_group,
                _("Enable &Highlighter Plus"),
                "highlighter_plus",
                bool,
            )
            self.highlighter_plus_checkbox.Bind(
                wx.EVT_CHECKBOX, self.on_highlighter_plus_change
            )

            self.hp_focus_checkbox = self._create_checkbox(
                highlighter_group,
                _("Highlight system fo&cus"),
                "highlighter_plus_focus_mode",
                bool,
                self._getter_by_highlighter_mode(
                    getattr(settings, "get_highlighter_plus_enabled_by_mode"),
                    "focus",
                ),
                self._setter_by_highlighter_mode(
                    getattr(settings, "set_highlighter_plus_enabled_by_mode"),
                    "focus",
                ),
            )

            self.hp_navigator_checkbox = self._create_checkbox(
                highlighter_group,
                _("Highlight &navigator object"),
                "highlighter_plus_navigator_mode",
                bool,
                self._getter_by_highlighter_mode(
                    getattr(settings, "get_highlighter_plus_enabled_by_mode"),
                    "navigator",
                ),
                self._setter_by_highlighter_mode(
                    getattr(settings, "set_highlighter_plus_enabled_by_mode"),
                    "navigator",
                ),
            )

            self.hp_browse_checkbox = self._create_checkbox(
                highlighter_group,
                _("Highlight &browse mode cursor"),
                "highlighter_plus_browse_mode",
                bool,
                self._getter_by_highlighter_mode(
                    getattr(settings, "get_highlighter_plus_enabled_by_mode"),
                    "browseMode",
                ),
                self._setter_by_highlighter_mode(
                    getattr(settings, "set_highlighter_plus_enabled_by_mode"),
                    "browseMode",
                ),
            )
            enabled_highlighter = self.highlighter_plus_checkbox.GetValue()
            self.hp_focus_checkbox.Enable(enabled_highlighter)
            self.hp_navigator_checkbox.Enable(enabled_highlighter)
            self.hp_browse_checkbox.Enable(enabled_highlighter)

            highlighter_color_wheel_button = highlighter_group.addItem(
                wx.Button(
                    highlighter_sizer.GetStaticBox(),
                    label=_("Manage highlighter styles"),
                )
            )

            highlighter_color_wheel_button.Bind(
                wx.EVT_BUTTON, self.on_highlighter_plus_color_wheel
            )

        except Exception as e:
            log.error(f"failed highlighter plus creation with err {e}")

        try:
            crosshair_width_text = _("Line width")
            self.crosshair_width = self.crosshair_group.addLabeledControl(
                crosshair_width_text,
                nvdaControls.SelectOnFocusSpinCtrl,
                min=1,
                max=10,
                initial=int(settings.get_crosshair_width()),
            )
            self.crosshair_width.Bind(wx.EVT_TEXT, self.on_crosshair_width_change)
        except Exception as e:
            log.error(f"failed crosshair width creation with err {e}")

        self._create_checkbox_wrapper(
            self.crosshair_group,
            _("Enable crosshair gap"),
            "crosshair_enable_center_radius",
            bool,
        )

        try:
            crosshair_center_radius_text = _("Gap")
            self.crosshair_center_radius = self.crosshair_group.addLabeledControl(
                crosshair_center_radius_text,
                nvdaControls.SelectOnFocusSpinCtrl,
                min=1,
                max=20,
                initial=int(settings.get_crosshair_center_radius()),
            )
            self.crosshair_center_radius.Bind(
                wx.EVT_TEXT, self.on_crosshair_center_radius_change
            )
        except Exception as e:
            log.error(f"failed crosshair gap creation with err {e}")

        try:
            crosshair_min_distance_text = _("Dead zone")
            self.crosshair_min_distance = self.crosshair_group.addLabeledControl(
                crosshair_min_distance_text,
                nvdaControls.SelectOnFocusSpinCtrl,
                min=1,
                max=20,
                initial=int(settings.get_crosshair_min_distance()),
            )
            self.crosshair_min_distance.Bind(
                wx.EVT_TEXT, self.on_crosshair_min_distance_change
            )
        except Exception as e:
            log.error(f"failed crosshair min distance creation with err {e}")

        tracking_box_sizer_helper = wx.StaticBoxSizer(
            wx.VERTICAL, self, label=_("Tracking settings")
        )
        tracking_group = guiHelper.BoxSizerHelper(self, sizer=tracking_box_sizer_helper)
        sHelper.addItem(tracking_group)

        self._create_checkbox_wrapper(
            tracking_group,
            _("Track to focus"),
            "tracking_focus",
            bool,
        )

        self._create_checkbox_wrapper(
            tracking_group,
            _("Track to foreground window change"),
            "tracking_foreground",
            bool,
        )

        self._create_checkbox_wrapper(
            tracking_group,
            _("Track to navigator object"),
            "tracking_navigator",
            bool,
        )

        self._create_checkbox_wrapper(
            tracking_group,
            _("Track to system caret"),
            "tracking_caret",
            bool,
        )
        self._create_checkbox_wrapper(
            tracking_group,
            _("Track to browse mode cursor"),
            "tracking_browsemode",
            bool,
        )
        self._create_checkbox_wrapper(
            tracking_group,
            _("Track to review cursor"),
            "tracking_review",
            bool,
        )
        self._create_checkbox_wrapper(
            tracking_group,
            _("Track to mouse cursor"),
            "tracking_mouse",
            bool,
        )
        try:
            self.boundary_level_choices = [
                ii.displayName for ii in list(settings._availableBoundaries.values())
            ]
            self.boundary_level_values = [
                ii.id for ii in list(settings._availableBoundaries.values())
            ]

            self._create_tracking_boundary_combobox(
                tracking_group,
                _("Focus tracking boundary"),
                "focus",
                settings.get_tracking_boundary_focus,
                settings.set_tracking_boundary_focus,
            )

            self._create_tracking_boundary_combobox(
                tracking_group,
                _("Foreground tracking boundary"),
                "foreground",
                settings.get_tracking_boundary_foreground,
                settings.set_tracking_boundary_foreground,
                allow_same_as_focus=True,
            )

            self._create_tracking_boundary_combobox(
                tracking_group,
                _("Navigator tracking boundary"),
                "navigator",
                settings.get_tracking_boundary_navigator,
                settings.set_tracking_boundary_navigator,
                allow_same_as_focus=True,
            )

            self._create_tracking_boundary_combobox(
                tracking_group,
                _("Caret tracking boundary"),
                "caret",
                settings.get_tracking_boundary_caret,
                settings.set_tracking_boundary_caret,
                allow_same_as_focus=True,
            )

            self._create_tracking_boundary_combobox(
                tracking_group,
                _("Browse mode cursor tracking boundary"),
                "browsemode",
                settings.get_tracking_boundary_browsemode,
                settings.set_tracking_boundary_browsemode,
                allow_same_as_focus=True,
            )

            self._create_tracking_boundary_combobox(
                tracking_group,
                _("Review tracking boundary"),
                "review",
                settings.get_tracking_boundary_review,
                settings.set_tracking_boundary_review,
                allow_same_as_focus=True,
            )

            self._create_tracking_boundary_combobox(
                tracking_group,
                _("Mouse tracking boundary"),
                "mouse",
                settings.get_tracking_boundary_mouse,
                settings.set_tracking_boundary_mouse,
                allow_same_as_focus=True,
            )

        except Exception as e:
            log.error(
                f"failed context-specific tracking boundary comboboxes creation with err {e}"
            )

        self._create_combobox_wrapper(
            tracking_group,
            _("Tracking alignment"),
            "tracking_alignment",
            str,
            [_("Left"), _("Center")],
            ["left", "center"],
        )

        # self._create_checkbox_wrapper(
        #     sHelper, _("Show unma&gnified system cursor"), "show_system_cursor", bool
        # )

        def object_reader_setter(val):
            settings.set_object_reader(val)
            usa_inst = USA.findUSAInstance()

            if val:
                usa_inst.objectReader.enable()
                settings.set_word_follower_global(True)
                if usa_inst == None:
                    return
                usa_inst.wordFollower.disable_wordFollower(False)
                usa_inst.wordFollower.enable_wordFollower(False)
                self.word_follower_global_checkbox.SetValue(True)
            else:
                usa_inst.objectReader.disable()

        self._create_checkbox(
            sHelper,
            _("Activate Object Reader mode"),
            "object_reader",
            bool,
            settings.get_object_reader,
            object_reader_setter,
        )

        # def word_follower_setter(val):
        #     settings.set_word_follower(val)
        #     usa_inst = USA.findUSAInstance()
        #     if usa_inst == None:
        #         return
        #     if val:
        #         usa_inst.wordFollower.enable_wordFollower(False)
        #     else:
        #         usa_inst.wordFollower.disable_wordFollower(False)

        # TODO: un-hide if someone misses it, otherwise remove it
        # change word follower checkbox setter and getter back to the normal one
        def word_follower_global_setter(val):
            settings.set_word_follower_global(val)
            usa_inst = USA.findUSAInstance()
            if usa_inst == None:
                return
            usa_inst.wordFollower.disable_wordFollower(False)
            if val or settings.get_word_follower():
                usa_inst.wordFollower.enable_wordFollower(False)

        # self._create_checkbox(sHelper, _("Enable following of words everywhere."), "word_follower_global", bool,
        # 			settings.get_word_follower_global,
        # 			word_follower_global_setter
        # 			)

        # This uses the global Word Follower setting as that one is also triggered
        # using the NVDA+alt+w shortcut
        self._create_checkbox(
            sHelper,
            _("Enable following of words in continuous reading mode"),
            "word_follower_global",
            bool,
            settings.get_word_follower_global,
            word_follower_global_setter,
        )

        self._create_checkbox_wrapper(
            sHelper,
            _("Let the caret jump with the read words in continuous reading mode"),
            "word_follower_caret_jump",
            bool,
        )

        self._create_checkbox_wrapper(
            sHelper,
            _("Enable highlighting of words in continuous reading mode"),
            "word_follower_show_rect",
            bool,
        )

        self._create_combobox_wrapper(
            sHelper,
            _("Word Follower rectangle color"),
            "word_follower_rect_color",
            str,
            [
                ii.displayName
                for ii in list(settings._get_availableRectcolors().values())
            ],
            [ii.id for ii in list(settings._get_availableRectcolors().values())],
        )

        try:
            word_follower_rect_thickness_text = _("Word Follower rectangle thickness")
            self.word_follower_rect_thickness = sHelper.addLabeledControl(
                word_follower_rect_thickness_text,
                nvdaControls.SelectOnFocusSpinCtrl,
                min=1,
                max=10,
            )
            self.word_follower_rect_thickness.SetValue(
                settings.get_word_follower_rect_thickness()
            )
            self.word_follower_rect_thickness.Bind(
                wx.EVT_TEXT, self.on_word_follower_rect_thickness_change
            )
        except Exception as e:
            log.error(
                f"failed Word Follower rect thickness control creation with err {e}"
            )

        self._create_checkbox_wrapper(sHelper, _("Follow caret"), "follow_caret", bool)

        self._create_checkbox_wrapper(
            sHelper,
            _("Enable bitmap &smoothing to increase magnification quality"),
            "bitmap_smoothing",
            bool,
        )

        self._create_checkbox_wrapper(
            sHelper, _("Smooth magnification"), "smooth_magnification", bool
        )

        self._create_combobox_wrapper(
            sHelper,
            _("How smoothly the screen moves when zooming or panning"),
            "lerping_speed",
            int,
            [
                ii.displayName
                for ii in list(
                    settings._get_availableMagnificationsmoothnesss().values()
                )
            ],
            [
                int(ii.displayName)
                for ii in list(
                    settings._get_availableMagnificationsmoothnesss().values()
                )
            ],
        )

        self._create_checkbox_wrapper(
            sHelper,
            _("Let USA set the magnification level by fixed levels"),
            "magnification_fixed_levels",
            bool,
        )

        self._create_combobox_wrapper(
            sHelper,
            _("Step Level"),
            "magnification_step_level",
            float,
            [ii.displayName for ii in list(settings.availableSteplevels.values())],
            [
                float(ii.displayName)
                for ii in list(settings.availableSteplevels.values())
            ],
        )

        self.enable_docked_none_mode_checkbox = sHelper.addItem(
            wx.CheckBox(
                self, label=_("Enable Docked and No Magnifier mode (requires restart)")
            )
        )
        self.enable_docked_none_mode_checkbox.SetValue(
            settings.get_show_system_cursor()
        )
        self.enable_docked_none_mode_checkbox.Bind(
            wx.EVT_CHECKBOX, self.on_enable_docked_none_mode_changed
        )

        try:
            displays = [
                wx.Display(i).GetGeometry() for i in range(wx.Display.GetCount())
            ]
            number_editable_windows = len(displays)
            while number_editable_windows > settings.get_total_amount_windows():
                settings.add_window_to_json()

            editable_window_combo_box_text = _("Adjusted screen")
            self.editable_window_strings = [
                _("Screen ") + str(windowIndex)
                for windowIndex in range(number_editable_windows)
            ]
            self.editable_window_combo_box = sHelper.addLabeledControl(
                editable_window_combo_box_text,
                wx.Choice,
                choices=self.editable_window_strings,
            )
            self.editable_window_combo_box.SetSelection(0)
            self.editable_window_combo_box.Bind(
                wx.EVT_CHOICE, self.on_editable_window_changed
            )
        except Exception as e:
            log.error(
                f"failed creating list of editable windows combo box with err {e}"
            )

        try:
            mag_mode_combobox_text = _("Magnifier mode")
            self.mag_mode_strings = settings.get_magnifier_mode_strings()
            self.mag_mode_values = settings.get_magnifier_mode_values()
            self.mag_mode_combo_box = sHelper.addLabeledControl(
                mag_mode_combobox_text, wx.Choice, choices=self.mag_mode_strings
            )
            try:
                mag_mode_index = self.mag_mode_values.index(
                    str(
                        settings.get_magnifier_mode_by_window(
                            self.editable_window_combo_box.CurrentSelection
                        )
                    )
                )
            except:
                mag_mode_index = 0
            self.mag_mode_combo_box.SetSelection(mag_mode_index)
            self.mag_mode_combo_box.Bind(wx.EVT_CHOICE, self.on_mag_mode_changed)
        except Exception as e:
            log.error(f"failed mag mode with err {e}")

        # try:
        # 	mag_level_combo_box_text = _("Magnification &level")
        # 	self.mag_level_strings = [ii.displayName for ii in list(settings._get_availableMagnificationlevels().values())]
        # 	self.mag_level_combo_box = sHelper.addLabeledControl(mag_level_combo_box_text, wx.Choice, choices=self.mag_level_strings)
        # 	try:
        # 		mag_level_index = self.mag_level_strings.index(str(settings.get_magnification_level_by_window(self.editable_window_combo_box.CurrentSelection)))
        # 	except:
        # 		mag_level_index = 2
        # 	self.mag_level_combo_box.SetSelection(mag_level_index)
        # 	self.mag_level_combo_box.Bind(wx.EVT_CHOICE, self.on_mag_level_changed)
        # except Exception as e:
        # 	log.error(f'failed mag level with err {e}')
        self.mag_level_strings = [
            ii.displayName
            for ii in list(settings._get_availableMagnificationlevels().values())
        ]
        mag_level_values = [
            float(ii.displayName)
            for ii in list(settings._get_availableMagnificationlevels().values())
        ]

        current_mag_level = settings.get_magnification_level_by_window(0)
        if current_mag_level not in mag_level_values:
            # Custom mag level
            current_mag_level = round(current_mag_level, 2)
            settings.set_magnification_level_by_window(0, current_mag_level)
            mag_level_values.insert(0, current_mag_level)
            self.mag_level_strings.insert(0, str(current_mag_level))

        self._create_combobox(
            sHelper,
            _("Magnification &level"),
            "magnification_level",
            float,
            self.mag_level_strings,
            mag_level_values,
            self._getter_by_window(
                getattr(settings, "get_magnification_level_by_window")
            ),
            self._setter_by_window(
                getattr(settings, "set_magnification_level_by_window")
            ),
        )

        try:
            color_scheme_combo_box_text = _("Color sc&heme")
            self.color_scheme_strings = [
                ii.displayName
                for ii in list(
                    settings._get_available_color_schemes_by_window(
                        self.editable_window_combo_box.CurrentSelection
                    ).values()
                )
            ]
            self.color_scheme_key_strings = [
                ii
                for ii in list(
                    settings._get_available_color_schemes_by_window(
                        self.editable_window_combo_box.CurrentSelection
                    ).keys()
                )
            ]
            self.color_scheme_combo_box = sHelper.addLabeledControl(
                color_scheme_combo_box_text,
                wx.Choice,
                choices=self.color_scheme_strings,
            )
            self.color_scheme_combo_box.SetSelection(
                self.color_scheme_key_strings.index(
                    str(
                        settings.get_color_scheme_by_window(
                            self.editable_window_combo_box.CurrentSelection
                        )
                    )
                )
            )
            self.color_scheme_combo_box.Bind(
                wx.EVT_CHOICE, self.on_selected_color_scheme_change
            )
        except Exception as e:
            log.error(f"failed color scheme with err {e}")

        try:
            self.brightness_ctrl = sHelper.addLabeledControl(
                _("Brightness for all color schemes (%)"),
                nvdaControls.SelectOnFocusSpinCtrl,
                min=1,
                max=200,
                initial=int(
                    settings.get_brightness_by_window(
                        self.editable_window_combo_box.CurrentSelection
                    )
                ),
            )
            self.brightness_ctrl.Bind(wx.EVT_TEXT, self.onSpinCtrlChange)

            # Color scheme selection
            label = _("Color scheme settings")

            BoxSizerHelper = wx.StaticBoxSizer(wx.VERTICAL, self, label=label)
            colorschemeBox = BoxSizerHelper.GetStaticBox()
            self.colorschemeGroup = guiHelper.BoxSizerHelper(self, sizer=BoxSizerHelper)
            self.colorSchemeDict = settings._get_all_color_schemes_by_window(
                self.editable_window_combo_box.CurrentSelection
            )
            sHelper.addItem(self.colorschemeGroup)

            modifierBoxLabel = _("&Select color schemes")

            self.modifierChoices = [
                self.colorSchemeDict[key].displayName for key in self.colorSchemeDict
            ]
            self.modifierList = self.colorschemeGroup.addLabeledControl(
                modifierBoxLabel,
                nvdaControls.CustomCheckListBox,
                choices=self.modifierChoices,
            )
            checked_items = []
            try:
                get_scs = settings.get_selected_color_scheme_by_window_key
                if (
                    get_scs(self.editable_window_combo_box.CurrentSelection, "default")
                    == "True"
                ):
                    checked_items.append(
                        self.modifierChoices.index(
                            self.colorSchemeDict["default"].displayName
                        )
                    )
                if (
                    get_scs(self.editable_window_combo_box.CurrentSelection, "inverted")
                    == "True"
                ):
                    checked_items.append(
                        self.modifierChoices.index(
                            self.colorSchemeDict["inverted"].displayName
                        )
                    )
                if (
                    get_scs(self.editable_window_combo_box.CurrentSelection, "gray")
                    == "True"
                ):
                    checked_items.append(
                        self.modifierChoices.index(
                            self.colorSchemeDict["gray"].displayName
                        )
                    )
                if (
                    get_scs(
                        self.editable_window_combo_box.CurrentSelection, "high-contrast"
                    )
                    == "True"
                ):
                    checked_items.append(
                        self.modifierChoices.index(
                            self.colorSchemeDict["high-contrast"].displayName
                        )
                    )
                if (
                    get_scs(
                        self.editable_window_combo_box.CurrentSelection,
                        "inverted-high-contrast",
                    )
                    == "True"
                ):
                    checked_items.append(
                        self.modifierChoices.index(
                            self.colorSchemeDict["inverted-high-contrast"].displayName
                        )
                    )
                if (
                    get_scs(
                        self.editable_window_combo_box.CurrentSelection, "blue-light"
                    )
                    == "True"
                ):
                    checked_items.append(
                        self.modifierChoices.index(
                            self.colorSchemeDict["blue-light"].displayName
                        )
                    )
                if (
                    get_scs(self.editable_window_combo_box.CurrentSelection, "custom")
                    == "True"
                ):
                    checked_items.append(
                        self.modifierChoices.index(
                            self.colorSchemeDict["custom"].displayName
                        )
                    )
            except Exception as e:
                log.error(f"creating failed with err {e}")

            self.modifierList.CheckedItems = checked_items
            self.modifierList.Select(0)
            self.modifierList.Bind(wx.EVT_CHECKLISTBOX, self.on_checkbox_changed)

            # # Translators: Button
            self.colorButton = self.colorschemeGroup.addItem(
                wx.Button(colorschemeBox, label=_("Manage custom color sch&eme..."))
            )
            self.colorButton.Bind(wx.EVT_BUTTON, self.onColor)
            self.colorButton.Enable("custom" in self.color_scheme_key_strings)

            # TODO: un-hide once implemented
            # button for fg depending color schemes
            # self.fgColorButton = self.colorschemeGroup.addItem(wx.Button(colorschemeBox, label=_("Manage App-depending color schemes")))
            # self.fgColorButton.Bind(wx.EVT_BUTTON, self.onFGColor)
            # self.fgColorButton.Enable(bool(settings.supportedSettings))
            self.Layout()

        except Exception as e:
            log.error(f"brightness err: {e}")

        cam_viewer_box_sizer_helper = wx.StaticBoxSizer(
            wx.VERTICAL, self, label=_("Cam-Viewer settings")
        )
        cam_viewer_box = cam_viewer_box_sizer_helper.GetStaticBox()
        self.cam_viewer_group = guiHelper.BoxSizerHelper(
            self, sizer=cam_viewer_box_sizer_helper
        )
        sHelper.addItem(self.cam_viewer_group)

        webcam_values = list(range(0, settings.get_cam_viewer_max_webcam_index() + 1))
        webcam_choices = [f"Camera {i + 1}" for i in webcam_values]
        self._create_combobox_wrapper(
            self.cam_viewer_group,
            _("Camera selection"),
            "cam_viewer_webcam_index",
            int,
            webcam_choices,
            webcam_values,
        )

        self._create_checkbox_wrapper(
            self.cam_viewer_group,
            _("Cam-Viewer grayscale"),
            "cam_viewer_grayscale",
            bool,
        )
        self._create_checkbox_wrapper(
            self.cam_viewer_group, _("Cam-Viewer inverse"), "cam_viewer_inverse", bool
        )

        try:
            cam_viewer_brightness_text = _("Cam-Viewer brightness")
            self.cam_viewer_brightness = self.cam_viewer_group.addLabeledControl(
                cam_viewer_brightness_text,
                nvdaControls.SelectOnFocusSpinCtrl,
                min=-255,
                max=255,
                initial=int(settings.get_cam_viewer_brightness()),
            )
            self.cam_viewer_brightness.Bind(
                wx.EVT_TEXT, self.on_cam_viewer_brightness_change
            )
        except Exception as e:
            log.error(f"failed cam viewer brightness creation with err {e}")

        try:
            cam_viewer_contrast_text = _("Cam-Viewer contrast")
            self.cam_viewer_contrast = self.cam_viewer_group.addLabeledControl(
                cam_viewer_contrast_text,
                nvdaControls.SelectOnFocusSpinCtrl,
                min=-255,
                max=255,
                initial=int(settings.get_cam_viewer_contrast()),
            )
            self.cam_viewer_contrast.Bind(
                wx.EVT_TEXT, self.on_cam_viewer_contrast_change
            )
        except Exception as e:
            log.error(f"failed cam viewer contrast creation with err {e}")

        try:
            cam_viewer_color_offset_text = _("Cam-Viewer color offset:")
            self.cam_viewer_color_label = self.cam_viewer_group.addItem(
                wx.StaticText(cam_viewer_box, label=cam_viewer_color_offset_text)
            )
            color_offset = settings.get_cam_viewer_color_offset()
            r = color_offset[0]
            g = color_offset[1]
            b = color_offset[2]
            self.cam_viewer_color_offset = self.cam_viewer_group.addItem(
                wx.ColourPickerCtrl(
                    cam_viewer_box, name="cam_viewer_color_offset_picker"
                )
            )
            self.cam_viewer_color_offset.SetColour(wx.Colour(r, g, b))
            self.cam_viewer_color_offset.Bind(
                wx.EVT_COLOURPICKER_CHANGED, self.on_cam_viewer_color_offset_change
            )
        except Exception as e:
            log.error(f"failed cam viewer color offset control creation with err {e}")

        return True

    # Function override to be able to show setting without having provider settings configured
    # https://github.com/nvaccess/nvda/blob/c32e0e78a7a627a800fbf58ca3a4eade9c3c4477/source/gui/settingsDialogs.py#L5162
    def _updateOptionsVisibility(self):
        if bool(self._providerControl.getProviderInstance()) == False:
            self.settingsSizer.Hide(self._optionsSizer, recursive=True)
            self._sendLayoutUpdatedEvent()
        else:
            self.settingsSizer.Show(self._optionsSizer, recursive=True)
            self._sendLayoutUpdatedEvent()

    def on_highlighter_plus_color_wheel(self, evt):
        settings = self._providerSettings.getSettings()
        with HighlighterPlusStyleDialog(self, settings) as dlg:
            dlg.ShowModal()

    # creates checkbox list, sets the values correctly.
    def build_selected_color_scheme_settings(self):
        settings = self._providerSettings.getSettings()
        try:
            color_scheme_dict = settings._get_all_color_schemes_by_window(
                self.editable_window_combo_box.CurrentSelection
            )
            set_scsk = settings.set_selected_color_scheme_by_window_key

            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "default",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            color_scheme_dict["default"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "inverted",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            color_scheme_dict["inverted"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "gray",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            color_scheme_dict["gray"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "high-contrast",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            color_scheme_dict["high-contrast"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "inverted-high-contrast",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            color_scheme_dict["inverted-high-contrast"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "custom",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            color_scheme_dict["custom"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "blue-light",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            color_scheme_dict["blue-light"].displayName
                        )
                    )
                ),
            )

            # Update available color schemes for selected color scheme
            available_color_scheme = settings._get_available_color_schemes_by_window(
                self.editable_window_combo_box.CurrentSelection
            )
            self.color_scheme_strings = [
                ii.displayName for ii in list(available_color_scheme.values())
            ]
            self.color_scheme_key_strings = [
                ii for ii in list(available_color_scheme.keys())
            ]
            value = settings.get_color_scheme_by_window(
                self.editable_window_combo_box.CurrentSelection
            )
            log.debug(f"color scheme values {self.color_scheme_strings}")
            log.debug(f"color scheme keys {self.color_scheme_key_strings}")
            log.debug(f"color scheme by window {value}")
            log.debug(
                f"color scheme by index {self.color_scheme_key_strings.index(str(value))}"
            )
            self.color_scheme_combo_box.Clear()
            self.color_scheme_combo_box.AppendItems(self.color_scheme_strings)
            self.color_scheme_combo_box.SetSelection(
                self.color_scheme_key_strings.index(str(value))
            )

            self.colorButton.Enable("custom" in self.color_scheme_key_strings)

            log.debug(f"new combobox choices {self.color_scheme_combo_box}")
        except Exception as e:
            log.error(f"error on color_scheme_combo_box {e}")

    def onSpinCtrlChange(self, evt):
        brightness_value = float(evt.GetInt())
        settings = self._providerSettings.getSettings()
        if brightness_value == settings.get_brightness_by_window(
            self.editable_window_combo_box.CurrentSelection
        ):
            log.debug("brightness_value unchanged, skip")
            return

        settings.backup_json()
        brightness_value = max(1, min(200, brightness_value))
        settings.set_brightness_by_window(
            self.editable_window_combo_box.CurrentSelection, brightness_value
        )
        settings.UpdateConfig()

    def on_selected_color_scheme_change(self, evt):
        settings = self._providerSettings.getSettings()
        settings.backup_json()
        try:
            colorSchemeIndex = self.color_scheme_combo_box.CurrentSelection
            colorSchemeValue = str(self.color_scheme_key_strings[colorSchemeIndex])
            settings.set_color_scheme_by_window(
                self.editable_window_combo_box.CurrentSelection, colorSchemeValue
            )
            settings.UpdateConfig()
        except Exception as e:
            log.error(f"on_selected_color_scheme_change err: {e}")
        evt.Skip()

    def on_crosshair_color_change(self, evt):
        settings = self._providerSettings.getSettings()
        settings.backup_json()
        crosshair_color_value: wx.Colour = evt.GetColour()
        r = crosshair_color_value.Red()
        g = crosshair_color_value.Green()
        b = crosshair_color_value.Blue()
        a = crosshair_color_value.Alpha()
        color_value = (a << 24) | (r << 16) | (g << 8) | b
        try:
            settings.set_crosshair_color(color_value)
            log.debug(f" set_crosshair_color {color_value} succeeded")
        except Exception as e:
            log.error(f" set_crosshair_color {color_value} failed with err {e}")
        settings.UpdateConfig()

    def on_crosshair_width_change(self, evt):
        crosshair_width_value = float(evt.GetInt())
        settings = self._providerSettings.getSettings()
        if crosshair_width_value == settings.get_crosshair_width():
            log.debug("crosshair_width_value unchanged, skip")
            return

        settings.backup_json()
        # EVT_TEXT does not validate the value, so we clamp it here
        crosshair_width_value = max(1, min(10, crosshair_width_value))
        try:
            settings.set_crosshair_width(crosshair_width_value)
            log.debug(f" set_crosshair_width {crosshair_width_value} succeeded")
        except Exception as e:
            log.error(
                f" set_crosshair_width {crosshair_width_value} failed with err {e}"
            )
        settings.UpdateConfig()

    def on_crosshair_center_radius_change(self, evt):
        crosshair_center_radius_value = evt.GetInt()
        settings = self._providerSettings.getSettings()
        if crosshair_center_radius_value == settings.get_crosshair_center_radius():
            log.debug("crosshair_center_radius_value unchanged, skip")
            return

        settings.backup_json()
        crosshair_center_radius_value = max(1, min(20, crosshair_center_radius_value))
        try:
            settings.set_crosshair_center_radius(crosshair_center_radius_value)
            log.debug(
                f" set_crosshair_center_radius {crosshair_center_radius_value} succeeded"
            )
        except Exception as e:
            log.error(
                f" set_crosshair_center_radius {crosshair_center_radius_value} failed with err {e}"
            )
        settings.UpdateConfig()

    def on_crosshair_min_distance_change(self, evt):
        crosshair_min_distance_value = evt.GetInt()
        settings = self._providerSettings.getSettings()
        if crosshair_min_distance_value == settings.get_crosshair_min_distance():
            log.debug("crosshair_min_distance_value unchanged, skip")
            return

        settings.backup_json()
        crosshair_min_distance_value = max(1, min(20, crosshair_min_distance_value))
        try:
            settings.set_crosshair_min_distance(crosshair_min_distance_value)
            log.debug(
                f" set_crosshair_min_distance {crosshair_min_distance_value} succeeded"
            )
        except Exception as e:
            log.error(
                f" set_crosshair_min_distance {crosshair_min_distance_value} failed with err {e}"
            )
        settings.UpdateConfig()

    def on_word_follower_rect_thickness_change(self, evt):
        word_follower_rect_thickness_value = evt.GetInt()
        settings = self._providerSettings.getSettings()
        if (
            word_follower_rect_thickness_value
            == settings.get_word_follower_rect_thickness()
        ):
            log.debug("word_follower_rect_thickness_value unchanged, skip")
            return

        settings.backup_json()
        word_follower_rect_thickness_value = max(
            1, min(10, word_follower_rect_thickness_value)
        )
        try:
            settings.set_word_follower_rect_thickness(
                word_follower_rect_thickness_value
            )
            log.debug(
                f" set_word_follower_rect_thickness {word_follower_rect_thickness_value} succeeded"
            )
        except Exception as e:
            log.error(
                f" set_word_follower_rect_thickness {word_follower_rect_thickness_value} failed with err {e}"
            )
        settings.UpdateConfig()

    def on_checkbox_changed(self, evt):
        self.build_selected_color_scheme_settings()

    def on_mag_mode_changed(self, evt):
        settings = self._providerSettings.getSettings()
        settings.backup_json()
        mag_mode_index = self.mag_mode_combo_box.CurrentSelection
        mag_mode_value = self.mag_mode_values[mag_mode_index]
        try:
            settings.set_magnifier_mode_by_window(
                self.editable_window_combo_box.CurrentSelection, mag_mode_value
            )
            log.debug(f" set_magnifier_mode_by_window {mag_mode_value} suceeded")
        except Exception as e:
            log.error(
                f" set_magnifier_mode_by_window {mag_mode_value} failed with err {e}"
            )
        settings.UpdateConfig()

    def build_color_scheme_checkbox_settings(self):
        settings = self._providerSettings.getSettings()

        checked_items = []
        try:
            get_scs = settings.get_selected_color_scheme_by_window_key
            if (
                get_scs(self.editable_window_combo_box.CurrentSelection, "default")
                == "True"
            ):
                checked_items.append(
                    self.modifierChoices.index(
                        self.colorSchemeDict["default"].displayName
                    )
                )
            if (
                get_scs(self.editable_window_combo_box.CurrentSelection, "inverted")
                == "True"
            ):
                checked_items.append(
                    self.modifierChoices.index(
                        self.colorSchemeDict["inverted"].displayName
                    )
                )
            if (
                get_scs(self.editable_window_combo_box.CurrentSelection, "gray")
                == "True"
            ):
                checked_items.append(
                    self.modifierChoices.index(self.colorSchemeDict["gray"].displayName)
                )
            if (
                get_scs(
                    self.editable_window_combo_box.CurrentSelection, "high-contrast"
                )
                == "True"
            ):
                checked_items.append(
                    self.modifierChoices.index(
                        self.colorSchemeDict["high-contrast"].displayName
                    )
                )
            if (
                get_scs(
                    self.editable_window_combo_box.CurrentSelection,
                    "inverted-high-contrast",
                )
                == "True"
            ):
                checked_items.append(
                    self.modifierChoices.index(
                        self.colorSchemeDict["inverted-high-contrast"].displayName
                    )
                )
            if (
                get_scs(self.editable_window_combo_box.CurrentSelection, "blue-light")
                == "True"
            ):
                checked_items.append(
                    self.modifierChoices.index(
                        self.colorSchemeDict["blue-light"].displayName
                    )
                )
            if (
                get_scs(self.editable_window_combo_box.CurrentSelection, "custom")
                == "True"
            ):
                checked_items.append(
                    self.modifierChoices.index(
                        self.colorSchemeDict["custom"].displayName
                    )
                )

        except Exception as e:
            log.error(f"creating failed with err {e}")

        self.modifierList.CheckedItems = checked_items
        self.modifierList.Select(0)

    def on_enable_docked_none_mode_changed(self, evt):
        settings = self._providerSettings.getSettings()
        is_enabled = self.enable_docked_none_mode_checkbox.GetValue()

        # We will need to restart NVDA for a consistent system cursor behaviour
        if (
            gui.messageBox(
                _("NVDA will be restarted to apply this setting."),
                _("Information"),
                wx.OK | wx.CANCEL | wx.ICON_EXCLAMATION,
            )
            == wx.CANCEL
        ):
            self.enable_docked_none_mode_checkbox.SetValue(not is_enabled)
            return

        try:
            settings.set_show_system_cursor(is_enabled)
            log.debug(f" set_show_system_cursor {is_enabled} succeeded")
        except Exception as e:
            log.error(f" set_show_system_cursor {is_enabled} failed with err {e}")

        # If we disable the system cursor make sure that on no magnification window docked
        # or none mode is activated
        if not is_enabled:
            windows = settings.get_windows()
            for i, w in enumerate(windows):
                if w["mode"] in ["docked", "none"]:
                    settings.set_magnifier_mode_by_window(i, "per-screen")

        settings.UpdateConfig()

        # We already notified the user about the pending restart, now we just have to restart
        queueHandler.queueFunction(queueHandler.eventQueue, core.restart)

    # This should reset the values of the comboboxes and inputfields.
    # Additionally, it needs to retrieve the available color schemes and update the checkboxes.
    def on_editable_window_changed(self, event):
        settings = self._providerSettings.getSettings()
        self.mag_mode_combo_box.SetSelection(
            self.mag_mode_values.index(
                str(
                    settings.get_magnifier_mode_by_window(
                        self.editable_window_combo_box.CurrentSelection
                    )
                )
            )
        )
        try:
            self.magnification_level_combobox.SetSelection(
                self.mag_level_strings.index(
                    str(
                        settings.get_magnification_level_by_window(
                            self.editable_window_combo_box.CurrentSelection
                        )
                    )
                )
            )
        except ValueError:
            # Custom magnification level
            current_mag_level = settings.get_magnification_level_by_window(
                self.editable_window_combo_box.CurrentSelection
            )
            current_mag_level = round(current_mag_level, 2)
            settings.set_magnification_level_by_window(
                self.editable_window_combo_box.CurrentSelection, current_mag_level
            )
            current_mag_level_str = str(current_mag_level)
            self.mag_level_strings.insert(0, current_mag_level_str)
            # Update checkbox-internal choices
            self.magnification_level_combobox.Set(self.mag_level_strings)
            self.magnification_level_combobox.SetSelection(
                self.mag_level_strings.index(current_mag_level_str)
            )

        self.brightness_ctrl.Value = int(
            settings.get_brightness_by_window(
                self.editable_window_combo_box.CurrentSelection
            )
        )

        self.build_color_scheme_checkbox_settings()
        self.build_selected_color_scheme_settings()
        available_color_scheme = settings._get_available_color_schemes_by_window(
            self.editable_window_combo_box.CurrentSelection
        )
        self.color_scheme_key_strings = [
            ii for ii in list(available_color_scheme.keys())
        ]
        log.debug(f"self.color_schsneme_key_strings {self.color_scheme_key_strings}")
        log.debug(
            f"string of current selected {str(settings.get_color_scheme_by_window(self.editable_window_combo_box.CurrentSelection))}"
        )
        log.debug(
            f"string of index = {self.color_scheme_key_strings.index(str(settings.get_color_scheme_by_window(self.editable_window_combo_box.CurrentSelection)))}"
        )
        log.debug(
            f"window button {self.editable_window_combo_box.CurrentSelection + 1} has been selected"
        )
        self.color_scheme_combo_box.SetSelection(
            self.color_scheme_key_strings.index(
                str(
                    settings.get_color_scheme_by_window(
                        self.editable_window_combo_box.CurrentSelection
                    )
                )
            )
        )

    def on_cam_viewer_brightness_change(self, evt):
        cam_viewer_brightness_value = float(evt.GetInt())
        settings = self._providerSettings.getSettings()
        if cam_viewer_brightness_value == settings.get_cam_viewer_brightness():
            log.debug("set_cam_viewer_brightness unchanged, skip")
            return

        settings.backup_json()
        cam_viewer_brightness_value = max(-255, min(255, cam_viewer_brightness_value))
        try:
            settings.set_cam_viewer_brightness(cam_viewer_brightness_value)
            log.debug(
                f" set_cam_viewer_brightness {cam_viewer_brightness_value} succeeded"
            )
        except Exception as e:
            log.error(
                f" set_cam_viewer_brightness {cam_viewer_brightness_value} failed with err {e}"
            )
        settings.write_cam_viewer_conf()

    def on_cam_viewer_contrast_change(self, evt):
        cam_viewer_contrast_value = float(evt.GetInt())
        settings = self._providerSettings.getSettings()
        if cam_viewer_contrast_value == settings.get_cam_viewer_brightness():
            log.debug("set_cam_viewer_brightness unchanged, skip")
            return

        settings.backup_json()
        cam_viewer_contrast_value = max(-255, min(255, cam_viewer_contrast_value))
        try:
            settings.set_cam_viewer_contrast(cam_viewer_contrast_value)
            log.debug(f" set_cam_viewer_contrast {cam_viewer_contrast_value} succeeded")
        except Exception as e:
            log.error(
                f" set_cam_viewer_contrast {cam_viewer_contrast_value} failed with err {e}"
            )
        settings.write_cam_viewer_conf()

    def on_cam_viewer_color_offset_change(self, evt):
        settings = self._providerSettings.getSettings()
        settings.backup_json()
        cam_viewer_color_offset: wx.Colour = evt.GetColour()
        r = cam_viewer_color_offset.Red()
        g = cam_viewer_color_offset.Green()
        b = cam_viewer_color_offset.Blue()
        try:
            settings.set_cam_viewer_color_offset((r, g, b))
            log.debug(f" set_cam_viewer_color_offset {(r, g, b)} succeeded")
        except Exception as e:
            log.error(f" set_cam_viewer_color_offset {(r, g, b)} failed with err {e}")
        settings.write_cam_viewer_conf()

    # FIGURE OUT WHEN IT SAVES AND WHEN IT SHOULD BE SETTING THE VARIABLES PROPERLY RATHER THAN ALL OVER THE PLACE CONSTANTLY.
    def _enableToggle(self, evt):
        super()._enableToggle(evt)
        settings = self._providerSettings.getSettings()
        self.colorButton.Enable(bool(settings.supportedSettings))

    def onLicense(self, evt):
        # get the license dialog from the global plugin, since we also need it there.
        allPlugins = list(globalPluginHandler.runningPlugins)
        usaPlugin = None
        for ii in allPlugins:
            try:
                # just call a function that we know
                enabled = ii.isUSAEnabled()
                usaPlugin = ii
            except:
                pass
        if usaPlugin:
            USALicenseDialog = usaPlugin.getLicenseDialog()
            with USALicenseDialog(self) as dlg:
                dlg.ShowModal()

    def onColor(self, evt):
        settings = self._providerSettings.getSettings()
        with USACustomColorSchemeDialog(
            self, settings, self.editable_window_combo_box.CurrentSelection
        ) as dlg:
            dlg.ShowModal()

    def onFGColor(self, evt):
        settings = self._providerSettings.getSettings()
        with USAFGColorSchemeDialog(self, settings) as dlg:
            dlg.ShowModal()

    def onSave(self):
        # save colorscheme selection
        settings = self._providerSettings.getSettings()
        colorSchemeDict = settings._get_all_color_schemes_by_window(
            self.editable_window_combo_box.CurrentSelection
        )
        conf = settings.conf
        try:
            log.debug(
                f"self.editable_window val: {self.editable_window_combo_box.CurrentSelection}"
            )
            set_scsk = settings.set_selected_color_scheme_by_window_key
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "default",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["default"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "inverted",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["inverted"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "gray",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(colorSchemeDict["gray"].displayName)
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "high-contrast",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["high-contrast"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "inverted-high-contrast",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["inverted-high-contrast"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "custom",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["custom"].displayName
                        )
                    )
                ),
            )
            set_scsk(
                self.editable_window_combo_box.CurrentSelection,
                "blue-light",
                str(
                    self.modifierList.IsChecked(
                        self.modifierChoices.index(
                            colorSchemeDict["blue-light"].displayName
                        )
                    )
                ),
            )

            settings.set_brightness_by_window(
                self.editable_window_combo_box.CurrentSelection,
                self.brightness_ctrl.Value,
            )

            magLevelIndex = self.magnification_level_combobox.CurrentSelection
            magLevelValue = float(self.mag_level_strings[magLevelIndex])
            try:
                settings.set_magnification_level_by_window(
                    self.editable_window_combo_box.CurrentSelection, magLevelValue
                )
                log.debug(
                    f" set_magnification_level_by_window {magLevelValue} suceeded"
                )
            except Exception as e:
                log.error(
                    f" set_magnification_level_by_window {magLevelValue} failed with err {e}"
                )

            colorSchemeIndex = self.color_scheme_combo_box.CurrentSelection
            colorSchemeValue = str(self.color_scheme_key_strings[colorSchemeIndex])
            try:
                settings.set_color_scheme_by_window(
                    self.editable_window_combo_box.CurrentSelection, colorSchemeValue
                )
                # settings.set_custom_color_scheme_by_window(self.editable_window_combo_box.CurrentSelection, colorSchemeDict[colorSchemeValue].value.transform)
                log.debug(f" set_color_scheme_by_window {colorSchemeValue} suceeded")
            except Exception as e:
                log.error(
                    f" set_color_scheme_by_window {colorSchemeValue} failed with err {e}"
                )
        except Exception as err:
            log.error(f"welp something went wrong err: {err}")
        if self.isRemote.GetValue() and not settings.remoteHandler:
            ui.messageBox(
                _(
                    "Unicorn NVDA addon cannot be found! USA cannot run in remote if Unicorn is not active!"
                )
            )
            settings.isRemote = False
        else:
            settings.isRemote = self.isRemote.GetValue()

        self._providerSettings.getSettings().delete_backup()
        self._providerSettings.getSettings().UpdateConfig()
        super().onSave()

    def onDiscard(self):
        self._providerSettings.getSettings().restore_from_backup()
        self._providerSettings.getSettings().UpdateConfig()
        self._providerSettings.getSettings().write_cam_viewer_conf()

        super().onDiscard()


class USACustomColorSchemeDialog(gui.settingsDialogs.SettingsDialog):

    def makeSettings(self, settingsSizer):
        main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, sizer=settingsSizer)
        self.initialScheme = copy.deepcopy(
            self.settings.get_custom_color_scheme_by_window(self.windowIndex)
        )

        # # Translators: Button
        self.matrixButton = main_sizer_helper.addItem(
            wx.Button(self, label=_("Advanced settings"))
        )
        self.matrixButton.Bind(wx.EVT_BUTTON, self.onMatrix)

        cs_settings = self.settings.get_custom_color_scheme_settings_by_window(
            self.windowIndex
        )
        self.cs_settings_backup = copy.deepcopy(cs_settings)

        brightnessLabelText = _("Brightness (%)")
        self.BrightnessCtrl = main_sizer_helper.addLabeledControl(
            brightnessLabelText,
            nvdaControls.SelectOnFocusSpinCtrl,
            min=0,
            max=200,
            initial=int(cs_settings["brightness"] * 100),
        )
        self.BrightnessCtrl.Bind(wx.EVT_TEXT, self.onSpinCtrlChange)

        contrastLabelText = _("Contrast (%)")
        self.ContrastCtrl = main_sizer_helper.addLabeledControl(
            contrastLabelText,
            nvdaControls.SelectOnFocusSpinCtrl,
            min=0,
            max=200,
            initial=int(cs_settings["contrast"] * 100),
        )
        self.ContrastCtrl.Bind(wx.EVT_TEXT, self.onSpinCtrlChange)

        saturationLabelText = _("Saturation (%)")
        self.SaturationCtrl = main_sizer_helper.addLabeledControl(
            saturationLabelText,
            nvdaControls.SelectOnFocusSpinCtrl,
            min=0,
            max=500,
            initial=int(cs_settings["saturation"] * 100),
        )
        self.SaturationCtrl.Bind(wx.EVT_TEXT, self.onSpinCtrlChange)

        self.HueScaling = 360
        HueRotationLabelText = _("Hue rotation (degrees)")
        self.HueRotationCtrl = main_sizer_helper.addLabeledControl(
            HueRotationLabelText,
            nvdaControls.SelectOnFocusSpinCtrl,
            min=0,
            max=self.HueScaling,
            initial=int(cs_settings["hue-rotation"]),
        )
        self.HueRotationCtrl.Bind(wx.EVT_TEXT, self.onSpinCtrlChange)

        blueLightLabelText = _("Blue light (%)")
        self.BlueLightCtrl = main_sizer_helper.addLabeledControl(
            blueLightLabelText,
            nvdaControls.SelectOnFocusSpinCtrl,
            min=0,
            max=150,
            initial=int(cs_settings["blue-light"] * 100),
        )
        self.BlueLightCtrl.Bind(wx.EVT_TEXT, self.onSpinCtrlChange)

        self.checkboxInverse = main_sizer_helper.addItem(
            (wx.CheckBox(self, label=_("Inverse colors")))
        )
        self.checkboxInverse.SetValue(cs_settings["inverse-colors"])
        self.checkboxInverse.Bind(wx.EVT_CHECKBOX, self.onSpinCtrlChange)

        self.checkboxOffset = main_sizer_helper.addItem(
            (wx.CheckBox(self, label=_("Color offset")))
        )
        self.checkboxOffset.SetValue(cs_settings["add-color-offset"])
        self.checkboxOffset.Bind(wx.EVT_CHECKBOX, self.onSpinCtrlChange)

        self.colorWheel = main_sizer_helper.addItem(
            wx.ColourPickerCtrl(self, name="colorwheel")
        )

        self.colorWheel.Bind(wx.EVT_COLOURPICKER_CHANGED, self.onSpinCtrlChange)

        self.colorWheel.SetColour(wx.Colour(cs_settings["color-offset"]))

    def __init__(self, parent, settings, windowIndex):
        self.settings = settings
        self.windowIndex = windowIndex
        super().__init__(parent, multiInstanceAllowed=True)

    title = _("Color scheme configuration")

    def hsv_to_rgb(self, h):
        v = 1
        s = 1
        i = int(h * 6.0)  # XXX assume int() truncates!
        f = (h * 6.0) - i
        p, q, t = (
            int(255 * (v * (1.0 - s))),
            int(255 * (v * (1.0 - s * f))),
            int(255 * (v * (1.0 - s * (1.0 - f)))),
        )
        v *= 255
        i %= 6
        if i == 0:
            return (v, t, p)
        if i == 1:
            return (q, v, p)
        if i == 2:
            return (p, v, t)
        if i == 3:
            return (p, q, v)
        if i == 4:
            return (t, p, v)
        if i == 5:
            return (v, p, q)

    def getBSCMatrix(self):
        # from https://docs.rainmeter.net/tips/colormatrix-guide/
        # https://flylib.com/books/en/2.72.1/matrix_operations_in_image_processing.html
        # https://www.c-sharpcorner.com/uploadfile/mahesh/color-mapping-using-color-objects-in-gdi/
        B = (self.BrightnessCtrl.GetValue() - 100.0) / 100.0
        C = self.ContrastCtrl.GetValue() / 100.0
        S = self.SaturationCtrl.GetValue() / 100.0
        rgbOffset = self.colorWheel.GetColour().Get()
        h = self.HueRotationCtrl.GetValue() / self.HueScaling
        addOffsetColor = self.checkboxOffset.GetValue()
        invert = self.checkboxInverse.GetValue()
        blueLight = self.BlueLightCtrl.GetValue() / 100.0
        sumRGB = (rgbOffset[0] + rgbOffset[1] + rgbOffset[2]) / 3.0
        rOffset = (rgbOffset[0] - sumRGB) / 255.0 if addOffsetColor else 0.0
        gOffset = (rgbOffset[1] - sumRGB) / 255.0 if addOffsetColor else 0.0
        bOffset = (rgbOffset[2] - sumRGB) / 255.0 if addOffsetColor else 0.0

        t = (1.0 - C) / 2.0

        lumR = 0.3086
        lumG = 0.6094
        lumB = 0.0820
        sr = (1.0 - S) * lumR
        sg = (1.0 - S) * lumG
        sb = (1.0 - S) * lumB

        RGBmats = [
            [C * (sr + S) - (blueLight * 0.2), C * sr, C * sr],
            [C * sg, C * (sg + S) - (blueLight * 0.5), C * sg],
            [C * sb, C * sb, C * (sb + S) - blueLight],
        ]

        RGBmat = mo.rotateHue(2 * mo.pi * h)
        RGBmat = mo.matrix_multiply(RGBmat, RGBmats)
        rowR = [RGBmat[0][0], RGBmat[0][1], RGBmat[0][2], 0, 0]
        rowG = [RGBmat[1][0], RGBmat[1][1], RGBmat[1][2], 0, 0]
        rowB = [RGBmat[2][0], RGBmat[2][1], RGBmat[2][2], 0, 0]
        rowA = [0, 0, 0, 1, 0]
        rowW = [t + B + rOffset, t + B + gOffset, t + B + bOffset, 0, 1]
        BSCMatrix = [rowR, rowG, rowB, rowA, rowW]

        if invert:
            BSCMatrix = mo.matrix_multiply(mo.inverseMatrix, BSCMatrix)

        log.debug(f"USA: {RGBmats} \n {BSCMatrix}")
        return BSCMatrix

    def save_config_values(self):
        cs_settings = copy.copy(
            self.settings.get_custom_color_scheme_settings_by_window(self.windowIndex)
        )
        cs_settings["brightness"] = float(self.BrightnessCtrl.GetValue() / 100)
        cs_settings["contrast"] = float(self.ContrastCtrl.GetValue() / 100)
        cs_settings["saturation"] = float(self.SaturationCtrl.GetValue() / 100)
        cs_settings["hue-rotation"] = self.HueRotationCtrl.GetValue()
        cs_settings["inverse-colors"] = bool(self.checkboxInverse.GetValue())
        cs_settings["add-color-offset"] = bool(self.checkboxOffset.GetValue())
        cs_settings["blue-light"] = float(self.BlueLightCtrl.GetValue() / 100)
        cs_settings["color-offset"] = self.colorWheel.GetColour().GetRGBA()

        previous_cs_settings = self.settings.get_custom_color_scheme_settings_by_window(
            self.windowIndex
        )
        if cs_settings == previous_cs_settings:
            log.debug("cs_settings unchanged, skip")
            return

        self.settings.set_custom_color_scheme_settings_by_window(
            self.windowIndex, cs_settings
        )

        BCSMatrix = self.getBSCMatrix()
        self.settings.set_custom_color_scheme_by_window(self.windowIndex, BCSMatrix)
        self.settings.UpdateConfig()

    def onSpinCtrlChange(self, evt):
        self.save_config_values()
        evt.Skip()

    def onMatrix(self, evt):
        with USACustomColorSchemeMatrix(
            self, self.settings, self.windowIndex, self.initialScheme
        ) as dlg:
            dlg.ShowModal()

        super().onCancel(evt)

    def onOk(self, evt):
        self.save_config_values()

        super().onOk(evt)

    def onCancel(self, evt):
        self.settings.set_custom_color_scheme_settings_by_window(
            self.windowIndex, self.cs_settings_backup
        )
        self.settings.set_custom_color_scheme_by_window(
            self.windowIndex, self.initialScheme
        )
        self.settings.UpdateConfig()

        super().onCancel(evt)


class USACustomColorSchemeMatrix(wx.Dialog):

    def __init__(self, parent, settings, windowIndex, initialScheme):
        super().__init__(parent, id=wx.ID_ANY, title=_("Change color scheme"))
        self.settings = settings
        self.windowIndex = windowIndex

        self.initialScheme = initialScheme
        self.scheme = self.settings.get_custom_color_scheme_by_window(self.windowIndex)

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, sizer=main_sizer)
        self.rowColNaming = ("r", "g", "b", "a", "t")
        gridSizer = wx.GridBagSizer(
            hgap=guiHelper.SPACE_BETWEEN_BUTTONS_HORIZONTAL,
            vgap=guiHelper.SPACE_BETWEEN_BUTTONS_VERTICAL,
        )
        gridSizer.Rows = 6
        gridSizer.Cols = 6
        self.spinControls = []
        for r in range(gridSizer.Rows):
            if r != 0:
                l = []
                self.spinControls.append(l)
            for c in range(gridSizer.Cols):
                if r == 0 and c == 0:
                    continue
                if r == 0 and c != 0:
                    label = wx.StaticText(self, label=self.rowColNaming[c - 1])
                    gridSizer.Add(label, pos=(r, c))
                    continue
                if c == 0 and r != 0:
                    label = wx.StaticText(self, label=self.rowColNaming[r - 1])
                    gridSizer.Add(label, pos=(r, c))
                    continue
                control = wx.SpinCtrlDouble(
                    self, initial=0.0, inc=0.1, min=-10.0, max=10.0
                )
                control.Value = self.scheme[r - 1][c - 1]
                control.SetToolTip(f"{self.rowColNaming[r-1]}:{self.rowColNaming[c-1]}")
                control.Bind(
                    wx.EVT_SPINCTRLDOUBLE, self.onSpinCtrlChange, source=control
                )
                gridSizer.Add(control, pos=(r, c))
                l.append(control)
        main_sizer.Add(gridSizer, border=gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        main_sizer_helper.addDialogDismissButtons(
            self.CreateButtonSizer(wx.OK | wx.CANCEL)
        )
        self.Bind(wx.EVT_BUTTON, self.onOk, id=wx.ID_OK)
        self.Bind(wx.EVT_BUTTON, self.onCancel, id=wx.ID_CANCEL)
        main_sizer.Fit(self)
        self.SetSizer(main_sizer)
        self.Center(wx.BOTH)
        self.Show()

    def onSpinCtrlChange(self, evt):
        obj = evt.EventObject
        for r, lst in enumerate(self.spinControls):
            for c, control in enumerate(lst):
                if obj is control:
                    self.scheme[r][c] = evt.Value
                    break
            else:
                continue
            break
        else:
            log.error(f"{obj!r} not known to handler")
            return
        self.settings.set_custom_color_scheme_by_window(self.windowIndex, self.scheme)
        self.settings.UpdateConfig()
        evt.Skip()

    def onOk(self, evt):
        self.settings.set_custom_color_scheme_by_window(self.windowIndex, self.scheme)
        self.settings.UpdateConfig()
        evt.Skip()

    def onCancel(self, evt):
        self.settings.set_custom_color_scheme_by_window(
            self.windowIndex, self.initialScheme
        )
        self.settings.UpdateConfig()
        evt.Skip()


class HighlighterPlusStyleDialog(gui.settingsDialogs.SettingsDialog):
    contextStyleLevelStrings = [
        "1.0",
        "2.0",
        "3.0",
        "4.0",
        "5.0",
        "6.0",
        "7.0",
        "8.0",
        "9.0",
        "10.0",
    ]

    def makeSettings(self, settingsSizer):
        main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, sizer=settingsSizer)
        sub_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.HORIZONTAL)
        sub_sizer_text_helper = gui.guiHelper.BoxSizerHelper(
            self, orientation=wx.VERTICAL
        )
        sub_sizer_colorwheel_helper = gui.guiHelper.BoxSizerHelper(
            self, orientation=wx.VERTICAL
        )
        sub_sizer_thickness_helper = gui.guiHelper.BoxSizerHelper(
            self, orientation=wx.VERTICAL
        )
        sub_sizer_margin_helper = gui.guiHelper.BoxSizerHelper(
            self, orientation=wx.VERTICAL
        )
        sub_sizer_line_style_helper = gui.guiHelper.BoxSizerHelper(
            self, orientation=wx.VERTICAL
        )
        main_sizer_helper.addItem(sub_sizer_helper)

        # Order matters here
        sub_sizer_helper.addItem(sub_sizer_text_helper)
        sub_sizer_helper.addItem(sub_sizer_colorwheel_helper)
        sub_sizer_helper.addItem(sub_sizer_thickness_helper)
        sub_sizer_helper.addItem(sub_sizer_margin_helper)
        sub_sizer_helper.addItem(sub_sizer_line_style_helper)
        try:
            log.debug(f"HighlighterPlusStyleDialog::makeSettings() {self.settings}")
            for key, contextStyle in self.settings._contextStyles.items():
                text_sizer = wx.BoxSizer(wx.HORIZONTAL)
                text_sizer.Add(wx.StaticText(self, label=key))
                sub_sizer_text_helper.addItem(text_sizer, border=4, flag=wx.ALL)

                colorWheel = sub_sizer_colorwheel_helper.addItem(
                    wx.ColourPickerCtrl(self)
                )
                colorWheel.Bind(
                    wx.EVT_COLOURPICKER_CHANGED, contextStyle.on_spin_control_change
                )
                colorWheel.SetColour(
                    wx.Colour(
                        contextStyle.color[0],
                        contextStyle.color[1],
                        contextStyle.color[2],
                        255,
                    )
                )
                contextStyle.colorWheel = colorWheel

                thicknessCombobox = sub_sizer_thickness_helper.addLabeledControl(
                    _("Thickness"), wx.Choice, choices=self.contextStyleLevelStrings
                )
                thicknessCombobox.SetSelection(
                    self.contextStyleLevelStrings.index(str(contextStyle.thickness))
                )
                contextStyle.thicknessCtrl = thicknessCombobox

                marginCombobox = sub_sizer_margin_helper.addLabeledControl(
                    _("Margin"), wx.Choice, choices=self.contextStyleLevelStrings
                )
                marginCombobox.SetSelection(
                    self.contextStyleLevelStrings.index(str(contextStyle.margin))
                )
                contextStyle.marginCtrl = marginCombobox

                lineStyleCombobox = sub_sizer_line_style_helper.addLabeledControl(
                    _("Line style"),
                    wx.Choice,
                    choices=[_("Solid"), _("Dashed"), _("Dotted")],
                )
                lineStyleCombobox.SetSelection(
                    contextStyle.contextLineStyles.index(contextStyle.lineStyle)
                )
                contextStyle.lineStyleCtrl = lineStyleCombobox
        except Exception as e:
            log.error(
                f"HighlighterPlusStyleDialog::makeSettings() Failed to make settings: {e}"
            )

    def __init__(self, parent, settings):
        self.settings = settings
        self.title = _("Highlighter styles")
        super().__init__(parent, multiInstanceAllowed=True)

    def onOk(self, evt):
        for colSetting in self.settings._contextStyles.values():
            colSetting.on_ok(evt)
        self.settings.UpdateConfig()

        super().onOk(evt)

    def onCancel(self, evt):

        super().onCancel(evt)
