"""
Babbage B.V., 2019-2023

visionEnhancementProviders/_base/libUSA.py
Used to interact with the libUSA.dll
"""

import os
from ctypes import *
from ctypes.wintypes import *

import gui
import wx
from logHandler import log

from . import access4uReturnCodes

log.info("importing libUSA")


class MAGCOLOREFFECT(Structure):
    _fields_ = (("transform", c_float * 5 * 5),)


winLicenseError = 7056

MagnifierType = c_int
MAG_TYPE_SCREEN = 0
MAG_TYPE_WINDOW = 1

magWindows = dict()


def errCheck(result, func, args):
    if result > 0:
        raise WinError(result)
    return args


# Prototype declarations
InitializeFuncType = WINFUNCTYPE(DWORD, POINTER(HWND), MagnifierType)
InitializeArgTypes = (
    (2, "pHwnd"),
    (1, "magnifierType", MAG_TYPE_SCREEN),
)
ActivateLicenseFuncType = WINFUNCTYPE(c_wchar_p, c_wchar_p, POINTER(BOOL))
DeactivateLicenseFuncType = WINFUNCTYPE(c_wchar_p, POINTER(BOOL))
IsLicensedFuncType = WINFUNCTYPE(c_bool)

LicensingProductIdsFuncType = WINFUNCTYPE(POINTER(c_int))
LicensingProductNamesFuncType = WINFUNCTYPE(POINTER(c_char_p))
GetCurrentVersionProductIdFuncType = WINFUNCTYPE(c_int)

IsReleaseBuildFuncType = WINFUNCTYPE(c_bool)
SetupLoggingFuncType = WINFUNCTYPE(None)
InitFuncType = WINFUNCTYPE(None)
HelloFuncType = WINFUNCTYPE(None)
TerminateFuncType = WINFUNCTYPE(None)
ConfigUpdateType = WINFUNCTYPE(None, c_char_p)

SetTrackingRectangleType = WINFUNCTYPE(None, RECT, c_bool, c_char_p)

SetHighlighterFocusType = WINFUNCTYPE(None, RECT)
SetHighlighterNavigatorType = WINFUNCTYPE(None, RECT)
SetHighlighterBrowseModeType = WINFUNCTYPE(None, RECT)
SetHighlighterFocusNavigatorType = WINFUNCTYPE(None, RECT)


class base_libUSA_Handler:
    def __init__(self) -> None:
        self.libUSA = None
        self.libUSA_initialized = False
        pass

    def load_libUSA(self):
        try:
            self.libUSA = windll.libUSA
        except:
            self.libUSA = windll.LoadLibrary(
                os.path.join(os.path.abspath(os.path.dirname(__file__)), "usa.dll")
            )

        self._ActivateLicense = ActivateLicenseFuncType(
            ("USAActivateLicense", self.libUSA), ((1, "licenseKey"), (1, "success"))
        )
        self._DeactivateLicense = DeactivateLicenseFuncType(
            ("USADeactivateLicense", self.libUSA), ((1, "success"),)
        )
        self.IsLicensed = IsLicensedFuncType(("USAIsLicensed", self.libUSA))

        self.GetProductIds = LicensingProductIdsFuncType(
            ("USALicensingProductIds", self.libUSA)
        )
        self.GetProductNames = LicensingProductNamesFuncType(
            ("USALicensingProductNames", self.libUSA)
        )
        self.GetCurrentVersionProductId = GetCurrentVersionProductIdFuncType(
            ("USAGetCurrentVersionProductId", self.libUSA)
        )

        self.IsReleaseBuild = IsReleaseBuildFuncType(("IsReleaseBuild", self.libUSA))
        self.SetupLogging = SetupLoggingFuncType(("SetupLogging", self.libUSA))
        self.Init = InitFuncType(("USAInit", self.libUSA))
        self.Hello = HelloFuncType(("USAHello", self.libUSA))
        self.Terminate = TerminateFuncType(("USATerminate", self.libUSA))
        self._UpdateConfig = ConfigUpdateType(
            ("USAConfigUpdate", self.libUSA), ((1, "path"),)
        )

        self.SetTrackingRectangle = SetTrackingRectangleType(
            ("USASetTrackingRectangle", self.libUSA),
            ((1, "rect"), (1, "focus_change_only"), (1, "context")),
        )

        self.SetHighlighterFocus = SetHighlighterFocusType(
            ("USASetHighlighterFocus", self.libUSA), ((1, "rect"),)
        )

        self.SetHighlighterNavigator = SetHighlighterNavigatorType(
            ("USASetHighlighterNavigator", self.libUSA), ((1, "rect"),)
        )

        self.SetHighlighterBrowseMode = SetHighlighterBrowseModeType(
            ("USASetHighlighterBrowseMode", self.libUSA), ((1, "rect"),)
        )

        self.SetHighlighterFocusNavigator = SetHighlighterFocusNavigatorType(
            ("USASetHighlighterFocusNavigator", self.libUSA), ((1, "rect"),)
        )

    def initialize_libUSA(self):
        self.SetupLogging()
        self.Init()
        self.Hello()

        self.libUSA_initialized = True


class libUSA_Handler(base_libUSA_Handler):
    def __init__(self) -> None:
        self.initialized = False
        super().__init__()
        self.magWindows = {}

    def terminate(self) -> None:
        if self.libUSA:
            log.info("calling USATerminate")
            self.Terminate()
        self.initialized = False

    def Load(self):
        self.load_libUSA()

    def Initialize(self):

        if not self.libUSA:
            self.initialize_libUSA()
        try:
            log.info("Trying to init USA")
        except OSError as e:
            if e.winerror == 5:  # ERROR_ACCESS_DENIED
                gui.messageBox(
                    _("An error has occured:\n{error}").format(
                        error=_(
                            "Could not load the libUSA dll. Please check whether any other magnifiers are active on the system."
                        )
                    ),
                    _("Error"),
                    wx.OK | wx.ICON_WARNING,
                )
                raise WinError(e.winerror)
        self.initialized = True

    def ActivateLicense(self, licenseKey):
        success = BOOL()
        message = self._ActivateLicense(licenseKey, byref(success))
        return (success, message)

    def DeactivateLicense(self):
        success = BOOL()
        message = self._DeactivateLicense(byref(success))
        return (success, message)

    def UpdateConfig(self, path):
        if not self.initialized:
            return

        self._UpdateConfig(path)


handler: libUSA_Handler = libUSA_Handler()
