# A part of NonVisual Desktop Access (NVDA)
# This file is covered by the GNU General Public License.
# See the file COPYING for more details.
# Copyright (C) 2018-2023 NV Access Limited, Babbage B.V., Takuya Nishimoto


from ctypes.wintypes import RECT
from threading import Timer
from typing import Tuple

import api
import vision
import wx
from logHandler import log
from vision.constants import Context
from vision.util import getContextRect

from .libUSA import libUSA

_supportedContexts = (Context.FOCUS, Context.NAVIGATOR, Context.BROWSEMODE)


class InfiniteTimer:
    """A Timer class that does not stop, unless you want it to."""

    def __init__(self, seconds, target):
        self._should_continue = False
        self.is_running = False
        self.seconds = seconds
        self.target = target
        self.thread = None

    def _handle_target(self):
        self.is_running = True
        self.target()
        self.is_running = False
        self._start_timer()

    def _start_timer(self):
        if (
            self._should_continue
        ):  # Code could have been running when cancel was called.
            self.thread = Timer(self.seconds, self._handle_target)
            self.thread.start()

    def start(self):
        if not self._should_continue and not self.is_running:
            self._should_continue = True
            self._start_timer()
        else:
            print("Timer already started or running, please wait if you're restarting.")

    def cancel(self):
        if self.thread is not None:
            self._should_continue = (
                False  # Just in case thread is running and cancel fails.
            )
            self.thread.cancel()
        else:
            print("Timer never started or failed to initialize.")


class HighlighterPlus:
    _refreshInterval = 100
    enabledContexts: Tuple[Context]  # type info for autoprop: L{_get_enableContexts}

    def __init__(self, USAinstance):
        super().__init__()
        log.debug("Starting NVDAhighlighterPlus")

        self.started = False

        self.USA = USAinstance
        self.contextToRectMap = {}
        can_start = True
        providers = vision.handler.getActiveProviderInfos()
        for provider in providers:
            if provider.providerId == "NVDAHighlighter":
                log.debug("NVDA Highlighter is enabled")
                can_start = False
        if self.getSettings().get_highlighter_plus() and can_start:
            self.start()

    def start(self):
        self._timer = InfiniteTimer(0.1, self.refresh)
        self._timer.start()

        self.started = True

    def getSettings(self):
        return self.USA.getSettings()

    def terminate(self):
        log.debug("Terminating NVDAhighlighterPlus")

        if not self.started:
            return

        self.contextToRectMap.clear()
        self._timer.cancel()

    def handleUpdateThread(self):
        self.handleFocusChange(obj=api.getFocusObject())
        self.handleReviewMove()
        self.updateContexts()

    def updateContextRect(self, context, rect=None, obj=None):
        """Updates the position rectangle of the highlight for the specified context.
        If rect is specified, the method directly writes the rectangle to the contextToRectMap.
        Otherwise, it will call L{getContextRect}
        """
        if context not in self.get_enabledContexts():
            return
        if rect is None:
            try:
                rect = getContextRect(context, obj=obj)
            except (LookupError, NotImplementedError, RuntimeError, TypeError):
                rect = None

        self.contextToRectMap[context] = rect

    def handleFocusChange(self, obj):
        self.updateContextRect(context=Context.FOCUS, obj=obj)
        if not api.isObjectInActiveTreeInterceptor(obj):
            self.contextToRectMap.pop(Context.BROWSEMODE, None)
        else:
            self.handleBrowseModeMove()

    def handleReviewMove(self):
        self.updateContextRect(context=Context.NAVIGATOR)

    def handleBrowseModeMove(self):
        self.updateContextRect(context=Context.BROWSEMODE)

    def refresh(self):
        """Refreshes the screen positions of the enabled highlights."""
        wx.CallAfter(vision.handler.extensionPoints.post_coreCycle.notify)

    def get_enabledContexts(self):
        """Gets the contexts for which the highlighterPlus is enabled."""
        try:
            var = tuple(
                context
                for context in _supportedContexts
                if self.getSettings().get_highlighter_plus_enabled_by_mode(context)
            )
            return var
        except Exception as e:
            log.error(f"Failed to get enabledContexts: {e}")

    def updateContexts(self):
        contextRects = {}

        focus = api.getFocusObject()
        vbuf = focus.treeInterceptor

        for context in self.get_enabledContexts():
            rect = self.contextToRectMap.get(context)
            if not rect:
                continue
            # If vbuf.passThrough == True then we are in focus mode
            # See nvda/source/browseMode.py:reportPassThrough
            elif (
                context == Context.NAVIGATOR and contextRects.get(Context.FOCUS) == rect
            ):
                # When the focus overlaps the navigator object, which is usually the case,
                # show a different highlight style.
                # Focus is in contextRects, do not show the standalone focus highlight.
                contextRects.pop(Context.FOCUS)
                # Navigator object might be in contextRects as well
                contextRects.pop(Context.NAVIGATOR, None)
                context = Context.FOCUS_NAVIGATOR
            contextRects[context] = (
                rect  # Use libUSA highlighter functions instead of GDI+ painting
            )
        invalidRect = RECT(-1, -1, -1, -1)

        try:
            # Handle FOCUS_NAVIGATOR context
            if Context.FOCUS_NAVIGATOR in contextRects:
                rect = contextRects[Context.FOCUS_NAVIGATOR]
                rectStruct = rect.toRECT()
                libUSA.handler.SetHighlighterFocusNavigator(rectStruct)

                # When FOCUS_NAVIGATOR is active, disable individual FOCUS and NAVIGATOR
                libUSA.handler.SetHighlighterFocus(invalidRect)
                libUSA.handler.SetHighlighterNavigator(invalidRect)

                if vision._isDebug():
                    log.debug(
                        f"Set highlighter for FOCUS_NAVIGATOR context with rect {rect}"
                    )
                    log.debug("Set invalid rectangles for FOCUS and NAVIGATOR contexts")
            else:
                # No FOCUS_NAVIGATOR, disable it
                libUSA.handler.SetHighlighterFocusNavigator(invalidRect)

                # Handle individual FOCUS context
                if Context.FOCUS in contextRects:
                    rect = contextRects[Context.FOCUS]
                    rectStruct = rect.toRECT()
                    libUSA.handler.SetHighlighterFocus(rectStruct)

                    if vision._isDebug():
                        log.debug(f"Set highlighter for FOCUS context with rect {rect}")
                else:
                    libUSA.handler.SetHighlighterFocus(invalidRect)
                    if vision._isDebug():
                        log.debug("Set invalid rectangle for FOCUS context")

                # Handle individual NAVIGATOR context
                if Context.NAVIGATOR in contextRects:
                    rect = contextRects[Context.NAVIGATOR]
                    rectStruct = rect.toRECT()
                    libUSA.handler.SetHighlighterNavigator(rectStruct)

                    if vision._isDebug():
                        log.debug(
                            f"Set highlighter for NAVIGATOR context with rect {rect}"
                        )
                else:
                    libUSA.handler.SetHighlighterNavigator(invalidRect)
                    if vision._isDebug():
                        log.debug("Set invalid rectangle for NAVIGATOR context")

            # Handle BROWSEMODE context (independent of FOCUS_NAVIGATOR)
            if Context.BROWSEMODE in contextRects:
                rect = contextRects[Context.BROWSEMODE]
                rectStruct = rect.toRECT()
                libUSA.handler.SetHighlighterBrowseMode(rectStruct)

                if vision._isDebug():
                    log.debug(
                        f"Set highlighter for BROWSEMODE context with rect {rect}"
                    )
            else:
                libUSA.handler.SetHighlighterBrowseMode(invalidRect)
                if vision._isDebug():
                    log.debug("Set invalid rectangle for BROWSEMODE context")

        except Exception as e:
            log.error(f"Error setting highlighter: {e}")
