"""
Babbage B.V., 2019-2023

visionEhancementProviders/_base/baseSettings.py
Implements some logic on some controls in wx used in the settingspanel of NVDA vision.
The split in abstract USA settings and the regular USA settings in settings.py used to be because of the split in USA
and USA remote, but this was removed to reduce overhead.
"""

import weakref
from ctypes.wintypes import RECT
from operator import gt, lt
from typing import Optional

import addonHandler
import config
import globalPluginHandler
from autoSettingsUtils import driverSetting
from autoSettingsUtils.utils import StringParameterInfo
from logHandler import log
from vision.constants import Context
from vision.providerBase import (
    VisionEnhancementProvider,
    VisionEnhancementProviderSettings,
)

# We need to initialize translation and localization support:
addonHandler.initTranslation()
import ui
import winUser
from vision.util import getContextRect
from wx import Display

_contextOptionLabelsWithAccelerators = {
    Context.FOCUS: _("Follow &focus"),
    Context.FOREGROUND: _("Follow foreground &window changes"),
    Context.NAVIGATOR: _("Follow &navigator object"),
    Context.CARET: _("Follow system &caret"),
    Context.BROWSEMODE: _("Follow &browse mode cursor"),
    Context.REVIEW: _("Follow &review cursor"),
    Context.MOUSE: _("Follow &mouse"),
}

_contextBoundaryLabelsWithAccelerators = {
    Context.FOCUS: _("F&ocus boundary"),
    Context.CARET: _("C&aret boundary"),
    Context.BROWSEMODE: _("Browse mode cursor boundar&y"),
    Context.MOUSE: _("Mouse boun&dary"),
}

_contextBoundaryDefaults = {
    Context.FOCUS: (25, 25, 75, 75),
    Context.CARET: (10, 10, 90, 90),
    Context.BROWSEMODE: (20, 20, 80, 80),
    Context.MOUSE: (10, 10, 90, 90),
}


predefinedStepLevels = (
    0.01,
    0.05,
    0.1,
    0.15,
    0.2,
    0.25,
    0.3,
    0.40,
    0.5,
    0.6,
    0.75,
    1.0,
    1.25,
    1.5,
    1.75,
    2.0,
)

predefinedNumberOfMagnificationSteps = (
    1,
    5,
    10,
    15,
    20,
    25,
    30,
    35,
    40,
    45,
    50,
    55,
    60,
    65,
    70,
    75,
    80,
    85,
    90,
    95,
    100,
)

predefinedMagnificationStepTime_ms = (
    1,
    5,
    10,
    15,
    20,
    25,
    30,
    35,
    40,
    45,
    50,
    55,
    60,
    65,
    70,
    75,
    80,
    85,
    90,
    95,
    100,
)

predefinedMagnificationLevels = (
    1.0,
    1.25,
    1.5,
    1.75,
    2.0,
    2.5,
    3.0,
    4.0,
    5.0,
    6.5,
    8.0,
    10.0,
    12.5,
    15.0,
    17.5,
    20.0,
    25.0,
    30.0,
    37.5,
    45.0,
    52.5,
    60.0,
    70.0,
    80.0,
)

predefinedZoomFrames = (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)

predefinedMagnificationSmoothness = (1, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100)


class TupleDriverSetting(driverSetting.DriverSetting):

    def __init__(
        self,
        id,
        displayNameWithAccelerator,
        defaultVal,
        min: int = 4,
        max: int = 4,
        displayName: Optional[str] = None,
        useConfig: bool = True,
    ):
        super().__init__(
            id,
            displayNameWithAccelerator,
            availableInSettingsRing=False,
            defaultVal=defaultVal,
            displayName=displayName,
            useConfig=useConfig,
        )
        self.min = min
        self.max = max

    def _get_configSpec(self):
        return (
            f"int_list(default=list{self.defaultVal}, min={self.min}, max={self.max})"
        )


class FloatDriverSetting(driverSetting.DriverSetting):

    def _get_configSpec(self):
        return f"float(default={self.defaultVal})"


class StringWithValueParameterInfo(StringParameterInfo):
    def __init__(self, id, name, value):
        super().__init__(id, name)
        self.value = value


class AbstractUSASettings(VisionEnhancementProviderSettings):

    @classmethod
    def _get_conf(cls):
        return config.conf[cls._getConfigSection()][cls.getId()]

    def __init__(self):
        self._providerInstRef = lambda: None
        self.cachedMagnificationLevel = [None, None, None, None]
        self.json_conf = None
        super().__init__()

    def registerProviderInstance(self, inst: VisionEnhancementProvider):
        self._providerInstRef = weakref.ref(inst)
        # ensure any previously saved settings are loaded from config file:
        self._initSpecificSettings(self, self._getAvailableRuntimeSettings())

    def unregisterProviderInstance(self):
        if self._providerInstRef():
            self._saveSpecificSettings(self, self._getAvailableRuntimeSettings())
        self._providerInstRef = lambda: None

    def _get_supportedSettings(self):
        if not self._providerInstRef():
            return []
        return self._getAvailableRuntimeSettings()

    def _getAvailableRuntimeSettings(self):
        settings = [()]
        return settings

    def set_magnification_level_by_window(self, windowIndex: int, val: float): ...

    def get_magnification_level_by_window(self, windowIndex: int): ...

    def _setMagnificationByStepHelper(self, increase=True):
        # hardcoded the minimum and maximum level
        maxLevel = 100
        minLevel = 1
        window_index = self._get_current_window_index()
        if self.cachedMagnificationLevel[window_index]:
            level = self.cachedMagnificationLevel[window_index]
            self.cachedMagnificationLevel[window_index] = None
        else:
            curLevel = self.get_magnification_level_by_window(window_index)
            level = (
                curLevel * (1 + self.get_magnification_step_level())
                if increase
                else curLevel / (1 + self.get_magnification_step_level())
            )
            level = max(min(maxLevel, level), minLevel)
        self.set_magnification_level_by_window(window_index, level)
        return level

    def _setMagnificationByLevelHelper(self, increase=True):
        window_index = self._get_current_window_index()
        if self.cachedMagnificationLevel[window_index]:
            level = self.cachedMagnificationLevel[window_index]
            self.cachedMagnificationLevel[window_index] = None
        else:
            curLevel = self.get_magnification_level_by_window(window_index)
            levels = (iter if increase else reversed)(predefinedMagnificationLevels)
            for level in levels:
                if (gt if increase else lt)(level, curLevel):
                    break
            else:
                level = curLevel

        self.set_magnification_level_by_window(window_index, level)
        return level

    def _setMagnificationLevelHelper(self, increase=True):
        if self.get_magnification_fixed_levels():
            return self._setMagnificationByLevelHelper(increase)
        else:
            return self._setMagnificationByStepHelper(increase)

    def _setMagnificationSmoothnessHelper(self, increase=True):
        curLevel = self.get_lerping_speed()
        levels = (iter if increase else reversed)(predefinedMagnificationSmoothness)
        for level in levels:
            if (gt if increase else lt)(level, curLevel):
                self.set_lerping_speed(level)
                break
        else:
            self.set_lerping_speed(curLevel)
        return self.get_lerping_speed()

    def _toggleMagnifierHelper(self):
        window_index = self._get_current_window_index()
        if self.cachedMagnificationLevel[window_index]:
            self.set_magnification_level_by_window(
                window_index, self.cachedMagnificationLevel[window_index]
            )
            self.cachedMagnificationLevel[window_index] = None
            return True
        else:
            self.cachedMagnificationLevel[window_index] = (
                self.get_magnification_level_by_window(window_index)
            )
            self.set_magnification_level_by_window(window_index, 1.0)
            return False

    @classmethod
    def _get_availableMagnificationlevels(self):
        return {
            val: StringParameterInfo(val, str(val))
            for val in predefinedMagnificationLevels
        }

    @classmethod
    def _get_availableSteplevels(self):
        return {val: StringParameterInfo(val, str(val)) for val in predefinedStepLevels}

    @classmethod
    def _get_availableMagnificationstepss(self):
        return {
            val: StringParameterInfo(val, str(val))
            for val in predefinedNumberOfMagnificationSteps
        }

    @classmethod
    def _get_availableMagnificationsteptime_mss(self):
        return {
            val: StringParameterInfo(val, str(val))
            for val in predefinedMagnificationStepTime_ms
        }

    @classmethod
    def _get_availableMagnificationsmoothnesss(self):
        return {
            val: StringParameterInfo(val, str(val))
            for val in predefinedMagnificationSmoothness
        }

    def _get_current_window_json(self):
        displays = [Display(i).GetGeometry() for i in range(Display.GetCount())]
        if not self.get_mouse_focus_mode():
            try:
                coordinates = getContextRect("focus")
            except:
                coordinates = getContextRect("foreground")
        else:
            coordinates = winUser.getCursorPos()
        for i in range(len(displays)):
            if self.is_in_rect(displays[i], coordinates):
                return self.json_conf["usa"]["windows"][i]

    # Doesn't take mouse movement into account, might break with windows switching
    def _get_current_window_index(self):
        if self.is_fullscreen_magnifier_engine():
            # Always use index 0 for fullscreen magnifier engine
            return 0

        displays = [Display(i).GetGeometry() for i in range(Display.GetCount())]
        if not self.get_mouse_focus_mode():
            try:
                coordinates = getContextRect("focus")
            except:
                coordinates = getContextRect("foreground")
        else:
            coordinates = winUser.getCursorPos()
        for i in range(len(displays)):
            if self.is_in_rect(displays[i], coordinates):
                return i

    def is_in_rect(self, rect, coordinates):
        return (
            coordinates[0] >= rect[0]
            and coordinates[0] <= rect[0] + rect[2]
            and coordinates[1] >= rect[1]
            and coordinates[1] <= rect[1] + rect[3]
        )
