"""
Babbage B.V., 2019-2023

visionEnhancementProviders/_base/settings.py
contains the main code for the settings of the USA vision enhancement provider.
"""

import copy
import json
import os
from collections import namedtuple
from enum import Enum
from sys import exc_info

import globalVars
import winGDI
import wx  # type: ignore
from colors import RGB
from vision.constants import Context

from .._implementation.magnifierModes import magnifier_modes
from .baseSettings import *
from .baseSettings import (
    _contextBoundaryDefaults,
    _contextBoundaryLabelsWithAccelerators,
    _contextOptionLabelsWithAccelerators,
)
from .configSpec import configSpec
from .libUSA import libUSA

# We need to initialize translation and localization support:
addonHandler.initTranslation()


def getScreenEdgesString(i):
    # Translator: xgettext:no-python-format
    return _("within {i}% from screen edges").format(i=i)


class HighlightStyle(
    namedtuple("HighlightStyle", ("color", "width", "style", "margin"))
):
    """Represents the style of a highlight for a particular context.
    @ivar color: The color to use for the style
    @type color: L{RGB}
    @ivar width: The width of the lines to be drawn, in pixels.
            A higher width reduces the inner dimensions of the rectangle.
            Therefore, if you need to increase the outer dimensions of the rectangle,
            you need to increase the margin as well.
    @type width: int
    @ivar style: The style of the lines to be drawn;
            One of the C{winGDI.DashStyle*} enumeration constants.
    @type style: int
    @ivar margin: The number of pixels between the highlight's rectangle
            and the rectangle of the object to be highlighted.
            A higher margin stretches the highlight's rectangle.
            This value may also be negative.
    @type margin: int
    """


addonHandler.initTranslation()


class ContextStyle:
    margin = 10
    contextStyleLevelStrings = [
        "1.0",
        "2.0",
        "3.0",
        "4.0",
        "5.0",
        "6.0",
        "7.0",
        "8.0",
        "9.0",
        "10.0",
    ]

    contextLineStyles = [
        "solid",
        "dashed",
        "dotted",
    ]

    def __init__(self, lineStyle: str, context: Context, parent):
        log.debug("Initialize context styles")
        self.parent = parent
        self.context = context
        self.color = parent.get_highlighter_plus_color_by_mode(self.context)
        self.thickness = parent.get_highlighter_plus_thickness_by_mode(self.context)
        self.margin = parent.get_highlighter_plus_thickness_by_mode(self.context)
        self.lineStyle = lineStyle

        self.colorWheel = None
        self.thicknessCtrl = None
        self.marginCtrl = None
        self.lineStyleCtrl = None

    def update_thickness_control(self, evt):
        stepLevelIndex = self.thicknessCtrl.CurrentSelection
        self.thickness = float(self.contextStyleLevelStrings[stepLevelIndex])
        self.parent.set_highlighter_plus_thickness_by_mode(self.context, self.thickness)

    def update_margin_control(self, evt):
        stepLevelIndex = self.marginCtrl.CurrentSelection
        self.margin = float(self.contextStyleLevelStrings[stepLevelIndex])
        self.parent.set_highlighter_plus_margin_by_mode(self.context, self.margin)

    def update_line_style_control(self, evt):
        stepLevelIndex = self.lineStyleCtrl.CurrentSelection
        self.lineStyle = self.contextLineStyles[stepLevelIndex]
        self.parent.set_highlighter_plus_line_style_by_mode(
            self.context, self.lineStyle
        )

    def on_ok(self, evt):
        self.update_thickness_control(evt)
        self.update_margin_control(evt)
        self.update_line_style_control(evt)
        self.on_spin_control_change(evt)

    def on_spin_control_change(self, evt):
        self.color = self.colorWheel.GetColour().Get(False)
        self.parent.set_highlighter_plus_color_by_mode(self.context, self.color)


class USASettings(AbstractUSASettings):

    highlightPlusFocus = False
    highlightPlusNavigator = False
    highlightPlusBrowseMode = False

    @classmethod
    def getId(cls) -> str:
        return "usa"

    @classmethod
    def getDisplayName(cls) -> str:
        return "Ultimate Screen Access"

    def post_init_Settings(self):
        if self.json_initialized:
            self._contextStyles = {
                Context.FOCUS: ContextStyle("dashed", Context.FOCUS, self),
                Context.NAVIGATOR: ContextStyle("solid", Context.NAVIGATOR, self),
                Context.FOCUS_NAVIGATOR: ContextStyle(
                    "solid", Context.FOCUS_NAVIGATOR, self
                ),
                Context.BROWSEMODE: ContextStyle("solid", Context.BROWSEMODE, self),
            }

    def initSettings(self):
        super().initSettings()

        self.has_backup = False
        self.json_initialized = False

        # transfer the strings in customColorScheme to proper floats, since it is saved as strings
        if config.conf[self._getConfigSection()][self.getId()].get("customColorScheme"):
            for key in ["r", "g", "b", "a", "t"]:
                jj = [
                    eval(ii)
                    for ii in config.conf[self._getConfigSection()][self.getId()][
                        "customColorScheme"
                    ][key]
                ]
                config.conf[self._getConfigSection()][self.getId()][
                    "customColorScheme"
                ][key] = jj

        # fgColorschemes is not properly initialized the first time, so it is here
        if not config.conf[self._getConfigSection()][self.getId()].get(
            "fgColorschemes"
        ):
            config.conf[self._getConfigSection()][self.getId()]["fgColorschemes"] = {}
        else:
            # if we already have a aggerated section than we need to set the 'None'strings to actual None
            for key, item in list(
                config.conf[self._getConfigSection()][self.getId()][
                    "fgColorschemes"
                ].items()
            ):

                if item == "None":
                    config.conf[self._getConfigSection()][self.getId()][
                        "fgColorschemes"
                    ][key] = None
                # config.conf[self._getConfigSection()][self.getId()]['fgColorschemes'][key] = ii

        if not config.conf[self._getConfigSection()][self.getId()].get(
            "globalBrightness"
        ):
            config.conf[self._getConfigSection()][self.getId()][
                "globalBrightness"
            ] = 100

        # update form the configSpec file if needed
        config.conf[self._getConfigSection()][self.getId()].spec.update(configSpec)

        self.get_autostart, self.set_autostart = self._create_config_access_functions(
            bool, "autostart"
        )

        self.get_engine, self.set_engine = self._create_config_access_functions(
            str, "usa", "engine"
        )

        self.get_mouse_focus_mode, self.set_mouse_focus_mode = (
            self._create_config_access_functions(bool, "mouse-focus-mode")
        )

        self.get_tracking_options, self.set_tracking_options = (
            self._create_config_access_functions(dict[str, bool], "object-tracking")
        )

        self.get_tracking_focus, self.set_tracking_focus = (
            self._create_config_access_functions(bool, "object-tracking", "focus")
        )
        self.get_tracking_foreground, self.set_tracking_foreground = (
            self._create_config_access_functions(bool, "object-tracking", "foreground")
        )
        self.get_tracking_navigator, self.set_tracking_navigator = (
            self._create_config_access_functions(bool, "object-tracking", "navigator")
        )
        self.get_tracking_caret, self.set_tracking_caret = (
            self._create_config_access_functions(bool, "object-tracking", "caret")
        )
        self.get_tracking_browsemode, self.set_tracking_browsemode = (
            self._create_config_access_functions(bool, "object-tracking", "browsemode")
        )
        self.get_tracking_review, self.set_tracking_review = (
            self._create_config_access_functions(bool, "object-tracking", "review")
        )

        self.get_tracking_mouse, self.set_tracking_mouse = (
            self._create_config_access_functions(
                bool, "usa", "tracking", "track-to-mouse"
            )
        )

        # Context-specific tracking boundaries
        self.get_tracking_boundary_focus, self.set_tracking_boundary_focus = (
            self._create_config_access_functions(
                float, "usa", "tracking", "boundaries", "focus"
            )
        )
        self.get_tracking_boundary_foreground, self.set_tracking_boundary_foreground = (
            self._create_config_access_functions(
                float, "usa", "tracking", "boundaries", "foreground"
            )
        )
        self.get_tracking_boundary_navigator, self.set_tracking_boundary_navigator = (
            self._create_config_access_functions(
                float, "usa", "tracking", "boundaries", "navigator"
            )
        )
        self.get_tracking_boundary_caret, self.set_tracking_boundary_caret = (
            self._create_config_access_functions(
                float, "usa", "tracking", "boundaries", "caret"
            )
        )
        self.get_tracking_boundary_browsemode, self.set_tracking_boundary_browsemode = (
            self._create_config_access_functions(
                float, "usa", "tracking", "boundaries", "browsemode"
            )
        )
        self.get_tracking_boundary_review, self.set_tracking_boundary_review = (
            self._create_config_access_functions(
                float, "usa", "tracking", "boundaries", "review"
            )
        )
        self.get_tracking_boundary_mouse, self.set_tracking_boundary_mouse = (
            self._create_config_access_functions(
                float, "usa", "tracking", "boundaries", "mouse"
            )
        )
        self.get_tracking_alignment, self.set_tracking_alignment = (
            self._create_config_access_functions(str, "usa", "tracking", "alignment")
        )

        # Setters are explicitly defined
        self.get_magnifier_mode, _ = self._create_config_access_functions(
            bool, "usa", "windows", self._get_current_window_index, "mode"
        )
        self.get_magnifier_mode_by_window, _ = (
            self._create_config_access_functions_by_window(
                bool, "usa", "windows", "mode"
            )
        )

    _contextStyles = {}

    def _getAvailableRuntimeSettings(self):

        settings = []

        return settings
        # return sorted list of settings based on what it does
        # return sorted(settings, key=lambda x: x._get_configSpec() )

    _availableBoundaries = {
        val: StringParameterInfo(list(val), name)
        for val, name in {
            (0, 0, 100, 100): _("Unrestricted"),
            (5, 5, 95, 95): getScreenEdgesString(5),
            (10, 10, 90, 90): getScreenEdgesString(10),
            (15, 15, 85, 85): getScreenEdgesString(15),
            (20, 20, 80, 80): getScreenEdgesString(20),
            (25, 25, 75, 75): getScreenEdgesString(25),
            (30, 30, 70, 70): getScreenEdgesString(30),
            (35, 35, 65, 65): getScreenEdgesString(35),
            (40, 40, 60, 60): getScreenEdgesString(40),
            (45, 45, 55, 55): getScreenEdgesString(45),
            (50, 50, 50, 50): _("Centered on screen"),
            (-1, -1, -1, -1): _("Relative"),
        }.items()
    }

    def _get_isRemote(self):
        if self.conf.get("isRemote") == "1" or self.conf.get("isRemote") == 1:
            return True
        else:
            return False

    def _set_isRemote(self, bool):
        if bool:
            self.conf["isRemote"] = "1"
        else:
            self.conf["isRemote"] = "0"

    def _get_availableRectcolors(self):

        return {
            "Blue": StringParameterInfo("Blue", str(_("Blue"))),
            "Pink": StringParameterInfo("Pink", str(_("Pink"))),
            "Yellow": StringParameterInfo("Yellow", str(_("Yellow"))),
        }

    def _get_remoteHandler(self):
        plugin = next(
            (
                p
                for p in globalPluginHandler.runningPlugins
                if p.__module__ == "globalPlugins.usa"
            ),
            None,
        )
        remoteHandler = plugin.remoteHandler if plugin else None
        return remoteHandler

    def _get_colorScheme(self):
        if not self.conf.get("colorScheme"):
            self.conf["colorScheme"] = "default"
        return self.conf["colorScheme"]

    def _set_colorScheme(self, transformation):
        if self.isRemote:
            pass
        else:
            try:
                info = self.availableColorschemes[transformation]
            except KeyError:
                log.error(
                    "Transformation not in list! Continuing with first item in list"
                )
                # this is a fail safe which should not have to be used. If the error above shows up, than you should definetly fix something
                firstKey = list(self.availableColorschemes.keys())[0]
                info = self.availableColorschemes.get(firstKey)
            for hwnd in libUSA.handler.magWindows:
                try:
                    libUSA.handler.SetColorScheme(None, info.value)
                except WindowsError:
                    log.error("Couldn't set color scheme", exc_info=True)
                    return
            self.conf["colorScheme"] = transformation

    _showSystemCursor: bool = True

    def _get_showSystemCursor(self):
        return self._showSystemCursor

    def _set_showSystemCursor(self, state):
        if self.isRemote:
            pass
        else:
            for hwnd in libUSA.handler.magWindows:
                try:
                    libUSA.handler.ShowSystemCursor(None, state)
                except WindowsError:
                    log.error(
                        "Couldn't change show system cursor for window %s" % None,
                        exc_info=True,
                    )
                    return
            self._showSystemCursor = state

    # Getter defines all the transformations.
    def _get_all_color_schemes_by_window(self, window_index: int):
        # if customColorScheme is not defined yet, define it as follows:
        # this was previously in configSpec but moved to here since the latter did not work properly.
        b = (int(self.get_brightness_by_window(window_index)) - 100.0) / 100.0
        custom_color_scheme = self.json_conf["usa"]["windows"][window_index][
            "custom-color-scheme"
        ]
        transformations = [
            StringWithValueParameterInfo(
                "default",
                _("True colors"),
                libUSA.MAGCOLOREFFECT(
                    (
                        (1.0, 0.0, 0.0, 0.0, 0.0),
                        (0.0, 1.0, 0.0, 0.0, 0.0),
                        (0.0, 0.0, 1.0, 0.0, 0.0),
                        (0.0, 0.0, 0.0, 1.0, 0.0),
                        (0.0 + b, 0.0 + b, 0.0 + b, 0.0, 1.0),
                    )
                ),
            ),
            StringWithValueParameterInfo(
                "inverted",
                _("Inverted colors"),
                libUSA.MAGCOLOREFFECT(
                    (
                        (-1.0, 0.0, 0.0, 0.0, 0.0),
                        (0.0, -1.0, 0.0, 0.0, 0.0),
                        (0.0, 0.0, -1.0, 0.0, 0.0),
                        (0.0, 0.0, 0.0, 1.0, 0.0),
                        (1.0 + b, 1.0 + b, 1.0 + b, 0.0, 1.0),
                    )
                ),
            ),
            StringWithValueParameterInfo(
                "gray",
                _("Gray scale"),
                libUSA.MAGCOLOREFFECT(
                    (
                        (0.3, 0.3, 0.3, 0.0, 0.0),
                        (0.6, 0.6, 0.6, 0.0, 0.0),
                        (0.1, 0.1, 0.1, 0.0, 0.0),
                        (0.0, 0.0, 0.0, 1.0, 0.0),
                        (0.0 + b, 0.0 + b, 0.0 + b, 0.0, 1.0),
                    )
                ),
            ),
            StringWithValueParameterInfo(
                "high-contrast",
                _("High contrast"),
                libUSA.MAGCOLOREFFECT(
                    (
                        (2.0, 0.0, 0.0, 0.0, 0.0),
                        (0.0, 2.0, 0.0, 0.0, 0.0),
                        (0.0, 0.0, 2.0, 0.0, 0.0),
                        (0.0, 0.0, 0.0, 1.0, 0.0),
                        (-0.7 + b, -0.7 + b, -0.7 + b, 0.0, 1.0),
                    )
                ),
            ),
            StringWithValueParameterInfo(
                "inverted-high-contrast",
                _("Inverted high contrast"),
                libUSA.MAGCOLOREFFECT(
                    (
                        (-2.0, 0.0, 0.0, 0.0, 0.0),
                        (0.0, -2.0, 0.0, 0.0, 0.0),
                        (0.0, 0.0, -2.0, 0.0, 0.0),
                        (0.0, 0.0, 0.0, 1.0, 0.0),
                        (1.7 + b, 1.7 + b, 1.7 + b, 0.0, 1.0),
                    )
                ),
            ),
            StringWithValueParameterInfo(
                "blue-light",
                _("Blue light"),
                libUSA.MAGCOLOREFFECT(
                    (
                        (0.8, 0.0, 0.0, 0.0, 0.0),
                        (0.0, 0.5, 0.0, 0.0, 0.0),
                        (0.0, 0.0, 0.0, 0.0, 0.0),
                        (0.0, 0.0, 0.0, 1.0, 0.0),
                        (0.0 + b, 0.0 + b, 0.0 + b, 0.0, 1.0),
                    )
                ),
            ),
            StringWithValueParameterInfo(
                "custom",
                _("Custom"),
                libUSA.MAGCOLOREFFECT(
                    (
                        tuple(custom_color_scheme[0]),
                        tuple(custom_color_scheme[1]),
                        tuple(custom_color_scheme[2]),
                        tuple(custom_color_scheme[3]),
                        tuple(
                            (
                                max(custom_color_scheme[4][0] + b, -0.99),
                                max(custom_color_scheme[4][1] + b, -0.99),
                                max(custom_color_scheme[4][2] + b, -0.99),
                                custom_color_scheme[4][3],
                                custom_color_scheme[4][4],
                            )
                        ),
                    )
                ),
            ),
        ]
        return {transformation.id: transformation for transformation in transformations}

    def _get_available_color_schemes_by_window(self, window_index: int):
        # this function looks at the settings and makes a list of all possible transformations. If none are selected,
        # all are returned
        allTransformations = self._get_all_color_schemes_by_window(window_index)
        selectedTransformations = {}
        get_scs = self.get_selected_color_scheme_by_window_key
        if get_scs(window_index, "default") == "True":
            selectedTransformations["default"] = allTransformations["default"]
        if get_scs(window_index, "inverted") == "True":
            selectedTransformations["inverted"] = allTransformations["inverted"]
        if get_scs(window_index, "gray") == "True":
            selectedTransformations["gray"] = allTransformations["gray"]
        if get_scs(window_index, "high-contrast") == "True":
            selectedTransformations["high-contrast"] = allTransformations[
                "high-contrast"
            ]
        if get_scs(window_index, "blue-light") == "True":
            selectedTransformations["blue-light"] = allTransformations["blue-light"]
        if get_scs(window_index, "inverted-high-contrast") == "True":
            selectedTransformations["inverted-high-contrast"] = allTransformations[
                "inverted-high-contrast"
            ]
        if get_scs(window_index, "custom") == "True":
            selectedTransformations["custom"] = allTransformations["custom"]

        if not selectedTransformations:
            selectedTransformations["default"] = allTransformations["default"]
            selectedTransformations["inverted"] = allTransformations["inverted"]
            selectedTransformations["gray"] = allTransformations["gray"]
            selectedTransformations["high-contrast"] = allTransformations[
                "high-contrast"
            ]
            selectedTransformations["inverted-high-contrast"] = allTransformations[
                "inverted-high-contrast"
            ]
            selectedTransformations["blue-light"] = allTransformations["blue-light"]
            selectedTransformations["custom"] = allTransformations["custom"]
            return allTransformations

        return selectedTransformations

    def is_fullscreen_magnifier_engine(self):
        conf = self.json_conf["usa"]
        return conf["engine"] == "fullscreen"

    def get_magnifier_mode_strings(self):
        return (
            [
                _("Per screen"),
                _("Split"),
                _("Docked"),
                _("Lens"),
                _("Resizing lens"),
                _("None"),
                _("Camera"),
            ]
            if self.get_show_system_cursor()
            else [
                _("Per screen"),
                _("Split"),
                _("Lens"),
                _("Resizing lens"),
                _("Camera"),
            ]
        )

    def get_magnifier_mode_values(self):
        return (
            [
                "per-screen",
                "split",
                "docked",
                "lens",
                "resizing-lens",
                "none",
                "camera",
            ]
            if self.get_show_system_cursor()
            else [
                "per-screen",
                "split",
                "lens",
                "resizing-lens",
                "camera",
            ]
        )

    def to_json(self):
        return json.dumps(self.conf.dict())

    JSON_VERSION = 2

    def _same_json_version(self):
        return self.json_conf["version"] == self.JSON_VERSION

    # Example upgrade map with all available methods
    # JSON_UPGRADE_MAP = {
    #     2: [
    #         {"method": "once", "value": {"usa": {"new": "test"}}},
    #         {"method": "per-window", "value": {"new": "test"}},
    #         {"method": "per-mag", "value": {"new": "test"}},
    #         {"method": "del-once", "value": ["usa", "new"]},
    #         {"method": "del-per-window", "value": ["new"]},
    #         {"method": "del-per-mag", "value": ["new"]},
    #     ]
    # }
    JSON_UPGRADE_MAP = {
        1: [
            {"method": "once", "value": {"usa": {"engine": "multiwindow"}}},
        ],
        2: [
            {
                "method": "once",
                "value": {
                    "usa": {
                        "tracking": {
                            "boundaries": {
                                "focus": -1,
                                "foreground": -1,
                                "navigator": -1,
                                "caret": -1,
                                "browsemode": -1,
                                "review": -1,
                                "mouse": -1,
                            }
                        },
                        "highlighter-plus": {
                            "enabled": False,
                            "focus": {
                                "enabled": True,
                                "color": [255, 0, 0],
                                "thickness": 5.0,
                                "margin": 5.0,
                                "line-style": "dashed",
                            },
                            "navigator": {
                                "enabled": True,
                                "color": [0, 255, 0],
                                "thickness": 5.0,
                                "margin": 5.0,
                                "line-style": "solid",
                            },
                            "browseMode": {
                                "enabled": True,
                                "color": [255, 0, 255],
                                "thickness": 5.0,
                                "margin": 5.0,
                                "line-style": "solid",
                            },
                            "focusNavigator": {
                                "enabled": True,
                                "color": [0, 0, 255],
                                "thickness": 5.0,
                                "margin": 5.0,
                                "line-style": "solid",
                            },
                        },
                    }
                },
            },
        ],
    }

    def _upgrade_json(self):
        # https://stackoverflow.com/a/7205107
        def merge(a: dict, b: dict, path=[]):
            for key in b:
                if key in a:
                    if isinstance(a[key], dict) and isinstance(b[key], dict):
                        merge(a[key], b[key], path + [str(key)])
                    elif a[key] != b[key]:
                        raise Exception("Conflict at " + ".".join(path + [str(key)]))
                else:
                    a[key] = b[key]
            return a

        def delete_key_by_path(d, path):
            for key in path[:-1]:
                if key in d and isinstance(d[key], dict):
                    d = d[key]
                else:
                    raise Exception(f"Path {path} not found")
            target_key = path[-1]
            if target_key in d:
                del d[target_key]

        if self.json_conf["version"] > self.JSON_VERSION:
            log.warn(
                f"Config json version {self.json_conf['version']} is greater than JSON_VERSION {self.JSON_VERSION}"
            )
            return

        while self.json_conf["version"] < self.JSON_VERSION:
            version = self.json_conf["version"] + 1

            upgrade_list = self.JSON_UPGRADE_MAP[version]

            try:
                for upgrade in upgrade_list:
                    method = upgrade["method"]
                    value = upgrade["value"]
                    if method == "once":
                        merge(self.json_conf, value)
                    elif method == "per-window":
                        for w in self.json_conf["usa"]["windows"]:
                            merge(w, value)
                    elif method == "per-mag":
                        for w in self.json_conf["usa"]["windows"]:
                            for m in w["magnification-windows"]:
                                merge(m, value)
                    elif method == "del-once":
                        delete_key_by_path(self.json_conf, value)
                    elif method == "del-per-window":
                        for w in self.json_conf["usa"]["windows"]:
                            delete_key_by_path(w, value)
                    elif method == "del-per-mag":
                        for w in self.json_conf["usa"]["windows"]:
                            for m in w["magnification-windows"]:
                                delete_key_by_path(m, value)
                    else:
                        log.error(f"Invalid json upgrade method: {method}")
            except Exception as e:
                log.error(
                    f"Error while upgrading json config from version {self.json_conf['version']} to version {version}: {e}"
                )
                return

            self.json_conf["version"] = version

        self.write_to_json()

    def initialize_json(self, recursive_call: bool = False):
        if self.json_initialized:
            return True

        try:
            with open(self._get_config_file_path(), "r") as file:
                self.json_conf = json.load(file)

                if "converted-to-json" not in self.conf:
                    self.convert_conf_to_json()
                    self.conf["converted-to-json"] = True
                    config.conf.save()

                self.read_cam_viewer_conf()

                self.json_initialized = True

                # Add default version 0 if no version is present
                if "version" not in self.json_conf:
                    self.json_conf["version"] = 0
                # Check and upgrade json config version
                if not self._same_json_version():
                    self._upgrade_json()
                return True
        except FileNotFoundError:
            config_path = self._get_config_file_path()
            log.warning(f"The file usa-config.json was not found at {config_path}.")
            if not recursive_call and libUSA.handler.IsReleaseBuild():
                import shutil

                addon_config_path = (
                    f"{globalVars.appArgs.configPath}\\addons\\usa\\usa-config.json"
                )
                shutil.copy2(addon_config_path, config_path)
                return self.initialize_json(True)
            return False
        return False

    def backup_json(self):
        if not self.has_backup:
            self.json_backup = copy.deepcopy(self.json_conf)
            self.cam_viewer_backup = copy.deepcopy(self.cam_viewer_conf)
            self.has_backup = True

    def restore_from_backup(self):
        if self.has_backup:
            self.json_conf = copy.deepcopy(self.json_backup)
            self.cam_viewer_conf = copy.deepcopy(self.cam_viewer_backup)
            self.has_backup = False

    def delete_backup(self):
        self.has_backup = False

    def _get_config_file_path(self):
        config_subfolder = "" if libUSA.handler.IsReleaseBuild() else "addons\\USA\\"
        return f"{globalVars.appArgs.configPath}\\{config_subfolder}usa-config.json"

    def write_to_json(self):
        try:
            if self.json_conf == None or not self.json_initialized:
                log.warn(
                    "Trying to write empty config to file, trying to initialize json."
                )
                self.initialize_json()
                if not self.json_initialized:
                    log.error("Failed to initialize json")
                    return
            with open(self._get_config_file_path(), "w") as file:
                json.dump(self.json_conf, file, indent=4)
                return True
        except FileNotFoundError:
            log.warning(
                f"The file usa-config.json was not found in the add-on root folder."
            )
        return False

    def UpdateConfig(self):
        self.write_to_json()

        libUSA.handler.UpdateConfig(self._get_config_file_path().encode())

    def convert_conf_to_json(self):
        conf = self.conf

        try:
            usa_json = self.json_conf["usa"]
            window = usa_json["windows"]

            # Global values
            usa_json["global-mode"] = "normal"
            usa_json["lerping-speed"] = float(conf["MagnificationSmoothness"])
            usa_json["smooth-magnification"] = True
            usa_json["magnification-step-level"] = float(conf["stepLevel"])

            self.json_conf["object-tracking"]["focus"] = bool(conf["trackToFocus"])
            self.json_conf["object-tracking"]["foreground"] = bool(
                conf["trackToForeground"]
            )
            self.json_conf["object-tracking"]["navigator"] = bool(
                conf["trackToNavigator"]
            )
            self.json_conf["object-tracking"]["caret"] = bool(conf["trackToCaret"])
            self.json_conf["object-tracking"]["browsemode"] = bool(
                conf["trackToBrowseMode"]
            )
            self.json_conf["object-tracking"]["review"] = bool(conf["trackToReview"])

            for i in range(len(usa_json["windows"])):
                window[i]["mode"] = "per-screen"
                window[i]["magnification-windows"][0]["area"]["left"] = 0.0
                window[i]["magnification-windows"][0]["area"]["top"] = 0.0
                window[i]["magnification-windows"][0]["area"]["right"] = 1.0
                window[i]["magnification-windows"][0]["area"]["bottom"] = 1.0
                window[i]["brightness"] = float(conf["globalBrightness"])
                window[i]["magnification-windows"][0]["magnification-level"] = float(
                    conf["magnificationLevel"]
                )

                # Custom color schemes
                window[i]["color-scheme"] = conf["colorScheme"]
                window[i]["selected-color-scheme"]["default"] = conf["selectedDefault"]
                window[i]["selected-color-scheme"]["inverted"] = conf[
                    "selectedInverted"
                ]
                window[i]["selected-color-scheme"]["gray"] = conf["selectedGray"]
                window[i]["selected-color-scheme"]["high-contrast"] = conf[
                    "selectedHighContrast"
                ]
                window[i]["selected-color-scheme"]["inverted-high-contrast"] = conf[
                    "selectedInvertedHighContrast"
                ]
                window[i]["selected-color-scheme"]["blue-light"] = conf[
                    "selectedBlueLight"
                ]
                window[i]["selected-color-scheme"]["custom"] = conf["selectedCustom"]

                # Color Scheme loop
                try:
                    rowToConfKeyMap = ("r", "g", "b", "a", "t")
                    for row_key in rowToConfKeyMap:
                        row_index = rowToConfKeyMap.index(row_key)
                        for col_index in range(len(conf["customColorScheme"][row_key])):
                            window[i]["custom-color-scheme"][row_index][col_index] = (
                                float(conf["customColorScheme"][row_key][col_index])
                            )
                except Exception as e:
                    log.warning(f"Passed on in incorrect matrix structure. Error: {e}")
            log.info(f"Finished converting ini to json")
        except Exception as e:
            log.error(f"Failed to convert ini to json: {e}")

    def add_window_to_json(self, copy_index=0):
        self.json_conf["usa"]["windows"].append(
            copy.deepcopy(self.json_conf["usa"]["windows"][copy_index])
        )
        added_entry_mag_windows = self.json_conf["usa"]["windows"][-1][
            "magnification-windows"
        ]
        for mag in added_entry_mag_windows:
            mag["panning-area-monitor-index"] = (
                len(self.json_conf["usa"]["windows"]) - 1
            )

    def _create_config_access_functions(self, type, *args):
        def get_variable() -> type:
            current = self.json_conf
            for arg in args:
                if callable(arg):
                    current = current[arg()]
                else:
                    current = current[arg]
            return current

        def set_variable(var: type):
            current = self.json_conf
            for arg in args[:-1]:
                if callable(arg):
                    current = current[arg()]
                else:
                    current = current[arg]
            current[args[-1]] = var

        return get_variable, set_variable

    def _create_config_access_functions_by_window(self, type, *args):
        def get_variable(window_idx: int) -> type:
            current = self.json_conf
            for arg in args:
                if callable(arg):
                    current = current[arg()]
                else:
                    current = current[arg]
                    if arg == "windows":
                        current = current[window_idx]
            return current

        def set_variable(window_idx: int, var: type):
            current = self.json_conf
            for arg in args[:-1]:
                if callable(arg):
                    current = current[arg()]
                else:
                    current = current[arg]
                    if arg == "windows":
                        current = current[window_idx]
            current[args[-1]] = var

        return get_variable, set_variable

    def _set_magnifier_mode_areas(self, window_json, mode: str):
        host_area = window_json["area"]
        mag_area_0 = window_json["magnification-windows"][0]["area"]
        mag_area_1 = window_json["magnification-windows"][1]["area"]

        # Set host area to whole screen here, overwrite later if necessary
        host_area["left"] = 0.0
        host_area["top"] = 0.0
        host_area["right"] = 1.0
        host_area["bottom"] = 1.0

        # Something goes wrong when switching to split screen mode, at least with 3 monitors

        if mode == "per-screen":
            mag_area_0["left"] = 0.0
            mag_area_0["top"] = 0.0
            mag_area_0["right"] = 1.0
            mag_area_0["bottom"] = 1.0
        elif mode == "split":
            mag_area_0["left"] = 0.0
            mag_area_0["top"] = 0.0
            mag_area_0["right"] = 0.5
            mag_area_0["bottom"] = 1.0

            mag_area_1["left"] = 0.5
            mag_area_1["top"] = 0.0
            mag_area_1["right"] = 1.0
            mag_area_1["bottom"] = 1.0
        elif mode in ["docked", "lens", "resizing-lens"]:
            mag_area_0["left"] = 0.0
            mag_area_0["top"] = 0.0
            mag_area_0["right"] = 1.0
            mag_area_0["bottom"] = 1.0

            host_area["left"] = 0.0
            host_area["top"] = 0.0
            host_area["right"] = 0.25
            host_area["bottom"] = 0.45
        elif mode == "camera":
            mag_area_0["left"] = 0.0
            mag_area_0["top"] = 0.0
            mag_area_0["right"] = 1.0
            mag_area_0["bottom"] = 1.0

            host_area["left"] = 0.0
            host_area["top"] = 0.0
            host_area["right"] = 0.5
            host_area["bottom"] = 1.0

    def switchMagnifierMode(self):
        if self.isRemote:
            self.remoteHandler.sendVisionCommandOverUnicorn(
                command="switchMagnifierMode"
            )
        else:
            if self.is_fullscreen_magnifier_engine():
                ui.message(_("Cannot adjust magnifier mode in fullscreen engine."))
                return

            json = self._get_current_window_json()
            host_area = json["area"]
            mag_area_0 = json["magnification-windows"][0]["area"]
            mag_area_1 = json["magnification-windows"][1]["area"]

            # Set host area to whole screen here, overwrite later if necessary
            host_area["left"] = 0.0
            host_area["top"] = 0.0
            host_area["right"] = 1.0
            host_area["bottom"] = 1.0

            if json["mode"] == "split":
                # vertical split Left to horizontal split top
                if mag_area_0["top"] == mag_area_1["top"] and mag_area_0["left"] == 0:
                    mag_area_0["left"] = 0.0
                    mag_area_0["top"] = 0.0
                    mag_area_0["right"] = 1.0
                    mag_area_0["bottom"] = 0.5

                    mag_area_1["left"] = 0.0
                    mag_area_1["top"] = 0.5
                    mag_area_1["right"] = 1.0
                    mag_area_1["bottom"] = 1.0

                # These 2 setups don't allow for proper border resizing.
                # horizontal split top to vertical split right
                elif (
                    mag_area_0["left"] == mag_area_1["left"] and mag_area_0["top"] == 0
                ):
                    mag_area_0["left"] = 0.5
                    mag_area_0["top"] = 0.0
                    mag_area_0["right"] = 1.0
                    mag_area_0["bottom"] = 1.0

                    mag_area_1["left"] = 0.0
                    mag_area_1["top"] = 0.0
                    mag_area_1["right"] = 0.5
                    mag_area_1["bottom"] = 1.0

                # vertical split right to horizontal split bottom
                elif mag_area_0["top"] == mag_area_1["top"] and mag_area_1["left"] == 0:
                    mag_area_0["left"] = 0.0
                    mag_area_0["top"] = 0.5
                    mag_area_0["right"] = 1.0
                    mag_area_0["bottom"] = 1.0

                    mag_area_1["left"] = 0.0
                    mag_area_1["top"] = 0.0
                    mag_area_1["right"] = 1.0
                    mag_area_1["bottom"] = 0.5

                # horizontal split bottom to docked
                # only if system cursor is enabled
                elif self.get_show_system_cursor():
                    json["mode"] = "docked"
                    mag_area_0["left"] = 0.0
                    mag_area_0["top"] = 0.0
                    mag_area_0["right"] = 1.0
                    mag_area_0["bottom"] = 1.0

                    host_area["left"] = 0.0
                    host_area["top"] = 0.0
                    host_area["right"] = 0.25
                    host_area["bottom"] = 0.45

                # skip 'docked' mode if system cursor is disabled
                else:
                    mag_area_0["left"] = 0.0
                    mag_area_0["top"] = 0.0
                    mag_area_0["right"] = 1.0
                    mag_area_0["bottom"] = 1.0

                    host_area["left"] = 0.0
                    host_area["top"] = 0.0
                    host_area["right"] = 0.25
                    host_area["bottom"] = 0.45

                    json["mode"] = "lens"

            elif json["mode"] == "docked":
                mag_area_0["left"] = 0.0
                mag_area_0["top"] = 0.0
                mag_area_0["right"] = 1.0
                mag_area_0["bottom"] = 1.0

                host_area["left"] = 0.0
                host_area["top"] = 0.0
                host_area["right"] = 0.25
                host_area["bottom"] = 0.45

                json["mode"] = "lens"

            elif json["mode"] == "lens":
                mag_area_0["left"] = 0.0
                mag_area_0["top"] = 0.0
                mag_area_0["right"] = 1.0
                mag_area_0["bottom"] = 1.0

                host_area["left"] = 0.0
                host_area["top"] = 0.0
                host_area["right"] = 0.25
                host_area["bottom"] = 0.45

                json["mode"] = "resizing-lens"

            elif json["mode"] == "resizing-lens":
                mag_area_0["left"] = 0.0
                mag_area_0["top"] = 0.0
                mag_area_0["right"] = 1.0
                mag_area_0["bottom"] = 1.0

                host_area["left"] = 0.0
                host_area["top"] = 0.0
                host_area["right"] = 0.5
                host_area["bottom"] = 1.0

                json["mode"] = "camera"

            # If system cursor is enabled
            elif json["mode"] == "camera" and self.get_show_system_cursor():
                json["mode"] = "none"

            # Skip 'none' mode if system cursor is disabled
            elif json["mode"] == "none" or (
                json["mode"] == "camera" and not self.get_show_system_cursor()
            ):
                mag_area_0["left"] = 0.0
                mag_area_0["top"] = 0.0
                mag_area_0["right"] = 1.0
                mag_area_0["bottom"] = 1.0

                json["mode"] = "per-screen"
            # per screen back to vertical split left
            else:
                json["mode"] = "split"
                mag_area_0["left"] = 0.0
                mag_area_0["top"] = 0.0
                mag_area_0["right"] = 0.5
                mag_area_0["bottom"] = 1.0

                mag_area_1["left"] = 0.5
                mag_area_1["top"] = 0.0
                mag_area_1["right"] = 1.0
                mag_area_1["bottom"] = 1.0

            m = next(
                m
                for m in magnifier_modes
                if m.name == json["mode"]
                and m.mag_area_0.left == mag_area_0["left"]
                and m.mag_area_0.top == mag_area_0["top"]
                and m.mag_area_0.right == mag_area_0["right"]
                and m.mag_area_0.bottom == mag_area_0["bottom"]
            )
            ui.message(_("Set magnifier mode to ") + m.display_name)

            self.UpdateConfig()

    def get_windows(self):
        return self.json_conf["usa"]["windows"]

    def set_color_scheme_by_window(self, windowIndex: int, val: str):
        conf = self.json_conf["usa"]
        conf["windows"][windowIndex]["color-scheme"] = val

    def set_color_scheme(self, val: str):
        self.set_color_scheme_by_window(self._get_current_window_index(), val)

    def set_magnification_level_by_window(self, windowIndex: int, val: float):
        conf = self.json_conf["usa"]
        conf["windows"][windowIndex]["magnification-windows"][0][
            "magnification-level"
        ] = val

    def set_magnification_level(self, val: float):
        self.set_magnification_level_by_window(self._get_current_window_index(), val)

    def set_magnification_step_level(self, val: float):
        conf = self.json_conf["usa"]
        conf["magnification-step-level"] = val

    def set_brightness_by_window(self, windowIndex: int, val: float):
        conf = self.json_conf["usa"]
        conf["windows"][windowIndex]["brightness"] = val

    def set_custom_color_scheme_by_window(self, windowIndex: int, val):
        conf = self.json_conf["usa"]
        try:
            for row_index in range(5):
                for col_index in range(len(val[row_index])):
                    conf["windows"][windowIndex]["custom-color-scheme"][row_index][
                        col_index
                    ] = float(val[row_index][col_index])
        except Exception as e:
            log.error(f"set_custom_color_scheme_by_window() {e}")

    def set_custom_color_scheme(self, val):
        self.set_custom_color_scheme_by_window(self._get_current_window_index(), val)

    def set_custom_color_scheme_settings_by_window(self, windowIndex: int, val):
        conf = self.json_conf["usa"]
        conf["windows"][windowIndex]["custom-color-scheme-settings"] = val

    def set_selected_color_scheme_by_window_key(
        self, windowIndex: int, key: str, val: str
    ):
        conf = self.json_conf["usa"]
        conf["windows"][windowIndex]["selected-color-scheme"][key] = val

    def set_lerping_speed(self, val: int):
        conf = self.json_conf["usa"]
        conf["lerping-speed"] = val

    def set_bitmap_smoothing(self, val: bool):
        conf = self.json_conf["usa"]
        conf["bitmap-smoothing"] = val

    def set_smooth_magnification(self, val: bool):
        conf = self.json_conf["usa"]
        conf["smooth-magnification"] = val

    def get_color_scheme_by_window(self, windowIndex: int):
        conf = self.json_conf["usa"]
        return conf["windows"][windowIndex]["color-scheme"]

    def get_color_scheme(self):
        return self.get_color_scheme_by_window(self._get_current_window_index())

    def get_magnification_level_by_window(self, windowIndex: int):
        conf = self.json_conf["usa"]
        return conf["windows"][windowIndex]["magnification-windows"][0][
            "magnification-level"
        ]

    def get_magnification_level(self):
        return self.get_magnification_level_by_window(self._get_current_window_index())

    def get_magnification_step_level(self):
        conf = self.json_conf["usa"]
        return conf["magnification-step-level"]

    def get_brightness_by_window(self, windowIndex: int):
        conf = self.json_conf["usa"]
        return conf["windows"][windowIndex]["brightness"]

    def get_custom_color_scheme_by_window(self, windowIndex: int):
        conf = self.json_conf["usa"]
        return conf["windows"][windowIndex]["custom-color-scheme"]

    def get_custom_color_scheme(self):
        return self.get_custom_color_scheme_by_window(self._get_current_window_index())

    def get_custom_color_scheme_settings_by_window(self, windowIndex: int):
        conf = self.json_conf["usa"]
        return conf["windows"][windowIndex]["custom-color-scheme-settings"]

    def get_custom_color_scheme_settings(self):
        return self.get_custom_color_scheme_settings_by_window(
            self._get_current_window_index()
        )

    def get_selected_color_scheme_by_window_key(self, windowIndex: int, key: str):
        conf = self.json_conf["usa"]
        return conf["windows"][windowIndex]["selected-color-scheme"][key]

    def get_selected_color_scheme_by_window(self, windowIndex: int):
        conf = self.json_conf["usa"]
        return conf["windows"][windowIndex]["selected-color-scheme"]

    def get_total_amount_windows(self):
        conf = self.json_conf["usa"]
        return len(conf["windows"])

    def get_lerping_speed(self):
        conf = self.json_conf["usa"]
        return conf["lerping-speed"]

    def get_bitmap_smoothing(self):
        conf = self.json_conf["usa"]
        return conf["bitmap-smoothing"]

    def get_smooth_magnification(self):
        conf = self.json_conf["usa"]
        return conf["smooth-magnification"]

    def set_show_system_cursor(self, var: bool):
        conf = self.json_conf["usa"]
        conf["show-system-cursor"] = var

    def get_show_system_cursor(self):
        conf = self.json_conf["usa"]
        return conf["show-system-cursor"]

    def set_quality_mode(self, var: bool):
        conf = self.json_conf["usa"]
        conf["quality-mode"] = var

    def get_quality_mode(self):
        conf = self.json_conf["usa"]
        return conf["quality-mode"]

    def set_magnifier_mode_by_window(self, window_index: int, var: str):
        conf = self.json_conf["usa"]
        conf["windows"][window_index]["mode"] = var

        m = next(m for m in magnifier_modes if m.name == var)
        ui.message(_("Set magnifier mode to ") + m.display_name)
        self._set_magnifier_mode_areas(conf["windows"][window_index], var)

    def set_magnifier_mode(self, var: str):
        self.set_magnifier_mode_by_window(self._get_current_window_index(), var)

    # Crosshair settings

    def set_crosshair_enable(self, var: bool):
        conf = self.json_conf["usa"]["crosshair"]
        conf["enable"] = var

    def get_crosshair_enable(self):
        conf = self.json_conf["usa"]["crosshair"]
        return conf["enable"]

    def set_crosshair_color(self, var: int):
        conf = self.json_conf["usa"]["crosshair"]
        conf["color"] = var

    def get_crosshair_color(self):
        conf = self.json_conf["usa"]["crosshair"]
        return conf["color"]

    def set_crosshair_width(self, var: float):
        conf = self.json_conf["usa"]["crosshair"]
        conf["width"] = var

    def get_crosshair_width(self):
        conf = self.json_conf["usa"]["crosshair"]
        return conf["width"]

    def set_crosshair_center_radius(self, var: int):
        conf = self.json_conf["usa"]["crosshair"]
        conf["center-radius"] = var

    def get_crosshair_center_radius(self):
        conf = self.json_conf["usa"]["crosshair"]
        return conf["center-radius"]

    def set_crosshair_min_distance(self, var: int):
        conf = self.json_conf["usa"]["crosshair"]
        conf["min-distance"] = var

    def get_crosshair_min_distance(self):
        conf = self.json_conf["usa"]["crosshair"]
        return conf["min-distance"]

    def set_crosshair_enable_center_radius(self, var: bool):
        conf = self.json_conf["usa"]["crosshair"]
        conf["enable-center-radius"] = var

    def get_crosshair_enable_center_radius(self):
        conf = self.json_conf["usa"]["crosshair"]
        return conf["enable-center-radius"]

    ###
    # Non-dll related settings
    ###

    # Word follower settings

    def set_word_follower(self, var: bool):
        self.json_conf["word-follower"]["enabled"] = var

    def get_word_follower(self):
        return self.json_conf["word-follower"]["enabled"]

    def set_word_follower_global(self, var: bool):
        self.json_conf["word-follower"]["enabled-global"] = var

    def get_word_follower_global(self):
        return self.json_conf["word-follower"]["enabled-global"]

    def set_word_follower_caret_jump(self, var: bool):
        self.json_conf["word-follower"]["caret-jump"] = var

    def get_word_follower_caret_jump(self):
        return self.json_conf["word-follower"]["caret-jump"]

    def set_word_follower_show_rect(self, var: bool):
        self.json_conf["word-follower"]["show-rect"] = var

    def get_word_follower_show_rect(self):
        return self.json_conf["word-follower"]["show-rect"]

    def set_word_follower_rect_color(self, var: str):
        self.json_conf["word-follower"]["rect-color"] = var

    def get_word_follower_rect_color(self):
        return self.json_conf["word-follower"]["rect-color"]

    def set_word_follower_rect_thickness(self, var: int):
        self.json_conf["word-follower"]["rect-thickness"] = var

    def get_word_follower_rect_thickness(self):
        return self.json_conf["word-follower"]["rect-thickness"]

    # End of word follower settings

    # highlighter plus settings
    def set_highlighter_plus(self, var: bool):
        self.json_conf["usa"]["highlighter-plus"]["enabled"] = var

    def get_highlighter_plus(self):
        return self.json_conf["usa"]["highlighter-plus"]["enabled"]

    def set_highlighter_plus_enabled_by_mode(self, mode: str, var: bool):
        self.json_conf["usa"]["highlighter-plus"][mode]["enabled"] = var

    def get_highlighter_plus_enabled_by_mode(self, mode: str):
        return self.json_conf["usa"]["highlighter-plus"][mode]["enabled"]

    def set_highlighter_plus_color_by_mode(self, mode: str, var):
        self.json_conf["usa"]["highlighter-plus"][mode]["color"] = var

    def get_highlighter_plus_color_by_mode(self, mode: str):
        return self.json_conf["usa"]["highlighter-plus"][mode]["color"]

    def set_highlighter_plus_margin_by_mode(self, mode: str, var):
        self.json_conf["usa"]["highlighter-plus"][mode]["margin"] = var

    def get_highlighter_plus_margin_by_mode(self, mode: str):
        return self.json_conf["usa"]["highlighter-plus"][mode]["margin"]

    def set_highlighter_plus_thickness_by_mode(self, mode: str, var):
        self.json_conf["usa"]["highlighter-plus"][mode]["thickness"] = var

    def get_highlighter_plus_thickness_by_mode(self, mode: str):
        return self.json_conf["usa"]["highlighter-plus"][mode]["thickness"]

    def set_highlighter_plus_line_style_by_mode(self, mode: str, var: str):
        self.json_conf["usa"]["highlighter-plus"][mode]["line-style"] = var

    def get_highlighter_plus_line_style_by_mode(self, mode: str):
        return self.json_conf["usa"]["highlighter-plus"][mode]["line-style"]

    # End of highlighter plus settings

    def set_object_reader(self, var: bool):
        self.json_conf["object-reader"] = var

    def get_object_reader(self):
        return self.json_conf["object-reader"]

    def set_magnification_fixed_levels(self, var: bool):
        self.json_conf["magnification-fixed-levels"] = var

    def get_magnification_fixed_levels(self):
        return self.json_conf["magnification-fixed-levels"]

    def set_quick_settings_on_startup(self, var: bool):
        self.json_conf["quick-settings-on-startup"] = var

    def get_quick_settings_on_startup(self):
        return self.json_conf["quick-settings-on-startup"]

    # def set_autostart(self, var: bool):
    # 	self.json_conf["autostart"] = var

    # def get_autostart(self):
    # 	return self.json_conf["autostart"]

    def set_follow_caret(self, val: bool):
        self.json_conf["follow-caret"] = val

    def get_follow_caret(self):
        return self.json_conf["follow-caret"]

    # Cam-Viewer settings

    def read_cam_viewer_conf(self):
        path = f"{globalVars.appArgs.configPath}\\addons\\USA\\cam-viewer.json"
        try:
            with open(path, "r") as file:
                self.cam_viewer_conf = json.load(file)
        except FileNotFoundError:
            log.warning(
                f"The file cam-viewer.json was not found in the add-on root folder."
            )

    def write_cam_viewer_conf(self):
        path = f"{globalVars.appArgs.configPath}\\addons\\USA\\cam-viewer.json"
        try:
            if self.cam_viewer_conf == None:
                import traceback

                traceback.print_stack()
                log.warn("Trying to write empty config to file, aborting.")
                return
            with open(path, "w") as file:
                json.dump(self.cam_viewer_conf, file, indent=4)
                return True
        except FileNotFoundError:
            log.warning(
                f"The file cam-viewer.json was not found in the add-on root folder."
            )
        return False

    def get_cam_viewer_brightness(self):
        return self.cam_viewer_conf["transformations"]["brightness"]

    def set_cam_viewer_brightness(self, val: float):
        self.cam_viewer_conf["transformations"]["brightness"] = val

    def get_cam_viewer_contrast(self):
        return self.cam_viewer_conf["transformations"]["contrast"]

    def set_cam_viewer_contrast(self, val: float):
        self.cam_viewer_conf["transformations"]["contrast"] = val

    def get_cam_viewer_grayscale(self):
        return self.cam_viewer_conf["transformations"]["grayscale"]

    def set_cam_viewer_grayscale(self, val: bool):
        self.cam_viewer_conf["transformations"]["grayscale"] = val

    def get_cam_viewer_inverse(self):
        return self.cam_viewer_conf["transformations"]["inverse"]

    def set_cam_viewer_inverse(self, val: bool):
        self.cam_viewer_conf["transformations"]["inverse"] = val

    def get_cam_viewer_color_offset(self):
        color_offset = self.cam_viewer_conf["transformations"]["color_offset"]
        return (
            int(color_offset["v2"]),
            int(color_offset["v1"]),
            int(color_offset["v0"]),
        )

    def set_cam_viewer_color_offset(self, val: tuple[int, int, int]):
        color_offset = self.cam_viewer_conf["transformations"]["color_offset"]
        color_offset["v2"] = val[0]
        color_offset["v1"] = val[1]
        color_offset["v0"] = val[2]

    def get_cam_viewer_webcam_index(self):
        return self.cam_viewer_conf["webcam-index"]

    def set_cam_viewer_webcam_index(self, val: int):
        self.cam_viewer_conf["webcam-index"] = val

    def get_cam_viewer_max_webcam_index(self):
        return self.cam_viewer_conf["max-webcam-index"]

    def get_cam_viewer_area(self):
        area = self.cam_viewer_conf["window"]
        return [area["x"], area["y"], area["width"], area["height"]]

    def set_cam_viewer_area(self, val):
        area = self.cam_viewer_conf["window"]
        area["x"] = val[0]
        area["y"] = val[1]
        area["width"] = val[2]
        area["height"] = val[3]

    def get_cam_viewer_orientation(self):
        return self.cam_viewer_conf["window"]["orientation"]

    def set_cam_viewer_orientation(self, val: int):
        self.cam_viewer_conf["window"]["orientation"] = val
