import ctypes
import re
import subprocess
import threading
import time

import api
import globalCommands
import globalVars
import NVDAObjects
import ui
import winUser
import wx
from controlTypes.role import Role
from logHandler import log

from .._base.settings import USASettings


class CamViewer:
    def __init__(self, settings: USASettings):
        self.settings = settings
        self.process = None

    def _set_area(self, axis: int, display_left: bool, display_area, cam_area):
        if axis == 0:
            display_width = display_area[2]

            if display_left:
                cam_area[0] = display_area[0]
            else:
                cam_area[0] = display_area[0] + display_width / 2

            cam_area[1] = display_area[1]
            cam_area[2] = display_width / 2
            cam_area[3] = display_area[3]
        else:
            display_height = display_area[3]

            if display_left:
                cam_area[1] = display_area[1]
            else:
                cam_area[1] = display_area[1] + display_height / 2

            cam_area[0] = display_area[0]
            cam_area[2] = display_area[2]
            cam_area[3] = display_height / 2

        self.settings.set_cam_viewer_area(cam_area)
        self.settings.write_cam_viewer_conf()

    def _set_window_camera_mode(self, window):
        cam_viewer_orientation = self.settings.get_cam_viewer_orientation()
        window["mode"] = "camera"
        host_area = window["area"]
        mag_0_area = window["magnification-windows"][0]["area"]
        mag_0_area["left"] = 0.0
        mag_0_area["top"] = 0.0
        mag_0_area["right"] = 1.0
        mag_0_area["bottom"] = 1.0
        if cam_viewer_orientation < 2:
            host_area["left"] = 0.0 if cam_viewer_orientation == 0 else 0.5
            host_area["top"] = 0.0
            host_area["right"] = 0.5 if cam_viewer_orientation == 0 else 1.0
            host_area["bottom"] = 1.0
        else:
            host_area["left"] = 0.0
            host_area["top"] = 0.0 if cam_viewer_orientation == 2 else 0.5
            host_area["right"] = 1.0
            host_area["bottom"] = 0.5 if cam_viewer_orientation == 2 else 1.0
        return host_area

    def set_next_position(self, keep_position: bool = False):
        displays = [wx.Display(i).GetGeometry() for i in range(wx.Display.GetCount())]

        cur_area = self.settings.get_cam_viewer_area()

        for i, d in enumerate(displays):
            if self.settings.is_in_rect(
                d, (cur_area[0], cur_area[1])
            ) and self.settings.is_in_rect(
                d, (cur_area[0] + cur_area[2], cur_area[1] + cur_area[3])
            ):
                cur_display = d
                cur_display_idx = i
                break

        if cur_display == None:
            log.error("script_nextCamViewerPosition: Failed to find current monitor")
            return

        # 0 for left/right side of screen, 1 for top/bottom
        axis = 0 if self.settings.get_cam_viewer_orientation() < 2 else 1

        if cur_area[axis] == cur_display[axis]:
            display_left = False
        else:
            display_left = True
            if not keep_position:
                cur_display_idx = (cur_display_idx + 1) % len(displays)
                cur_display = displays[cur_display_idx]

        self._set_area(axis, display_left, cur_display, cur_area)

        self.settings.set_cam_viewer_area(cur_area)
        self.settings.write_cam_viewer_conf()

    def start(self):
        ui.message(_("Starting Cam-Viewer"))
        path = f"{globalVars.appArgs.configPath}\\addons\\USA\\lib"

        if not self.process or self.process.poll() != None:
            # Set cam viewer position to screen where camera magnifier mode is active
            windows = self.settings.get_windows()
            cam_viewer_orientation = self.settings.get_cam_viewer_orientation()
            cam_magnifier_monitor_idx = 0

            for i, w in enumerate(windows):
                if w["mode"] == "camera":
                    cam_magnifier_monitor_idx = i
                    host_area = self._set_window_camera_mode(w)
                    break
            else:
                cam_magnifier_monitor_idx = self.settings._get_current_window_index()
                w = windows[cam_magnifier_monitor_idx]
                host_area = self._set_window_camera_mode(w)

            display = wx.Display(cam_magnifier_monitor_idx).GetGeometry()

            cam_area = self.settings.get_cam_viewer_area()
            # 0 for left/right side of screen, 1 for top/bottom
            axis = 0 if cam_viewer_orientation < 2 else 1

            if host_area["left" if axis == 0 else "right"] == 0:
                display_left = False
            else:
                display_left = True

            self._set_area(axis, display_left, display, cam_area)

            self.settings.UpdateConfig()

            self.process = subprocess.Popen([path + "\\cam-viewer.exe"], cwd=path)
            # Give the cam viewer 5s time to discover all webcams
            wx.CallLater(5000, self.settings.read_cam_viewer_conf)
        else:
            # Find cam viewer window handle and setforegroundwindow on it
            window_handle = None
            WNDENUMPROC = ctypes.WINFUNCTYPE(
                ctypes.wintypes.BOOL, ctypes.wintypes.HWND, ctypes.wintypes.LPARAM
            )

            @WNDENUMPROC
            def callback(window, data):
                nonlocal window_handle
                pid, _ = winUser.getWindowThreadProcessID(window)
                if pid == self.process.pid:
                    window_handle = window
                    return False
                return True

            ctypes.windll.user32.EnumWindows(callback, 0)
            if window_handle != None:
                ui.message(_("Set focus to Cam-Viewer"))
                winUser.setForegroundWindow(window_handle)

    def stop(self):
        ui.message(_("Stopping Cam-Viewer"))
        if not self.process or self.process.poll() != None:
            return

        self.process.terminate()

        windows = self.settings.get_windows()
        for i, w in enumerate(windows):
            if w["mode"] == "camera":
                self.settings.set_magnifier_mode_by_window(i, "per-screen")
        self.settings.UpdateConfig()

    def switch_orientation(self):
        cam_viewer_orientation = self.settings.get_cam_viewer_orientation()
        cam_viewer_orientation = (cam_viewer_orientation + 1) % 4
        self.settings.set_cam_viewer_orientation(cam_viewer_orientation)

        # Adjust the camera mode host area
        windows = self.settings.get_windows()

        for w in windows:
            if w["mode"] == "camera":
                self._set_window_camera_mode(w)

        self.settings.UpdateConfig()

        self.set_next_position(keep_position=True)


class A4UReadAble:
    def __init__(self, settings: USASettings):
        self.settings = settings
        self.processes = []
        self.check_thread = None
        self.saved_magnification_level = 1.0

    def _find_descendant_windows(
        self, parent, visible=None, controlID=None, className=None
    ):
        result = []
        WNDENUMPROC = ctypes.WINFUNCTYPE(
            ctypes.wintypes.BOOL, ctypes.wintypes.HWND, ctypes.wintypes.LPARAM
        )

        @WNDENUMPROC
        def callback(window, data):
            if (
                (visible is None or winUser.isWindowVisible(window) == visible)
                and (not controlID or winUser.getControlID(window) == controlID)
                and (not className or winUser.getClassName(window) == className)
            ):
                result.append(window)
                return True
            return True

        ctypes.windll.user32.EnumChildWindows(parent, callback, 0)
        try:
            return result
        except IndexError:
            raise LookupError("No matching descendant window found")

    def _get_chrome_current_url(self, fg, nav):
        windows = self._find_descendant_windows(
            fg.parent.parent.windowHandle, className="Chrome_RenderWidgetHostHWND"
        )
        # Find documents with URL
        documents = {}
        for w in windows:
            document = NVDAObjects.IAccessible.getNVDAObjectFromEvent(
                w,
                winUser.OBJID_CLIENT,
                0,
            )
            acc_value: str = document.IAccessibleObject.accValue(0)
            if acc_value != None and re.match("(https?)|(file)://", acc_value):
                documents[document] = acc_value

        return next(
            (documents[d] for d in documents if d.windowHandle == nav.windowHandle),
            None,
        )

    def _test_for_running_processes(self):
        while True:
            if self.settings.is_fullscreen_magnifier_engine():
                # Filter all terminated A4U ReadAble process handles out
                self.processes = list(p for p in self.processes if p.poll() == None)
                if len(self.processes) == 0:
                    # Reanable magnification
                    self.settings.set_magnification_level_by_window(
                        0, self.saved_magnification_level
                    )
                    self.settings.UpdateConfig()
                    return
            # Check if NVDA+space was unbound for Accessbility Viewer
            # elif "kb:nvda+space" not in globalCommands.commands._gestureMap:
            #     # Filter all terminated A4U ReadAble process handles out
            #     self.processes = list(p for p in self.processes if p.poll() == None)
            #     # No more running A4U ReadAble processes
            #     if len(self.processes) == 0:
            #         # Rebind NVDA+space
            #         globalCommands.commands.bindGesture(
            #             "kb:nvda+space", "toggleVirtualBufferPassThrough"
            #         )
            #         return
            else:
                return
            time.sleep(2)

    def start(self):
        windows = self.settings.get_windows()

        # Check if there is a window with camera mode
        i = next((i for i, el in enumerate(windows) if el["mode"] == "camera"), -1)

        # If there is at least one window with camera mode, pick the first without camera mode
        if i != -1:
            i = next((i for i, el in enumerate(windows) if el["mode"] != "camera"), 0)

        # Get current URL of browser
        fg = api.getForegroundObject()
        window_class_name = fg.windowClassName
        if window_class_name == "Chrome_WidgetWin_1":
            nav = api.getNavigatorObject()
            url = self._get_chrome_current_url(fg, nav)
        elif window_class_name == "MozillaWindowClass":
            toolbar = next(
                x
                for x in fg.children
                if x.role == Role.TOOLBAR and x.name == "Navigation"
            )
            combobox = next(x for x in toolbar.children if x.role == Role.COMBOBOX)
            url = combobox.children[0].value
        else:
            url = None

        # Chrome: find document with "Chrome_RenderWidgetHostHWND" class name
        # document.IAccessibleObject.accValue(document.IAccessibleChildID)

        # Firefox: role: toolbar name: navigation
        # only combobox, first children value

        log.info(f"Opening A4U ReadAble with url: {url}")
        ui.message(_("Opening the A4U ReadAble. This may take a few seconds."))

        path = f"{globalVars.appArgs.configPath}\\addons\\USA\\lib"
        # Unbind "NVDA+space" shortcut
        # try:
        #     globalCommands.commands.removeGestureBinding("kb:nvda+space")
        # except KeyError:
        #     # Shortcut is not bound, ignore
        #     pass

        launch_arguments = [path + "\\A4UReadAble.exe", "--browser", "true"]

        # Window with camera mode found, pin A4U ReadAble to specific screen
        if i != -1:
            launch_arguments.extend(["--screen", str(i)])

        if url != None:
            launch_arguments.extend(["--source", url])
        else:
            # When starting in "usage manual" mode (without url), it restarts itself and the process handle
            # is therefore not usable.
            # Therefore we manually start it with the manual document as source
            launch_arguments.extend(
                [
                    "--source",
                    f"{globalVars.appArgs.configPath}\\addons\\USA\\a4u-readable-manual_nl.html",
                ]
            )

        self.processes.append(
            subprocess.Popen(
                launch_arguments,
                cwd=path,
            )
        )

        # Disable magnification if fullscreen magnifier engine is active
        if self.settings.is_fullscreen_magnifier_engine():
            self.saved_magnification_level = (
                self.settings.get_magnification_level_by_window(0)
            )
            self.settings.set_magnification_level_by_window(0, 1.0)
            self.settings.UpdateConfig()

        if self.check_thread != None and self.check_thread.is_alive():
            # Thread is already running
            return

        self.check_thread = threading.Thread(target=self._test_for_running_processes)
        self.check_thread.start()
