import addonHandler
import ui

from .._base.settings import USASettings
from .magnifierModes import magnifier_modes
from .windowModification import WindowModification

addonHandler.initTranslation()


# USA Rotor is the user faced name for the NavTool
class NavTool:
    def __init__(self, settings: USASettings, window_modification: WindowModification):
        self.settings = settings
        self.window_modification = window_modification

        self.index = 0
        self.setting_strings = [
            _("Magnification level"),
            _("Color scheme"),
            _("Magnifier mode"),
        ]
        self.option_index = 0
        self.modification_mode_index = 0
        self.modification_modes = [
            _("General settings mode"),
            _("Window resize mode"),
            _("Window move mode"),
        ]
        self.mag_mode_index = 0

    def _is_modification_mode_general_settings(self):
        return self.modification_mode_index == 0

    def _is_modification_mode_resize(self):
        return self.modification_mode_index == 1

    def _is_modification_mode_move(self):
        return self.modification_mode_index == 2

    def _get_magnification_level_options(self):
        return [
            float(ii.displayName)
            for ii in list(self.settings._get_availableMagnificationlevels().values())
        ]

    def _get_color_scheme_options(self):
        return [
            ii
            for ii in list(
                self.settings._get_available_color_schemes_by_window(
                    self.settings._get_current_window_index()
                ).keys()
            )
        ]

    def _get_color_scheme_option_strings(self):
        return [
            ii.displayName
            for ii in list(
                self.settings._get_available_color_schemes_by_window(
                    self.settings._get_current_window_index()
                ).values()
            )
        ]

    def _get_magnifier_mode_options(self):
        return (
            magnifier_modes
            if self.settings.get_show_system_cursor()
            else [m for m in magnifier_modes if m.name not in ["docked", "none"]]
        )

    # Get current option index based on config settings
    def _get_option_index(self, option_values, getter):
        try:
            return option_values.index(getter())
        except ValueError:
            return 0

    def _set_current_setting_option_index(self):
        if self.index == 0:  # Magnification level
            options = self._get_magnification_level_options()
            self.option_index = self._get_option_index(
                options, self.settings.get_magnification_level
            )
        elif self.index == 1:  # Color scheme
            options = self._get_color_scheme_options()
            self.option_index = self._get_option_index(
                options, self.settings.get_color_scheme
            )
        elif self.index == 2:  # Magnifier mode
            options = self._get_magnifier_mode_options()
            current_mode = self.settings.get_magnifier_mode()
            mode_index = next(
                (i for i, x in enumerate(options) if x.name == current_mode)
            )
            # Split modes are index 0 to 3
            # If we are currently in a split mode we dont want to set mag_mode_index
            # as it is either still storing the correct split mode from the last adjustment
            # or returns to 0 as a default value
            if mode_index > 3:
                self.mag_mode_index = mode_index

    def _set_magnifier_mode(self):
        if self.settings.is_fullscreen_magnifier_engine():
            ui.message(_("Cannot adjust magnifier mode in fullscreen engine."))
            return

        m = self._get_magnifier_mode_options()[self.mag_mode_index]
        w = self.settings._get_current_window_json()
        w["mode"] = m.name
        w["area"] = m.host_area.to_dict()
        mag_area_0 = w["magnification-windows"][0]["area"]
        mag_area_1 = w["magnification-windows"][1]["area"]
        mag_area_0["left"] = m.mag_area_0.left
        mag_area_0["top"] = m.mag_area_0.top
        mag_area_0["right"] = m.mag_area_0.right
        mag_area_0["bottom"] = m.mag_area_0.bottom
        mag_area_1["left"] = m.mag_area_1.left
        mag_area_1["top"] = m.mag_area_1.top
        mag_area_1["right"] = m.mag_area_1.right
        mag_area_1["bottom"] = m.mag_area_1.bottom
        ui.message(_("Set magnifier mode to ") + m.display_name)
        self.settings.UpdateConfig()

    # Set option index and save in config setting
    def _set_setting_option(self, option_values, index, setter):
        setter(option_values[index])
        self.settings.UpdateConfig()

    def next_setting(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="navToolNextSetting"
            )
            return
        if self._is_modification_mode_general_settings():
            self.index = (self.index + 1) % len(self.setting_strings)
            ui.message(self.setting_strings[self.index] + _(" setting selected"))
        elif self._is_modification_mode_resize():
            if self.settings.get_magnifier_mode() == "split":
                self.window_modification._move_split_border(1)
            else:
                self.window_modification._resize_lens_window(1)
        elif self._is_modification_mode_move():
            self.window_modification._move_lens_window(1)

    def prev_setting(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="navToolPrevSetting"
            )
            return
        if self._is_modification_mode_general_settings():
            self.index = (self.index - 1) % len(self.setting_strings)
            ui.message(self.setting_strings[self.index] + _(" setting selected"))
        elif self._is_modification_mode_resize():
            if self.settings.get_magnifier_mode() == "split":
                self.window_modification._move_split_border(0)
            else:
                self.window_modification._resize_lens_window(0)
        elif self._is_modification_mode_move():
            self.window_modification._move_lens_window(0)

    def next_option(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="navToolNextOption"
            )
            return
        if self._is_modification_mode_general_settings():
            self._set_current_setting_option_index()
            if self.index == 0:  # Magnification level
                options = self._get_magnification_level_options()
                self.option_index = min(self.option_index + 1, len(options))
                self._set_setting_option(
                    options,
                    self.option_index,
                    self.settings.set_magnification_level,
                )
                ui.message(
                    _("Increased magnification level to ")
                    + str(options[self.option_index])
                )
            elif self.index == 1:  # Color scheme
                options = self._get_color_scheme_options()
                self.option_index = (self.option_index + 1) % len(options)
                self._set_setting_option(
                    options, self.option_index, self.settings.set_color_scheme
                )
                ui.message(
                    _("Set color scheme to ")
                    + self._get_color_scheme_option_strings()[self.option_index]
                )
            elif self.index == 2:  # Magnifier Mode
                options = self._get_magnifier_mode_options()
                self.mag_mode_index = (self.mag_mode_index + 1) % len(options)
                self._set_magnifier_mode()
        elif self._is_modification_mode_resize():
            if self.settings.get_magnifier_mode() == "split":
                self.window_modification._move_split_border(2)
            else:
                self.window_modification._resize_lens_window(2)
        elif self._is_modification_mode_move():
            self.window_modification._move_lens_window(2)

    def prev_option(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="navToolPrevOption"
            )
            return
        if self._is_modification_mode_general_settings():
            self._set_current_setting_option_index()
            if self.index == 0:  # Magnification level
                options = self._get_magnification_level_options()
                self.option_index = max(self.option_index - 1, 0)
                self._set_setting_option(
                    options,
                    self.option_index,
                    self.settings.set_magnification_level,
                )
                ui.message(
                    _("Decreased magnification level to ")
                    + str(options[self.option_index])
                )
            elif self.index == 1:  # Color scheme
                options = self._get_color_scheme_options()
                self.option_index = (self.option_index - 1) % len(options)
                self._set_setting_option(
                    options, self.option_index, self.settings.set_color_scheme
                )
                ui.message(
                    _("Set color scheme to ")
                    + self._get_color_scheme_option_strings()[self.option_index]
                )
            elif self.index == 2:  # Magnifier Mode
                options = self._get_magnifier_mode_options()
                self.mag_mode_index = (self.mag_mode_index - 1) % len(options)
                self._set_magnifier_mode()
        elif self._is_modification_mode_resize():
            if self.settings.get_magnifier_mode() == "split":
                self.window_modification._move_split_border(3)
            else:
                self.window_modification._resize_lens_window(3)
        elif self._is_modification_mode_move():
            self.window_modification._move_lens_window(3)

    def next_modification_mode(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="navToolNextModificationMode"
            )
            return
        self.modification_mode_index = (self.modification_mode_index + 1) % len(
            self.modification_modes
        )
        ui.message(
            _("Modification mode is")
            + f" {self.modification_modes[self.modification_mode_index]}"
        )

    def prev_modification_mode(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="navToolPrevModificationMode"
            )
            return
        self.modification_mode_index = (self.modification_mode_index - 1) % len(
            self.modification_modes
        )
        ui.message(
            _("Modification mode is")
            + f" {self.modification_modes[self.modification_mode_index]}"
        )
