import json
from pathlib import Path
from typing import Dict, Optional

import addonHandler
import config
import globalVars
import ui
from logHandler import log
from synthDriverHandler import SynthDriver, changeVoice, getSynth, setSynth

# Initialize translation and localization support:
addonHandler.initTranslation()


def create_synth_profile(index: int):
    # Retrieve current synthesizer
    synth: Optional[SynthDriver] = getSynth()
    if not synth:
        ui.message(_("Current voice synthesizer not found"))
        log.error("Failed to create synth profile: getSynth() returned None")
        return

    # Get current synthesizer and voice settings
    settings: Dict[str, any] = {}
    for s in synth.supportedSettings:
        settings[s.id] = getattr(synth, s.id)

    settings["synthesizer"] = synth.name
    settings["name"] = synth.availableVoices[synth.voice].displayName

    # Save them to the profile json configuration file
    config = Path(
        f"{globalVars.appArgs.configPath}\\profiles\\voice\\profile{index}.json"
    )
    config.parent.mkdir(exist_ok=True, parents=True)
    config.write_text(json.dumps(settings))

    ui.message(_("Created voice profile {}").format(index))


def announce_synth_voice_name(index: int):
    # Load synthesizer profile from config file
    config_path = Path(
        f"{globalVars.appArgs.configPath}\\profiles\\voice\\profile{index}.json"
    )

    if not config_path.is_file():
        ui.message(_("Voice profile {} does not exit").format(index))
        log.info(f"Failed to announce synth profile: {config_path} is not a valid file")
        return

    settings = json.loads(config_path.read_text())

    try:
        ui.message(settings["name"])
    except KeyError as e:
        log.error(f"Failed to announce synth profile {index} name: {e}")


def load_synth_profile(index: int):
    # Load synthesizer profile from config file
    config_path = Path(
        f"{globalVars.appArgs.configPath}\\profiles\\voice\\profile{index}.json"
    )

    if not config_path.is_file():
        ui.message(_("Voice profile {} does not exit").format(index))
        log.error(f"Failed to load synth profile: {config_path} is not a valid file")
        return

    settings = json.loads(config_path.read_text())

    try:
        # Set synthesizer
        synth_name = settings["synthesizer"]
        if not setSynth(synth_name):
            ui.message(_("Failed to load voice synthesizer") + " " + synth_name)
            log.error(
                f"Failed to load synth profile: setSynth('{synth_name}') returned False"
            )
            return

        synth = getSynth()

        synth.rate = settings["rate"]

        synth.rateBoost = settings["rateBoost"]

        synth.pitch = settings["pitch"]

        synth.volume = settings["volume"]

        changeVoice(synth, settings["voice"])

        # Should save the internal settings to the nvda config.conf settings object
        synth.saveSettings()

        # ======== DO NOT REMOVE THE LINE BELOW - IT IS NEEDED ========

        # With this it works as intended, without this is resets the settings on
        # NVDA settings menu cancel minus the voice settings
        ui.message(
            str(
                synth.availableVoices[
                    config.conf[synth._getConfigSection()][synth.getId()]["voice"]
                ].displayName
            )
        )
        # =============================================================

        # Should save the config.conf settings to the .ini config file
        config.conf.save()

        ui.message(_("Loaded voice profile {}").format(index))

    except KeyError as e:
        ui.message(_("Failed to load voice synthesizer"))
        log.error(f"Failed to load synth profile: {e}")


def delete_synth_profile(index: int):
    config = Path(
        f"{globalVars.appArgs.configPath}\\profiles\\voice\\profile{index}.json"
    )
    try:
        config.unlink()
        ui.message(_("Deleted voice profile {}").format(index))
    except FileNotFoundError:
        ui.message(
            _("Failed to delete voice profile {}. It does not exist").format(index)
        )
