import sys
import threading
from ctypes import *
from ctypes.wintypes import *
from time import sleep

import addonHandler
import api
import browseMode
import locationHelper
import speech
import ui
import winUser
import wx
from logHandler import log
from vision.constants import Context

from .._base.libUSA import libUSA
from .._base.settings import USASettings
from .navTool import NavTool

# Initialize translation and localization support:
addonHandler.initTranslation()


class MousePositionSwapper:
    def __init__(self, settings: USASettings):
        self.settings = settings

        self.cached_mouse_pos = []
        displays = [wx.Display(i).GetGeometry() for i in range(wx.Display.GetCount())]
        for i in range(len(displays)):
            self.cached_mouse_pos.append([])
            self.cached_mouse_pos[i].append(
                int((displays[i][2] * 0.5) + displays[i][0])
            )
            self.cached_mouse_pos[i].append(
                int((displays[i][3] * 0.5) + displays[i][1])
            )

    def swap(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="swapMousePosition"
            )
            return

        displays = [wx.Display(i).GetGeometry() for i in range(wx.Display.GetCount())]

        num_windows = min(self.settings.get_total_amount_windows(), len(displays))

        def getAbsoluteCursorPosition(cursor_pos):

            monitor_x_min = sys.maxsize
            monitor_x_max = -sys.maxsize
            monitor_y_min = sys.maxsize
            monitor_y_max = -sys.maxsize

            for d in displays:
                monitor_x_min = min(monitor_x_min, d[0])
                monitor_x_max = max(monitor_x_max, d[0] + d[2])

                monitor_y_min = min(monitor_y_min, d[1])
                monitor_y_max = max(monitor_y_max, d[1] + d[3])

            if monitor_x_min < 0:
                monitor_x_max -= monitor_x_min
            if monitor_y_min < 0:
                monitor_y_max -= monitor_y_min

            cursor_abs_x = 65535 * (cursor_pos[0] - monitor_x_min) / monitor_x_max
            cursor_abs_y = 65535 * (cursor_pos[1] - monitor_y_min) / monitor_y_max

            return [int(cursor_abs_x), int(cursor_abs_y)]

        new_mouse_pos = winUser.getCursorPos()

        next_index = 0
        for current_index in range(len(displays)):
            display = displays[current_index]
            next_index = (current_index + 1) % num_windows

            mouse_is_in_current_window = self.settings.is_in_rect(
                display, new_mouse_pos
            )

            if mouse_is_in_current_window:
                mouse_input = winUser.Input()
                mouse_input.type = winUser.INPUT_MOUSE
                absolute_cursor_pos = getAbsoluteCursorPosition(
                    self.cached_mouse_pos[next_index]
                )
                mouse_input.ii.mi.dx = absolute_cursor_pos[0]
                mouse_input.ii.mi.dy = absolute_cursor_pos[1]
                MOUSEEVENTF_VIRTUALDESK = 0x4000
                mouse_input.ii.mi.dwFlags = (
                    winUser.MOUSEEVENTF_MOVE
                    | MOUSEEVENTF_VIRTUALDESK
                    | winUser.MOUSEEVENTF_ABSOLUTE
                )
                mouse_input.ii.mi.time = 0
                winUser.SendInput([mouse_input])

                self.cached_mouse_pos[current_index][0] = new_mouse_pos[0]
                self.cached_mouse_pos[current_index][1] = new_mouse_pos[1]
                return
            log.debug(f"Failed to find display for cursor position {new_mouse_pos}")


class Utilities:
    def __init__(self, settings: USASettings, nav_tool: NavTool):
        self.settings = settings
        self.nav_tool = nav_tool
        self.last_tracking_rect = locationHelper.RectLTRB(0, 0, 0, 0)

    def toggle_crosshair(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="toggleCrosshair"
            )
            return
        if self.settings.get_crosshair_enable():
            self.settings.set_crosshair_enable(False)
            ui.message(_("Crosshair disabled"))
        else:
            self.settings.set_crosshair_enable(True)
            ui.message(_("Crosshair enabled"))

        self.settings.UpdateConfig()

    def toggle_mouse_or_focus_for_settings(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="toggleMouseOrFocusForSettings"
            )
            return
        use_mouse_for_getting_window = not self.settings.get_mouse_focus_mode()
        if use_mouse_for_getting_window:
            ui.message(_("Enabling mouse position based window selected mode"))
        else:
            ui.message(_("Disabling mouse position based window selected mode"))
        self.settings.set_mouse_focus_mode(use_mouse_for_getting_window)

    def stop_speech_and_browse_mode(self, gesture):
        if not self.settings.get_object_reader():
            gesture.send()
            return
        # cancel speech as default
        speech.speech.cancelSpeech()

        # turn speech off when relevant
        if not speech.speech._speechState.speechMode == speech.speech.SpeechMode.off:
            log.debugWarning("turning off speech by script_stopSpeechAndBrowseMode")
            speech.speech.setSpeechMode(speech.speech.SpeechMode.off)
            return

        # turn off browse mode
        nav = api.getNavigatorObject()
        if nav.treeInterceptor:
            if isinstance(nav.treeInterceptor, browseMode.BrowseModeTreeInterceptor):
                if not nav.treeInterceptor.passThrough:
                    log.debugWarning(
                        "turning off browseMode by script_stopSpeechAndBrowseMode"
                    )
                    nav.treeInterceptor.passThrough = True
                    nav.treeInterceptor.disableAutoPassThrough = True
                    return

        gesture.send()

    def toggle_magnifier(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="toggleMagnifier"
            )
            return
        enabled = self.settings._toggleMagnifierHelper()
        if enabled:
            levelString = "{:.1f}".format(self.settings.get_magnification_level())
            ui.message(_("Magnifier on, level {level}").format(level=levelString))
        else:
            ui.message(_("Magnifier off"))
        self.settings.UpdateConfig()

    def set_magnification_smoothness_helper(self, increase: bool = True):
        log.debug(
            f"USA: setMagnificationSmoothnessHelper {'increase' if increase else 'decrease'}"
        )
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="magnificationSmoothnessHelper", increase=increase
            )
            return
        level = self.settings._setMagnificationSmoothnessHelper(increase=increase)
        speech.cancelSpeech()

        levelString = f"{level}"
        ui.message(
            _("Magnification smoothness speed {level}%").format(level=levelString)
        )
        self.settings.UpdateConfig()

    def set_magnification_level_helper(self, increase: bool = True):
        log.debug(f"USA: setMagnifierLevel {'increase' if increase else 'decrease'}")

        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="magnificationLevelHelper", increase=increase
            )
            return
        # Sync with NavTool if magnification level settings is activated
        if self.nav_tool.index == 0:
            self.nav_tool.option_index += 1 if increase else -1
        level = self.settings._setMagnificationLevelHelper(increase=increase)
        speech.cancelSpeech()
        levelString = "{:.1f}".format(level)
        self.settings.UpdateConfig()
        ui.message(_("Magnification level {level}").format(level=levelString))

    def toggle_bitmap_smoothing(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="toggleBitmapSmoothing"
            )
            return
        oldState = self.settings.get_bitmap_smoothing()
        self.settings.set_bitmap_smoothing(not oldState)
        bitmap_smoothing = self.settings.get_bitmap_smoothing()
        if bitmap_smoothing == oldState:
            ui.message(_("Bitmap smoothing not available"))
            return
        stateMsg = _("on") if bitmap_smoothing else _("off")
        ui.message(_("Bitmap smoothing {state}").format(state=stateMsg))
        self.settings.UpdateConfig()

    def track_to_rectangle(self, rect, context=""):
        if self.settings.isRemote:
            if isinstance(rect, locationHelper.RectLTWH):
                rect = rect.toLTRB()
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="trackToRectangle", rect=rect, context=context
            )
            return
        if not rect:
            log.warning("No rectangle, unexpected behavior")
            return

        r = rect.toRECT()

        enabled_contexts = [
            key
            for key, value in self.settings.get_tracking_options().items()
            if value is True
        ]

        if context.lower() not in enabled_contexts:
            return

        if self.settings.get_follow_caret() and context in [
            Context.CARET,
            Context.BROWSEMODE,
        ]:
            libUSA.handler.SetTrackingRectangle(r, True, context.lower().encode())
        elif self.settings.get_magnifier_mode() != "resizing-lens" or context not in [
            Context.CARET,
            Context.BROWSEMODE,
        ]:
            if self.last_tracking_rect == rect:
                return
            self.last_tracking_rect = rect
            libUSA.handler.SetTrackingRectangle(r, False, context.lower().encode())

    def toggle_color_scheme(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="toggleColorScheme"
            )
            return

        # Sync with NavTool if color scheme setting is activated
        if self.nav_tool.index == 1:
            self.nav_tool.option_index += 1
        window_index = self.settings._get_current_window_index()
        transformations = list(
            self.settings._get_available_color_schemes_by_window(window_index).values()
        )
        IDs = list(transformation.id for transformation in transformations)
        try:
            index = IDs.index(self.settings.get_color_scheme())
        except ValueError:
            index = 0
        newIndex = (index + 1) % len(IDs)
        self.settings.set_color_scheme(IDs[newIndex])
        self.settings.UpdateConfig()
        # Translators: Reports the newly activated color scheme.

        ui.message(_("Color scheme {name}").format(name=IDs[newIndex]))


class ProcessIterator:
    # Constants
    _TH32CS_SNAPPROCESS = 0x00000002
    _INVALID_HANDLE_VALUE = -1

    # Define the PROCESSENTRY32 structure
    class _PROCESSENTRY32(Structure):
        _fields_ = [
            ("dwSize", DWORD),
            ("cntUsage", DWORD),
            ("th32ProcessID", DWORD),
            ("th32DefaultHeapID", POINTER(ULONG)),
            ("th32ModuleID", DWORD),
            ("cntThreads", DWORD),
            ("th32ParentProcessID", DWORD),
            ("pcPriClassBase", LONG),
            ("dwFlags", DWORD),
            ("szExeFile", c_char * 260),
        ]

    # Load necessary DLLs
    _kernel32 = WinDLL("kernel32", use_last_error=True)
    _psapi = WinDLL("Psapi", use_last_error=True)

    def __init__(self):
        self.watch_thread = None
        self.should_exit = False

    def list_processes(self):
        # Create a snapshot of all processes
        handle_process_snapshot = self._kernel32.CreateToolhelp32Snapshot(
            self._TH32CS_SNAPPROCESS, 0
        )
        if handle_process_snapshot == self._INVALID_HANDLE_VALUE:
            raise Exception("Failed to create process snapshot")

        process_entry = self._PROCESSENTRY32()
        process_entry.dwSize = sizeof(self._PROCESSENTRY32)

        # Retrieve the first process
        if not self._kernel32.Process32First(
            handle_process_snapshot, byref(process_entry)
        ):
            self._kernel32.CloseHandle(handle_process_snapshot)
            raise Exception("Failed to retrieve first process")

        process_list = []

        # Iterate through the processes
        while self._kernel32.Process32Next(
            handle_process_snapshot, byref(process_entry)
        ):
            process_list.append(
                {
                    "pid": process_entry.th32ProcessID,
                    "name": process_entry.szExeFile.decode().lower(),
                }
            )

        # Close the handle
        self._kernel32.CloseHandle(handle_process_snapshot)

        return process_list

    def _watch_for_process(self, process_name: str, callback):
        while not self.should_exit:
            try:
                process_list = self.list_processes()
            except Exception as e:
                log.error(f"Failed to iterate processes: {e}")

            if any(process["name"] == process_name for process in process_list):
                callback()
                # Sleep an additional 10 seconds
                sleep(10)

            sleep(1)

    def run_watch_thread(self, process_name: str, callback):
        self.terminate_watch_thread()

        self.watch_thread = threading.Thread(
            target=self._watch_for_process, args=(process_name, callback)
        )
        self.watch_thread.start()

    def terminate_watch_thread(self):
        if self.watch_thread and self.watch_thread.is_alive():
            self.should_exit = True
            self.watch_thread.join()
            self.should_exit = False
