import addonHandler
import ui
from logHandler import log

from .._base.settings import USASettings

addonHandler.initTranslation()


class WindowModification:
    def __init__(self, settings: USASettings):
        self.settings = settings
        self.position_edit_mode = False
        self.size_edit_mode = False

    def _adjust_border_with_directions(self, moveVertical: bool, subtract_from: bool):
        json = self.settings._get_current_window_json()
        if json["mode"] != "split":
            return

        mag_area_0 = json["magnification-windows"][0]["area"]
        mag_area_1 = json["magnification-windows"][1]["area"]

        # Set strings for vertical or horizontal border adjustment
        if moveVertical:
            border_dir_0 = "top"
            border_dir_1 = "bottom"
        else:
            border_dir_0 = "left"
            border_dir_1 = "right"

        # Index 0 is by default the only window to have magnification.
        # This means that when the mag0's position is emulated to be on mag1's, the border will not be mag0 right/bottom but that of mag1
        mag_0_is_first = mag_area_0[border_dir_0] < mag_area_1[border_dir_0]
        width_or_height = (
            mag_area_1[border_dir_1] if mag_0_is_first else mag_area_0[border_dir_1]
        )
        if mag_area_0[border_dir_0] == mag_area_1[border_dir_0]:
            return

        # The amount of steps that the the border can be adjust in each direction
        number_of_steps = 20
        adjust_by = 0.5 / number_of_steps
        if mag_0_is_first:
            border = (
                mag_area_0[border_dir_1] - adjust_by
                if subtract_from
                else mag_area_0[border_dir_1] + adjust_by
            )
            mag_string_0 = border_dir_1
            mag_string_1 = border_dir_0
        else:
            border = (
                mag_area_1[border_dir_1] - adjust_by
                if subtract_from
                else mag_area_1[border_dir_1] + adjust_by
            )
            mag_string_0 = border_dir_0
            mag_string_1 = border_dir_1

        # clamp to -1 step on each side
        if border == adjust_by or border == width_or_height - adjust_by:
            return

        mag_area_0[mag_string_0] = border
        mag_area_1[mag_string_1] = border
        self.settings.UpdateConfig()

    def _move_lens_window(self, direction: int):
        window_cfg = self.settings.json_conf["usa"]["windows"][
            self.settings._get_current_window_index()
        ]
        if window_cfg["mode"] not in ["docked", "lens", "resizing-lens", "camera"]:
            return
        host_area = window_cfg["area"]
        step_size = 0.05

        width = host_area["right"] - host_area["left"]
        height = host_area["bottom"] - host_area["top"]

        if direction == 0:
            host_area["left"] = max(0.0, host_area["left"] - step_size)
            host_area["right"] = host_area["left"] + width
        elif direction == 1:
            host_area["right"] = min(1.0, host_area["right"] + step_size)
            host_area["left"] = host_area["right"] - width
        elif direction == 2:
            host_area["top"] = max(0.0, host_area["top"] - step_size)
            host_area["bottom"] = host_area["top"] + height
        elif direction == 3:
            host_area["bottom"] = min(1.0, host_area["bottom"] + step_size)
            host_area["top"] = host_area["bottom"] - height

        self.settings.UpdateConfig()

    def _resize_lens_window(self, direction):
        window_cfg = self.settings.json_conf["usa"]["windows"][
            self.settings._get_current_window_index()
        ]
        if window_cfg["mode"] not in ["docked", "lens", "resizing-lens"]:
            return

        host_area = window_cfg["area"]
        step_size = 0.01
        min_size = 0.005

        if direction == 0:
            host_area["right"] = max(
                host_area["left"] + min_size, host_area["right"] - step_size
            )
        elif direction == 1:
            host_area["right"] = min(1.0, host_area["right"] + step_size)
        if direction == 2:
            host_area["bottom"] = max(
                host_area["top"] + min_size, host_area["bottom"] - step_size
            )
        elif direction == 3:
            host_area["bottom"] = min(1.0, host_area["bottom"] + step_size)

        self.settings.UpdateConfig()

    def _move_split_border(self, direction: int):
        if direction == 0:
            self._adjust_border_with_directions(False, True)
        elif direction == 1:
            self._adjust_border_with_directions(False, False)
        elif direction == 2:
            self._adjust_border_with_directions(True, True)
        elif direction == 3:
            self._adjust_border_with_directions(True, False)
        else:
            log.error("Invalid resizeWindow direction")

    # Direction: 0 - left, 1 - right, 2 - top, 3 - bottom
    def resizeWindow(self, direction: int):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="resizeWindow", direction=direction
            )
            return

        if self.settings.is_fullscreen_magnifier_engine():
            ui.message(_("Cannot adjust magnifier mode in fullscreen engine."))
            return

        if self.position_edit_mode:
            self._move_lens_window(direction)
        elif self.size_edit_mode:
            self._resize_lens_window(direction)
        else:
            self._move_split_border(direction)

    def toggle_position_edit_mode(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="positionEditMode"
            )
            return

        if self.settings.is_fullscreen_magnifier_engine():
            ui.message(_("Cannot adjust magnifier mode in fullscreen engine."))
            return

        self.position_edit_mode = not self.position_edit_mode
        state_str = _("enabled") if self.position_edit_mode else _("disabled")
        ui.message(_("{state_str} position edit mode").format(state_str=state_str))

    def toggle_size_edit_mode(self):
        if self.settings.isRemote:
            self.settings.remoteHandler.sendVisionCommandOverUnicorn(
                command="sizeEditMode"
            )
            return

        if self.settings.is_fullscreen_magnifier_engine():
            ui.message(_("Cannot adjust magnifier mode in fullscreen engine."))
            return

        self.size_edit_mode = not self.size_edit_mode
        state_str = _("enabled") if self.size_edit_mode else _("disabled")
        ui.message(_("{state_str} size edit mode").format(state_str=state_str))
