"""
Babbage B.V. 2019-2023

visionEnhancementProviders/usa.py
Contains the USA class. Also implements some functions for zooming in/out with the scroll-wheel and NVDA+mouseclick
"""

import ctypes
import ctypes.wintypes

import addonHandler
import api
import globalCommands
import globalPluginHandler
import globalVars
import gui
import keyboardHandler
import scriptHandler
import speech
import ui
import watchdog
import winInputHook
import wx  # type: ignore
from locationHelper import RectLTRB, RectLTWH
from logHandler import log
from scriptHandler import script
from vision import handler
from vision.util import *

from ._base import USABase, debug
from ._base.dialogs import dialogs
from ._base.HighlighterPlus import HighlighterPlus
from ._base.libUSA import libUSA
from ._base.settings import USASettings
from ._base.wordFollower import objectReader, wordFollower
from ._implementation.applications import A4UReadAble, CamViewer
from ._implementation.navTool import NavTool
from ._implementation.synthesizerProfiles import *
from ._implementation.utilities import MousePositionSwapper, ProcessIterator, Utilities
from ._implementation.windowModification import WindowModification

addonHandler.initTranslation()

oldMouseHook = (
    None  # function that is the mouse hook of NVDA itself. This overwritten by USA
)


def is_key_pressed(virtual_key_code):
    return bool(ctypes.windll.user32.GetKeyState(virtual_key_code) >> 15)


class USA(USABase):
    _settings = USASettings()

    @classmethod
    def getSettingsPanelClass(cls):
        return dialogs.USASettingsPanel

    @classmethod
    def canStart(cls) -> bool:
        log.debug("checking if USA can start")
        try:
            USA_GlobalPlugin = next(
                p
                for p in globalPluginHandler.runningPlugins
                if p.__module__ == "globalPlugins.usa"
            )
        except StopIteration:
            return False
        if not USA_GlobalPlugin:
            log.warn("could not find USA global plugin")
            return False
        if not USA_GlobalPlugin.canStartUSA_VEP:
            log.warn("canStartUSA_VEP false")
            return False

        # isFullScreenMagnificationAvailable is deprecated in NVDA 2024 and not useful anyway. It just returns true.
        fullScreenAvailable = True  # winVersion.isFullScreenMagnificationAvailable()
        if cls.getSettings().isRemote:
            log.debug("isRemote is true, USA can start")
            return True
        if fullScreenAvailable:
            if globalVars.appArgs.secure:
                return True
            isLicensed = watchdog.cancellableExecute(libUSA.handler.IsLicensed)
            return isLicensed
        return False

    def initialize_libUSA(self):
        try:
            libUSA.handler.initialize_libUSA()
        except Exception as e:
            log.error(f"could not load or initialize libUSA.dll. \nReturned error: {e}")
            gui.messageBox(
                _("An error has occured:\n{error}").format(
                    error=_(
                        "Could not load the libUSA  dll. Please check the log or contact support."
                    )
                ),
                _("Error"),
                wx.OK | wx.ICON_WARNING,
            )
            raise e
        if not libUSA.handler.IsLicensed():
            log.info("returning early due to unlicensed USA")
            return

        libUSA.handler.Initialize()
        self.getSettings().UpdateConfig()

    def __init__(self):
        log.info("starting USA")
        addonManifest = addonHandler.getCodeAddon().manifest
        log.info(f"USA version: {addonManifest['version']}")

        self.getSettings().initialize_json()
        self.getSettings().post_init_Settings()
        if not self.getSettings().isRemote:
            log.info("loading and initializing libUSA")
            if not libUSA.handler.initialized:
                self.initialize_libUSA()

        self.hookOurMouseFunction()

        super().__init__()
        self.objectReader = objectReader.objectReader(self)
        self.wordFollower = wordFollower.wordFollower(self)
        self.highlighterPlus = HighlighterPlus(self)

        self.window_modification = WindowModification(self.getSettings())
        self.nav_tool = NavTool(self.getSettings(), self.window_modification)
        self.utilities = Utilities(self.getSettings(), self.nav_tool)
        self.cam_viewer = CamViewer(self.getSettings())
        self.a4u_readable = A4UReadAble(self.getSettings())
        self.mouse_position_swapper = MousePositionSwapper(self.getSettings())
        self.process_iterator = ProcessIterator()
        self.process_iterator.run_watch_thread(
            "magnify.exe",
            lambda: ui.message(_("Please terminate the Windows Magnifier")),
        )

        providers = handler.getActiveProviderInfos()
        for provider in providers:
            if provider.providerId == "NVDAHighlighter":
                self.getSettings().set_highlighter_plus(False)

        log.info("usa init complete")

    def terminate(self):
        log.info("terminating USA")
        self.unhookOurMouseFunction()
        self.wordFollower.terminate()
        self.objectReader.terminate()
        self.highlighterPlus.terminate()
        self.process_iterator.terminate_watch_thread()
        super().terminate()
        libUSA.handler.terminate()
        log.info("USA terminated")

    def handleForegroundChange(self, obj):
        log.debug(f"USA: handleForegroundChange {obj}")
        self.wordFollower.handleForegroundChange(obj)
        self.objectReader.handleForegroundChange(obj)

        # Foreground dependent color schemes are not supported in USA 2025 (yet)
        return

        # for the focus dependent color schemes we need to change the color scheme if the appName is in the custom made
        # aggerated section of conf
        appName = api.getForegroundObject().appModule.appName
        if self.getSettings().conf["fgColorschemes"].get(appName):
            self.getSettings()._set_colorScheme(
                self.getSettings().conf["fgColorschemes"].get(appName)
            )

        super().handleForegroundChange(obj)

    def handleReviewMove(self, context):
        log.debug(f"USA: handleReviewMove {context}")
        self.wordFollower.handleReviewMove(context)
        self.objectReader.handleReviewMove(context)
        self.highlighterPlus.handleReviewMove()
        super().handleReviewMove(context)

    def handleFocusChange(self, obj):
        log.debug(f"USA: handleFocusChange {obj}")
        self.wordFollower.handleFocusChange(obj)
        self.objectReader.handleFocusChange(obj)
        self.highlighterPlus.handleFocusChange(obj)
        super().handleFocusChange(obj)

    def handleBrowseModeMove(self, obj):
        log.debug(f"USA: handleBrowseModeMove {obj}")
        self.highlighterPlus.handleBrowseModeMove()
        super().handleBrowseModeMove(obj)

    def handleUpdateThread(self):
        log.debug(f"USA: handleUpdateThread")
        self.highlighterPlus.handleUpdateThread()
        super().handleUpdateThread()

    # @script(gesture="kb:alt+NVDA+upArrow", description=_("Scroll magnification area up"))
    # def script_scrollUp(self, gesture):
    # 	self._scrollScriptHelper( yDir=-1)

    # @script(gesture="kb:alt+NVDA+downArrow", description=_("Scroll magnification area down"))
    # def script_scrollDown(self, gesture):
    # 	self._scrollScriptHelper( yDir=1)

    # @script(gesture="kb:alt+NVDA+leftArrow", description=_("Scroll magnification area left"))
    # def script_scrollLeft(self, gesture):
    # 	self._scrollScriptHelper( xDir=-1)

    # @script(gesture="kb:alt+NVDA+rightArrow", description=_("Scroll magnification area right"))
    # def script_scrollRight(self, gesture):
    # 	self._scrollScriptHelper( xDir=1)

    @script(
        gesture="kb:NVDA+alt+r",
        description=_("Start reading all from object beneath the mouse"),
    )
    def script_readAtMousePos(self, gesture):
        log.debugWarning(
            f"USA: script_readAtMousePos \n \t self.objectReader.enabled = {self.objectReader.enabled}"
        )

        self.objectReader.readAllUnderMouse()

    @script(description=_("Toggles Word Follower of USA"))
    def script_toggleFollower(self, gesture):
        # script to toggle on/off the word follower
        settings = self.getSettings()
        follow_word_enabled = settings.get_word_follower()
        if follow_word_enabled:
            self.wordFollower.disable_wordFollower(False)
        else:
            self.wordFollower.enable_wordFollower(False)
        settings.set_word_follower(not follow_word_enabled)
        follow_word_enabled = settings.get_word_follower()
        debug(
            f"USA: script_toggleFollower \n \t settings[follow-word][enabled] = {follow_word_enabled}"
        )
        if follow_word_enabled:
            ui.message(_("Word Follower enabled"))
        elif not follow_word_enabled:
            ui.message(_("Word Follower disabled"))

    @script(
        gesture="kb:NVDA+alt+w", description=_("Toggles Word Follower globally of USA")
    )
    def script_toggleFollowerGlobal(self, gesture):
        # script to toggle on/off the word follower
        settings = self.getSettings()
        self.wordFollower.disable_wordFollower(False)

        settings.set_word_follower_global(not settings.get_word_follower_global())

        word_follower_global = settings.get_word_follower_global()
        if settings.get_word_follower() or word_follower_global:
            self.wordFollower.enable_wordFollower(False)
        if word_follower_global:
            ui.message(_("Global Word Follower enabled"))
        elif not word_follower_global:
            ui.message(_("Global Word Follower disabled"))
        settings.UpdateConfig()
        debug(
            f"USA: script_toggleFollowerGlobal \n \t settings[follow-word][enabled-global] = {word_follower_global}"
        )

    @script(description=_("Toggles rectangle for Word Follower of USA"))
    def script_toggleFollowerRect(self, gesture):
        # script to toggle on/off the rectangle for the word follower, only changes if wordfollower itself is enabled
        settings = self.getSettings()
        conf = settings.conf
        if settings.get_word_follower():
            if settings.get_word_follower_show_rect():
                self.wordFollower.disable_wordFollowerRect()
            else:
                self.wordFollower.enable_wordFollowerRect()
            settings.set_word_follower_show_rect(
                not settings.get_word_follower_show_rect()
            )

        debug(
            f"USA: script_toggleFollowerRect \n \t settings[follow-word][enabled] = {settings.get_word_follower()} \n \t settings[word-follower][show-rect] = {settings.get_word_follower_show_rect()}"
        )

    # @script(
    #     gesture="kb:NVDA+control+shift+p",
    #     description=_("Set current color scheme as default for current application"),
    # )
    # def script_setFGColorScheme(self, gesture):
    #     debug(
    #         f"USA: script_setFGColorScheme \n \t appName = {api.getForegroundObject().appModule.appName} \n \t colorScheme = {self.getSettings().colorScheme}"
    #     )
    #     # save the current color scheme for the current fg application, which is then used for the focus dependent color
    #     # scheme feature
    #     appName = api.getForegroundObject().appModule.appName
    #     self.getSettings().conf["fgColorschemes"][
    #         appName
    #     ] = self.getSettings().colorScheme
    #     ui.message(_("set current color scheme for this app as default"))

    def getUnicornCommandMapping(self):
        mapping = {
            "toggleColorScheme": self.utilities.toggle_color_scheme,
            "trackToRectangle": self.utilities.track_to_rectangle,
            # "reportOffsets": self.reportOffsets,
            "toggleBitmapSmoothing": self.utilities.toggle_bitmap_smoothing,
            "magnificationLevelHelper": self.utilities.set_magnification_level_helper,
            "magnificationSmoothnessHelper": self.utilities.set_magnification_smoothness_helper,
            "scrollScriptHelper": self._scrollScriptHelper,
            "toggleMagnifier": self.utilities.toggle_magnifier,
            "switchMagnifierMode": self.getSettings().switchMagnifierMode,
            "swapMousePosition": self.mouse_position_swapper.swap,
            "toggleMouseOrFocusForSettings": self.utilities.toggle_mouse_or_focus_for_settings,
            "toggleCrosshair": self.utilities.toggle_crosshair,
            "resizeWindow": self.window_modification.resizeWindow,
            "positionEditMode": self.window_modification.toggle_position_edit_mode,
            "sizeEditMode": self.window_modification.toggle_size_edit_mode,
            "navToolNextModificationMode": self.nav_tool.next_modification_mode,
            "navToolPrevModificationMode": self.nav_tool.prev_modification_mode,
            "navToolNextSetting": self.nav_tool.next_setting,
            "navToolPrevSetting": self.nav_tool.prev_setting,
            "navToolNextOption": self.nav_tool.next_option,
            "navToolPrevOption": self.nav_tool.prev_option,
        }
        return mapping

    def deserializeUnicornArguments(self, command, **kwargs):
        if command == "trackToRectangle":
            context = kwargs["context"]
            log.info(f" {kwargs['rect'] } ")
            rect = RectLTRB(*kwargs["rect"])
            return {"rect": rect, "context": context}
        elif command == "scrollScriptHelper":
            xDir = kwargs["xDir"]
            yDir = kwargs["yDir"]
            return {"xDir": xDir, "yDir": yDir}
        elif command in ("magnificationLevelHelper", "magnificationSmoothnessHelper"):
            increase = kwargs["increase"]
            return {"increase": increase}
        elif command == "resizeWindow":
            return {"direction": kwargs["direction"]}

    @script(
        gestures=["kb:alt+NVDA+backspace", "kb:control+backspace"],
        description=_("Cycles between the available color schemes"),
    )
    def script_toggleColorScheme(self, gesture):
        log.debug("USA: toggleColorScheme")
        self.utilities.toggle_color_scheme()

    # def reportOffsets(self):
    #     if self.getSettings().isRemote:
    #         self.getSettings().remoteHandler.sendVisionCommandOverUnicorn(
    #             command="reportOffsets"
    #         )
    #     else:
    #         point = libUSA.handler.GetTrackingOffsets(None)
    #         left = point.x
    #         top = point.y
    #         ui.message(str((left, top)))

    # @script(gesture="kb:NVDA+i", description=_("reports magnification offset"))
    # def script_reportOffsets(self, gesture):
    #     self.reportOffsets()

    @script(
        gesture="kb:alt+NVDA+d", description=_("Toggles the state of bitmap smoothing")
    )
    def script_toggleBitmapSmoothing(self, gesture):
        self.utilities.toggle_bitmap_smoothing()

    def _scrollScriptHelper(self, xDir=0, yDir=0):
        if self.getSettings().isRemote:
            self.getSettings().remoteHandler.sendVisionCommandOverUnicorn(
                command="scrollScriptHelper", xDir=xDir, yDir=yDir
            )
        else:
            currentRect = self.last_tracking_rect

            display = wx.Display(
                self.getSettings()._get_current_window_index()
            ).GetGeometry()
            screen = RectLTWH(
                display.GetLeft(),
                display.GetTop(),
                display.GetWidth(),
                display.GetHeight(),
            )

            xSize = round((screen.right - screen.left) / 20)
            ySize = round((screen.bottom - screen.top) / 20)
            left, top, right, bottom = currentRect

            rect_width = right - left
            rect_height = bottom - top

            if xDir > 0:
                right = min(screen.right, currentRect.right + xSize)
                left = right - rect_width
            elif xDir < 0:
                left = max(screen.left, currentRect.left - xSize)
                right = left + rect_width
            if yDir > 0:
                bottom = min(screen.bottom, currentRect.bottom + ySize)
                top = bottom - rect_height
            elif yDir < 0:
                top = max(screen.top, currentRect.top - ySize)
                bottom = top + rect_height
            newRect = RectLTRB(left, top, right, bottom)
            if newRect == currentRect:
                ui.message(_("Edge of screen"))
                return
            self.utilities.track_to_rectangle(newRect, Context.NAVIGATOR)

    def hookOurMouseFunction(self):
        # this function replaces the mousehook in NVDA with our own mousehook. we cannot use the mouseCallback since we
        # wont get the mousedata that we need (whether the scrollwheel goes up or down)
        global oldMouseHook
        oldMouseHook = winInputHook.mouseHook
        winInputHook.mouseHook = mouseHook2

        wx.CallLater(200, resetWinInputHook)

    def unhookOurMouseFunction(self):
        global oldMouseHook
        winInputHook.mouseHook = oldMouseHook

        wx.CallLater(200, resetWinInputHook)

    @script(
        gestures=["kb:alt+NVDA+-", "kb:control+pageDown", "kb:control+-"],
        description=_("Decreases the magnification level"),
    )
    def script_decreaseMagnification(self, gesture):
        self.utilities.set_magnification_level_helper(increase=False)

    # kb:alt+NVDA+shift+0 (plus on english keyboard) and alt+NVDA+plus is for german keyboard users
    @script(
        gestures=["kb:alt+NVDA+=", "kb:control+pageUp", "kb:control+="],
        description=_("Increases the magnification level"),
    )
    def script_increaseMagnification(self, gesture):
        self.utilities.set_magnification_level_helper(increase=True)

    @script(
        gestures=["kb:NVDA+control+="],
        description=_("Increases magnification smoothness"),
    )
    def script_increaseMagnificationSmoothness(self, gesture):
        self.utilities.set_magnification_smoothness_helper(increase=True)

    @script(
        gestures=["kb:NVDA+control+-"],
        description=_("Decreases magnification smoothness"),
    )
    def script_decreaseMagnificationSmoothness(self, gesture):
        self.utilities.set_magnification_smoothness_helper(increase=False)

    @script(
        gestures=["kb:alt+NVDA+0", "kb:control+0"],
        description=_("Toggles the magnifier"),
    )
    def script_toggleMagnifier(self, gesture):
        self.utilities.toggle_magnifier()

    @script(description=_("Stop speech and browsemode if applicable."))
    def script_stopSpeechAndBrowseMode(self, gesture):
        log.debugWarning("script_stopSpeechAndBrowseMode called")
        self.utilities.stop_speech_and_browse_mode(gesture)

    @script(
        gestures=[
            "kb:NVDA+downArrow",
        ],
        description=_("Sayall command in USA"),
    )
    def script_sayAllUSA(self, gesture):
        log.info("our own Sayall is called!")
        speech.speech.setSpeechMode(speech.speech.SpeechMode.talk)
        globalCommands.commands.script_sayAll(gesture)

    @script(
        gestures=[
            "kb:NVDA+Control+Alt+leftArrow",
        ],
        description=_("Magnification window modification: Left"),
    )
    def script_resizeWindowLeft(self, gesture):
        self.window_modification.resizeWindow(0)

    @script(
        gestures=[
            "kb:NVDA+Control+Alt+rightArrow",
        ],
        description=_("Magnification window modification: Right"),
    )
    def script_resizeWindowRight(self, gesture):
        self.window_modification.resizeWindow(1)

    @script(
        gestures=[
            "kb:NVDA+Control+Alt+upArrow",
        ],
        description=_("Magnification window modification: Up"),
    )
    def script_resizeWindowUp(self, gesture):
        self.window_modification.resizeWindow(2)

    @script(
        gestures=[
            "kb:NVDA+Control+Alt+downArrow",
        ],
        description=_("Magnification window modification: Down"),
    )
    def script_resizeWindowDown(self, gesture):
        self.window_modification.resizeWindow(3)

    @script(
        gestures=[
            "kb:NVDA+Alt+f",
        ],
        description=_("Switch magnifier mode"),
    )
    def script_switchMagnifierMode(self, gesture):
        # Keep Nav-Tool magnifier mode index in sync if its in magnifier mode setting
        if self.nav_tool.index == 2:
            self.nav_tool.option_index += 1

        self.getSettings().switchMagnifierMode()

    @script(
        gestures=[
            "kb:NVDA+alt+tab",
        ],
        description=_("Switch cursor position between windows"),
    )
    def script_swapMousePosition(self, gesture):
        self.mouse_position_swapper.swap()

    @script(
        gestures=[
            "kb:NVDA+alt+h",
        ],
        description=_(
            "Switches whether the mouse's or focus object's position are used to adjust settings"
        ),
    )
    def script_toggleMouseOrFocusForSettings(self, gesture):
        self.utilities.toggle_mouse_or_focus_for_settings()

    @script(
        gestures=[
            "kb:NVDA+alt+e",
        ],
        description=_("Toggle the crosshair"),
    )
    def script_toggleCrosshair(self, gesture):
        self.utilities.toggle_crosshair()

    @script(
        gesture="kb:NVDA+alt+p", description=_("Toggle magnifier position edit mode")
    )
    def script_positionEditMode(self, gesture):
        self.window_modification.toggle_position_edit_mode()

    @script(
        gesture="kb:NVDA+control+alt+r",
        description=_("Toggle magnifier size edit mode"),
    )
    def script_sizeEditingMode(self, gesture):
        self.window_modification.toggle_size_edit_mode()

    @script(gesture="kb:NVDA+alt+y", description=_("Start Cam-Viewer"))
    def script_startCamViewer(self, gesture):
        self.cam_viewer.start()

    @script(gesture="kb:NVDA+shift+alt+y", description=_("Stop Cam-Viewer"))
    def script_stopCamViewer(self, gesture):
        self.cam_viewer.stop()

    @script(gesture="kb:NVDA+alt+i", description=_("Select next Cam-Viewer position"))
    def script_nextCamViewerPosition(self, gesture):
        self.cam_viewer.set_next_position()

    @script(
        gesture="kb:NVDA+control+alt+i", description=_("Switch Cam-Viewer orientation")
    )
    def script_switchCamViewerOrientation(self, gesture):
        self.cam_viewer.switch_orientation()

    @script(gesture="kb:NVDA+alt+u", description=_("Open A4U ReadAble"))
    def script_openA4UReadAble(self, gesture):
        self.a4u_readable.start()

    @script(
        gesture="kb:NVDA+alt+rightArrow",
        description=_("Go to next setting in USA Rotor"),
    )
    def script_navToolNextSetting(self, gesture):
        self.nav_tool.next_setting()

    @script(
        gesture="kb:NVDA+alt+LeftArrow",
        description=_("Go to previous setting in USA Rotor"),
    )
    def script_navToolPrevSetting(self, gesture):
        self.nav_tool.prev_setting()

    @script(
        gesture="kb:NVDA+alt+UpArrow",
        description=_("Go to next option of current setting in USA Rotor"),
    )
    def script_navToolNextOption(self, gesture):
        self.nav_tool.next_option()

    @script(
        gesture="kb:NVDA+alt+DownArrow",
        description=_("Go to previous option of current setting in USA Rotor"),
    )
    def script_navToolPrevOption(self, gesture):
        self.nav_tool.prev_option()

    @script(
        gesture="kb:NVDA+alt+PageDown",
        description=_("Go to next settings modification mode"),
    )
    def script_nextModificationMode(self, gesture):
        self.nav_tool.next_modification_mode()

    @script(
        gesture="kb:NVDA+alt+PageUp",
        description=_("Go to previous settings modification mode"),
    )
    def script_prevModificationMode(self, gesture):
        self.nav_tool.prev_modification_mode()

    @script(
        description=_("Set voice synthesizer profile"),
        gestures=[
            "kb:NVDA+control+1",
            "kb:NVDA+control+2",
            "kb:NVDA+control+3",
            "kb:NVDA+control+4",
            "kb:NVDA+control+5",
            "kb:NVDA+control+6",
            "kb:NVDA+control+7",
            "kb:NVDA+control+8",
            "kb:NVDA+control+9",
            "kb:NVDA+control+0",
            "kb:NVDA+control+numpad1",
            "kb:NVDA+control+numpad2",
            "kb:NVDA+control+numpad3",
            "kb:NVDA+control+numpad4",
            "kb:NVDA+control+numpad5",
            "kb:NVDA+control+numpad6",
            "kb:NVDA+control+numpad7",
            "kb:NVDA+control+numpad8",
            "kb:NVDA+control+numpad9",
            "kb:NVDA+control+numpad0",
        ],
    )
    def script_setSynthProfile(self, gesture):
        profile_index = int(gesture._get_displayName()[-1])
        script_count = scriptHandler.getLastScriptRepeatCount()

        # Add profile
        if script_count == 0:
            create_synth_profile(profile_index)
        # Remove profile
        elif script_count == 1:
            delete_synth_profile(profile_index)

    def load_synth_profile(self, gesture):
        profile_index = int(gesture._get_displayName()[-1])
        script_count = scriptHandler.getLastScriptRepeatCount()

        # Announce profile voice name
        if script_count == 0:
            announce_synth_voice_name(profile_index)
        # Load profile
        elif script_count == 1:
            load_synth_profile(profile_index)

    descriptionString = _("Load voice synthesizer profile")

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+1", "kb:NVDA+shift+numpad1"],
    )
    def script_loadSynthProfile1(self, gesture):
        self.load_synth_profile(gesture)

    descriptionString = _("Load voice synthesizer profile")

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+2", "kb:NVDA+shift+numpad2"],
    )
    def script_loadSynthProfile2(self, gesture):
        self.load_synth_profile(gesture)

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+3", "kb:NVDA+shift+numpad3"],
    )
    def script_loadSynthProfile3(self, gesture):
        self.load_synth_profile(gesture)

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+4", "kb:NVDA+shift+numpad4"],
    )
    def script_loadSynthProfile4(self, gesture):
        self.load_synth_profile(gesture)

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+5", "kb:NVDA+shift+numpad5"],
    )
    def script_loadSynthProfile5(self, gesture):
        self.load_synth_profile(gesture)

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+6", "kb:NVDA+shift+numpad6"],
    )
    def script_loadSynthProfile6(self, gesture):
        self.load_synth_profile(gesture)

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+7", "kb:NVDA+shift+numpad7"],
    )
    def script_loadSynthProfile7(self, gesture):
        self.load_synth_profile(gesture)

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+8", "kb:NVDA+shift+numpad8"],
    )
    def script_loadSynthProfile8(self, gesture):
        self.load_synth_profile(gesture)

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+9", "kb:NVDA+shift+numpad9"],
    )
    def script_loadSynthProfile9(self, gesture):
        self.load_synth_profile(gesture)

    @script(
        description=descriptionString,
        gestures=["kb:NVDA+shift+0", "kb:NVDA+shift+numpad0"],
    )
    def script_loadSynthProfile0(self, gesture):
        self.load_synth_profile(gesture)


# get the NVDA keys beforehand toe make the code in the hook just a bit quicker.
NVDAKeys = keyboardHandler.getNVDAModifierKeys()


def resetWinInputHook():
    # only when hookThreadRefCount is 0, it makes sense to initialize wnnInputHook
    while winInputHook.hookThreadRefCount > 0:
        log.debugWarning(
            f"USA: terminating hookthread. current amount: {winInputHook.hookThreadRefCount}"
        )
        winInputHook.terminate()

    log.debugWarning(f"USA: re-initialize winInputHook")
    winInputHook.initialize()


# function used in mouseHook to check if an NVDA key is pressed
def checkIfNVDAKeyIsPressed():
    global NVDAKeys
    for ii in NVDAKeys:
        # the NVDAKeys and currentModifiers are tuples with the first element as vkCode
        if ii[0] in [kk[0] for kk in keyboardHandler.currentModifiers]:
            return True
    return False


# windows hook functie om het scroll wiel event op te vangen en USA in te laten zoomen als een toets + scroll wheel
# ingedrukt wordt. Deze functie werkt niet perfect in python en het kan zijn dat het af en toe de toetscombi + scrollen
# niet afvangt. Dit is een kopie met wat andere functies toegevoegd van de winInputHook variant die ook in NVDA zit.
@ctypes.WINFUNCTYPE(
    ctypes.c_long, ctypes.c_int, ctypes.wintypes.WPARAM, ctypes.wintypes.LPARAM
)
def mouseHook2(code, wParam, lParam):
    global USA_Instance
    try:
        HC_ACTION = 0
        # als NVDA vastloopt dan moet dit gewoon doorgaan
        if watchdog.isAttemptingRecovery or code != HC_ACTION:
            return oldMouseHook(code, wParam, lParam)

        USA_Instance = findUSAInstance()
        if not USA_Instance:
            return oldMouseHook(code, wParam, lParam)

        # This is before the other nvdakey check as we only want to check for MiddleMouseButton press
        if wParam == 519 and checkIfNVDAKeyIsPressed():
            wx.CallAfter(USA_Instance.swapMousePosition)
            return 1

        if not checkIfNVDAKeyIsPressed():
            return oldMouseHook(code, wParam, lParam)

        # controleert off dit een mousewheelevent is of clickevent
        if not wParam in (522, 514, 513):
            return oldMouseHook(code, wParam, lParam)

        if wParam == 522:
            msll = winInputHook.MSLLHOOKSTRUCT.from_address(lParam)
            if int(msll.mouseData) == 7864320:
                wx.CallAfter(scrollZoom, increase=True)
                return 1
            elif int(msll.mouseData) == 4287102976:
                wx.CallAfter(scrollZoom, increase=False)
                return 1
        if wParam == 513:
            return 1
        if wParam == 514:
            log.debugWarning(f"USA: mouseHook2; should execute callReadAtMousePos")
            wx.CallLater(200, callReadAtMousePos)
            return 1

    except Exception as e:
        return oldMouseHook(code, wParam, lParam)

    # sometimes the code gets here and I do not know how. Maybe because of the slowness of python. Just return 1 here
    # to prevent scroll events from going through when we do not want to
    return 1


def scrollZoom(increase=False):
    # deze functie is appart van de _setMagnificaitonLevelHelper vanwege wat problemen met wx.
    global USA_Instance
    USA_Instance = findUSAInstance()
    log.debugWarning(
        f"USA: scrollZoom; call {USA_Instance} method with increase {increase}"
    )
    USA_Instance.utilities.set_magnification_level_helper(increase)


def callReadAtMousePos():
    global USA_Instance
    USA_Instance = findUSAInstance()
    log.debugWarning(
        f"USA: callReadAtMousePos; call {USA_Instance} method to read from mouse pos."
    )
    wx.CallAfter(USA_Instance.script_readAtMousePos, None)


USA_Instance = None


def findUSAInstance():
    global USA_Instance
    if USA_Instance:
        return USA_Instance

    try:
        import vision

        instances = vision.handler.getActiveProviderInstances()
        for ii in instances:
            if ii.getSettings().getId().lower().strip() == "usa":
                return ii
    except:
        pass
    log.error("USA: findUSAInstance; could not find USA instance.")
    return None


VisionEnhancementProvider = USA
